// Copyright (C) 2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

// * The software is provided "as is" without any warranty, neither express
//   nor implied.
// * In no event will the author be held liable for any damage(s) arising
//   from the use of the software.
// * Redistribution of the software is allowed only in unmodified form.
// * Permission is granted to use the software solely for personal private
//   and non-commercial purposes.
// * An individuals use of the software in his or her capacity or function
//   as an agent, (independent) contractor, employee, member or officer of
//   a business, corporation or organization (commercial or non-commercial)
//   does not qualify as personal private and non-commercial purpose.
// * Without written approval from the author the software must not be used
//   for a business, for commercial, corporate, governmental, military or
//   organizational purposes of any kind, or in a commercial, corporate,
//   governmental, military or organizational environment of any kind.

// Minimalist Win32 Runtime Library for Microsoft C Compiler 20xy

// https://msdn.microsoft.com/en-us/library/wbeh5h91.aspx

#pragma include_alias(<ctype.h>, <NUL:>)
#pragma include_alias(<stdarg.h>, <NUL:>)
#pragma include_alias(<stddef.h>, <NUL:>)
#pragma include_alias(<stdint.h>, <NUL:>)
#pragma include_alias(<stdio.h>, <NUL:>)
#pragma include_alias(<stdlib.h>, <NUL:>)
#pragma include_alias(<string.h>, <NUL:>)

// https://msdn.microsoft.com/en-us/library/4141z1cx.aspx

#pragma once

#ifndef __NOMSVCRT_H__
#define __NOMSVCRT_H__

// https://msdn.microsoft.com/en-us/library/7f0aews7.aspx

#pragma comment(compiler)
#pragma comment(linker, "/DEFAULTLIB:NOMSVCRT.LIB")
#pragma comment(linker, "/DISALLOWLIB:LIBCMT.LIB")
#pragma comment(linker, "/DISALLOWLIB:LIBCMTD.LIB")
#pragma comment(linker, "/DISALLOWLIB:MSVCRT.LIB")
#pragma comment(linker, "/DISALLOWLIB:MSVCRTD.LIB")
#pragma comment(linker, "/DISALLOWLIB:OLDNAMES.LIB")
#pragma comment(user, __TIMESTAMP__)

// https://msdn.microsoft.com/en-us/library/b0084kay.aspx

#ifdef __MSVC_RUNTIME_CHECKS
#pragma comment(linker, "/DEFAULTLIB:RUNTMCHK.LIB")
#pragma message("Runtime checks via '/RTC*' compiler switch not supported!")
#endif

#ifndef _VC_NODEFAULTLIB
#pragma message("Use '/Zl' compiler switch!")
#endif

#ifndef IEEE_754
#ifdef _M_IX86
#if _M_IX86_FP > 1
#pragma message("Use '/arch:IA32' or '/arch:SSE' compiler switch; don't use '/arch:SSE2', '/arch:AVX' or '/arch:AVX2' compiler switch!")
#endif
#endif // _M_IX86
#else
#ifdef _M_IX86
#if _M_IX86_FP > 0
#pragma message("Use '/arch:IA32' compiler switch; don't use '/arch:SSE', '/arch:SSE2', '/arch:AVX' or '/arch:AVX2' compiler switch!")
#endif
#ifdef _M_FP_STRICT
#pragma message("Use '/fp:fast' or '/fp:precise', plus '/Oi', '/Qfast_transcendentals' and '/QIfist' compiler switches!")
#else
#pragma message("Use '/Oi', '/Qfast_transcendentals' and '/QIfist' compiler switches!")
#endif
#endif // _M_IX86
#endif // IEEE_754

// https://msdn.microsoft.com/en-us/library/6yh4a9k1.aspx

#define	__thread	__declspec(thread)

// https://msdn.microsoft.com/en-us/library/k15zsh48.aspx

#define WCHAR_MAX	(WCHAR) ~0			// = 65535
#define WCHAR_MIN	(WCHAR) 0

#ifndef SHRT_MAX
#define SHRT_MAX	(SHORT) ((USHORT) ~0 >> 1)	// = 32767
#endif
#ifndef SHRT_MIN
#define SHRT_MIN	(SHORT) ~((USHORT) ~0 >> 1)	// = -32768
#endif
#ifndef USHRT_MAX
#define USHRT_MAX	(USHORT) ~0			// = 65535
#endif
#define USHRT_MIN	(USHORT) 0

#ifndef INT_MAX
#define INT_MAX		(INT) (~0U >> 1)		// = 2147483647
#endif
#ifndef INT_MIN
#define INT_MIN		(INT) ~(~0U >> 1)		// = -2147483648
#endif
#ifndef UINT_MAX
#define UINT_MAX	~0U				// = 4294967295U
#endif
#define UINT_MIN	0U

#ifndef LONG_MAX
#define LONG_MAX	(LONG) (~0UL >> 1)		// = 2147483647L
#endif
#ifndef LONG_MIN
#define LONG_MIN	(LONG) ~(~0UL >> 1)		// = -2147483648L
#endif
#ifndef ULONG_MAX
#define ULONG_MAX	~0UL				// = 4294967295UL
#endif
#define ULONG_MIN	0UL

#ifndef LLONG_MAX
#define LLONG_MAX	(LONGLONG) (~0ULL >> 1)		// = 9223372036854775807LL
#endif
#ifndef LLONG_MIN
#define LLONG_MIN	(LONGLONG) ~(~0ULL >> 1)	// = -9223372036854775808LL
#endif
#ifndef ULLONG_MAX
#define ULLONG_MAX	~0ULL				// = 18446744073709551615ULL
#endif
#define ULLONG_MIN	0ULL

#ifdef IEEE_754
#define FLT_EVAL_METHOD	-2

// https://msdn.microsoft.com/en-us/library/6bs3y5ya.aspx
// https://msdn.microsoft.com/en-us/library/0b34tf65.aspx

#define FLT_RADIX	2
#if _MSC_VER < 1914 // MSC 19.14 alias 2017
#define FLT_DENORMAL	1.4012984643248171e-45F	// = 2**-149
//			1.1754942106924411e-38F	// = 2**-126 * (1 - 2**-24)
#define FLT_MINIMUM	1.1754943508222875e-38F	// = 2**-126
#define FLT_EPSILON	1.1920928955078125e-7F	// = 2**-23	EXACT: 2**-n has n fractional decimal digits!
#define FLT_OMICRON	9.9999994039535522e-1F	// = 1 - 2**-24
#define FLT_YPSILON	8388608.0F		// = 2**+23
#define FLT_MAXIMUM	3.4028234663852886e+38F	// = 2**+128 * (1 - 2**-24)
#if 0	// NOTE: MSC yields error C2124 for 1.0F / 0.0F and 0.0F / 0.0F!
#define FLT_INFINITY	(1.0F / 0.0F)
#define FLT_INDEFINITE	(0.0F / 0.0F)
#else
#define FLT_INFINITY	(1.0F / FLT_DENORMAL)
#define FLT_INDEFINITE	(0.0F * FLT_INFINITY)
#endif
#else
#define FLT_DENORMAL	0x0.000002p-126F
//			0x0.FFFFFEp-126F
#define FLT_MINIMUM	0x1.000000p-126F
#define FLT_EPSILON	0x1.000000p-23F		// = 0x0.000002p+0F
#define FLT_OMICRON	0x1.FFFFFEp-1F
#define FLT_YPSILON	0x1.000000p+23F
#define FLT_MAXIMUM	0x1.FFFFFEp+127F
#define FLT_INFINITY	0x1.000000p+128F
#define FLT_SNAN	0x1.400000p+128F	// signaling NaN
#define FLT_INDEFINITE -0x1.800000p+128F
#define FLT_QNAN	0x1.C00000p+128F	// quiet NaN
#endif // _MSC_VER

#define DBL_RADIX	FLT_RADIX
#if _MSC_VER < 1914 // MSC 19.14 alias 2017
#define DBL_DENORMAL	4.940656458412465441765687928682213724e-324	// = 2**-1074
//			2.225073858507200889024586876085859888e-308	// = 2**-1022 * (1 - 2**-53)
#define DBL_MINIMUM	2.225073858507201383090232717332404064e-308	// = 2**-1022
#define DBL_EPSILON	2.220446049250313080847263336181640625e-16	// = 2**-52	EXACT: 2**-n has n fractional decimal digits!
#define DBL_OMICRON	9.999999999999998889776975374843459576e-1	// = 1 - 2**-53
#define DBL_YPSILON	4503599627370496.0				// = 2**+52
#define DBL_MAXIMUM	1.797693134862315708145274237317043568e+308	// = 2**+1024 * (1 - 2**-53)
#if 0	// NOTE: MSC yields error C2124 for 1.0 / 0.0 and 0.0 / 0.0!
#define DBL_INFINITY	(1.0 / 0.0)
#define DBL_INDEFINITE	(0.0 / 0.0)
#else
#define DBL_INFINITY	(1.0 / DBL_DENORMAL)
#define DBL_INDEFINITE	(0.0 * DBL_INFINITY)
#endif
#else
#define DBL_DENORMAL	0x0.0000000000001p-1022
//			0x0.FFFFFFFFFFFFFp-1022
#define DBL_MINIMUM	0x1.0000000000000p-1022
#define DBL_EPSILON	0x1.0000000000000p-52	// = 0x0.0000000000001p+0
#define DBL_OMICRON	0x1.FFFFFFFFFFFFFp-1
#define DBL_YPSILON	0x1.0000000000000p+52
#define DBL_MAXIMUM	0x1.FFFFFFFFFFFFFp+1023
#define DBL_INFINITY	0x1.0000000000000p+1024
#define DBL_SNAN	0x1.4000000000000p+1024	// signaling NaN
#define DBL_INDEFINITE -0x1.8000000000000p+1024
#define DBL_QNAN	0x1.C000000000000p+1024	// quiet NaN
#endif // _MSC_VER

// https://msdn.microsoft.com/en-us/library/9cx8xs15.aspx

#if 0	// NOTE: MSC does not support long double alias binary80!
#define EXT_RADIX	FLT_RADIX
#if _MSC_VER < 1914 // MSC 19.14 alias 2017
#define EXT_DENORMAL	3.6451995318824746025284059336194e-4951L		// = 2**-16445
#define EXT_MINIMUM	3.36210314311209350626267781732175260259807934e-4932L	// = 2**-16382
#define EXT_EPSILON	1.08420217248550443400745280086994171142578125e-19L	// = 2**-63	EXACT: 2**-n has n fractional decimal digits!
#define EXT_OMICRON	9.99999999999999999945789891375724778299627360e-1L	// = 1 - 2**-64
#define EXT_YPSILON	9223372036854775808.0					// = 2**+63
#define EXT_MAXIMUM	1.18973149535723176502126385303097020516906332e+4932L	// = 2**+16384 * (1 - 2**-64)
#if 0	// NOTE: MSC yields error C2124 for 1.0L / 0.0L and 0.0L / 0.0L!
#define EXT_INFINITY	(1.0L / 0.0L)
#define EXT_INDEFINITE	(0.0L / 0.0L)
#else
#define EXT_INFINITY	(1.0L / EXT_DENORMAL)
#define EXT_INDEFINITE	(0.0L * EXT_INFINITY)
#endif
#else
#define EXT_DENORMAL	0x0.0000000000000002p-16382L
#define EXT_MINIMUM	0x1.0000000000000000p-16382L
#define EXT_EPSILON	0x1.0000000000000000p-63L	// = 0x0.0000000000000002p+0L
#define EXT_OMICRON	0x1.FFFFFFFFFFFFFFFEp-1L
#define EXT_YPSILON	0x1.0000000000000000p+63L
#define EXT_MAXIMUM	0x1.FFFFFFFFFFFFFFFEp+16383L
#define EXT_INFINITY	0x1.0000000000000000p+16384L
#define EXT_SNAN	0x1.4000000000000000p+16384L	// signaling NaN
#define EXT_INDEFINITE -0x1.8000000000000000p+16384L
#define EXT_QNAN	0x1.C000000000000000p+16384L	// quiet NaN
#endif // _MSC_VER
#endif

#if 0	// NOTE: MSC yields error C2124 for 1.0 / 0.0 and 0.0 / 0.0!
#define INFINITY	(1.0 / 0.0)
#define INDEFINITE	(0.0 / 0.0)
#elif 1
#define INFINITY	(1.0 / 0.5e-323)
#define INDEFINITE	(0.0 * INFINITY)
#else
#define INFINITY	(1.5e308 + 1.5e308)
#define INDEFINITE	(INFINITY - INFINITY)
#endif

// https://msdn.microsoft.com/en-us/library/4hwaceh6.aspx

#if _MSC_VER < 1914 // MSC 19.14 alias 2017
#define M_CATALAN	0.915965594177219015054603514932384111	// Catalan's Constant
#define M_DELTA		2.414213562373095048801688724209698079	// Silver Ratio
#define M_E		2.718281828459045235360287471352662498	// Euler's Number
#define M_SQRTE		1.648721270700128146848650787814163572
#define M_1_E		0.367879441171442321595523770161460867
#define M_1_SQRTE	0.606530659712633423603799534991180453
#define M_GAMMA		0.577215664901532860606512090082402431	// Euler-Mascheroni Constant
#define M_1_GAMMA	1.732454714600633473583025315860829681
#define M_LOG2E		1.442695040888963407359924681001892137
#define M_LOG2TEN	3.321928094887362347870319429489390176
#define M_LOG10E	0.434294481903251827651128918916605082
#define M_LOG10TWO	0.301029995663981195213738894724493027
#define M_LN2		0.693147180559945309417232121458176568
#define M_LN3		1.098612288668109691395245236922525705
#define M_LN5		1.609437912434100374600759333226187640
#define M_LN7		1.945910149055313305105352743443179730
#define M_LN10		2.302585092994045684017991454684364208
#define M_1_LN10	0.434294481903251827651128918916605082
#define M_LNPI		1.144729885849400174143427351353058712
#define M_PHI		1.618033988749894848204586834365638118	// Golden Ratio
#define M_1_PHI		0.618033988749894848204586834365638118
#define M_PI		3.141592653589793238462643383279502884	// Archimedes' Constant
#define M_PI_2		1.570796326794896619231321691639751442
#define M_PI_3		1.047197551196597746154214461093167628
#define M_PI_4		0.785398163397448309615660845819875721
#define M_PI_5		0.628318530717958647692528676655900577
#define M_PI_6		0.523598775598298873077107230546583814
#define	M_PI_8		0.392699081698724154807830422909937861
#define M_SQRTPI	1.772453850905516027298167483341145183
#define M_SQRT2PI	2.506628274631000502415765284811045253
#define M_1_PI		0.318309886183790671537767526745028724
#define M_2_PI		0.636619772367581343075535053490057448
#define	M_4_PI		1.273239544735162686151070106980114890
#define M_1_SQRTPI	0.564189583547756286948079451560772586
#define M_1_SQRT2PI	0.398942280401432677939946059934381868
#define M_2_SQRTPI	1.128379167095512573896158903121545172
#define M_PSI		1.465571231876768026656731225219939108	// Super-golden Ratio
#define M_RHO		1.324717957244746025960908854478097341	// Plastic Number
#define M_TAU		6.283185307179586476925286766559005768
#define M_SQRTTAU	2.506628274631000502415765284811045253
#define M_1_SQRT2TAU	0.398942280401432677939946059934381868
#define M_SQRT2		1.414213562373095048801688724209698079	// Pythagoras' Constant
#define M_SQRT2_2	0.707106781186547524400844362104849039
#define M_1_SQRT2	0.707106781186547524400844362104849039
#define M_SQRT3		1.732050807568877293527446341505872367	// Theodorus' Constant
#define M_1_SQRT3	0.577350269189625764509148780501957456
#define M_SQRT5		2.236067977499789696409173668731276235
#define M_1_SQRT5	0.447213595499957939281834733746255247
#define M_SQRT7		2.645751311064590590501615753639260426
#define M_1_SQRT7	0.377964473009227227214516536234180061
#define M_SQRT8		2.828427124746190097603377448419396157
#define M_1_SQRT8	0.353553390593273762200422181052424520
#define M_SQRT10	3.162277660168379331998893544432718534
#define M_1_SQRT10	0.316227766016837933199889354443271853
#else
#define M_CATALAN	0x1.D4F9713E8135D08A42B045C6FA65Ap-1	// = G
#define M_DELTA		0x2.6A09E667F3BCC908B2FB1366EA958p+0	// = delta = (delta**2 - 1) / 2
//      M_DELTA		0x1.3504F333F9DE6484597D89B3754ACp+1
#define M_E		0x1.5BF0A8B1457695355FB8AC404E7A8p+1	// = e
#define M_SQRTE		0x1.A61298E1E069BC972DFEFAB6DF340p+0	// = sqrt(e)
#define M_1_E		0x1.78B56362CEF37C6AEB7B1E0A4153Ep-2	// = 1 / e
#define M_1_SQRTE	0x1.368B2FC6F9609FE7ACEB46AA619BBp-1	// = 1 / sqrt(e)
#define M_GAMMA		0x1.2788CFC6FB618F49A37C7F0202A59p-1	// = gamma
#define M_1_GAMMA	0x1.BB8226F502BF7E854138C6D9978ACp+0	// = 1 / gamma
#define M_LOG2E		0x1.71547652B82FE1777D0FFDA0D23A8p+0	// = log2(e) = 1 / log(2)
#define M_LOG2TEN	0x1.A934F0979A3715FC9257EDFE9B5FBp+1	// = log2(10) = 1 / log10(2)
#define M_LOG10E	0x1.BCB7B1526E50E32A6AB7555F5A67Cp-2	// = log10(e) = 1 / log(10)
#define M_LOG10TWO	0x1.34413509F79FEF311F12B35816F92p-2	// = log10(2) = 1 / log2(10)
#define M_LN2		0x1.62E42FEFA39EF35793C7673007E5Fp-1	// = log(2) = 1 / log2(e)
#define M_LN3		0x1.193EA7AAD030A976A4198D55053B8p+0	// = log(3)
#define M_LN5		0x1.9C041F7ED8D336AFDF77A51607593p+0	// = log(5)
#define M_LN7		0x1.F2272AE325A57546F69496CF261BEp+0	// = log(7)
#define M_LN10		0x1.26BB1BBB5551582DD4ADAC5705A61p+1	// = log(10) = log(2) + log(5)
#define M_1_LN10	0x1.BCB7B1526E50E32A6AB7555F5A67Bp-2	// = 1 / log(10) = log10(e)
#define M_LNPI		0x1.250D048E7A1BD0BD5F956C6A843F5p+0	// = log(pi)
#define M_PHI		0x1.9E3779B97F4A7C15F39CC0605CEDDp+0	// = phi = phi**2 - 1
#define M_1_PHI		0x1.3C6EF372FE94F82BE73980C0B9DB9p-1	// = 1 / phi = phi - 1
#define M_PI		0x1.921FB54442D18469898CC51701B84p+1	// = pi
#define M_PI_2		0x1.921FB54442D18469898CC51701B84p+0	// = pi / 2
#define M_PI_3		0x1.0C152382D73658465BB32E0F567ADp+0	// = pi / 3
#define M_PI_4		0x1.921FB54442D18469898CC51701B84p-1	// = pi / 4
#define M_PI_5		0x1.41B2F769CF0E0387A13D6A78CE2D0p-1	// = pi / 5
#define M_PI_6		0x1.0C152382D73658465BB32E0F567ADp-1	// = pi / 6
#define M_PI_8		0x1.921FB54442D18469898CC51701B84p-2	// = pi / 8
#define M_SQRTPI	0x1.C5BF891B4EF6AA79C3B0520D5DB94p+0	// = sqrt(pi)
#define M_SQRT2PI	0x1.40D931FF627059657CA41FAE722CFp+1	// = sqrt(pi * 2)
#define M_1_PI		0x1.45F306DC9C882A53F84EAFA3EA69Cp-2	// = 1 / pi
#define M_2_PI		0x1.45F306DC9C882A53F84EAFA3EA69Cp-1	// = 2 / pi
#define	M_4_PI		0x1.45F306DC9C882A53F84EAFA3EA69Cp-0	// = 4 / pi
#define M_1_SQRTPI	0x1.20DD750429B6D11AE3A914FED7FD8p-1	// = 1 / sqrt(pi)
#define M_1_SQRT2PI	0x1.9884533D436508D0FCB3C500BAB8Ep-2	// = 1 / sqrt(pi * 2)
#define M_2_SQRTPI	0x1.20DD750429B6D11AE3A914FED7FD8p+0	// = 2 / sqrt(pi)
#define M_PSI		0x1.772FAD1EDE80B462113642B48A702p+2	// = psi = sqrt(psi**3 - 1) = sqrt(sqrt(psi)) + 1
#define M_RHO		0x1.5320B74ECA44ADAC178897C414613p+0	// = rho = rho**3 - 1
#define M_TAU		0x1.921FB54442D18469898CC51701B84p+2	// = tau = pi * 2
#define M_SQRTTAU	0x1.40D931FF627059657CA41FAE722CFp+1	// = sqrt(tau)
#define M_1_SQRTTAU	0x1.9884533D436508D0FCB3C500BAB8Ep-2	// = 1 / sqrt(tau)
#define M_SQRT2		0x1.6A09E667F3BCC908B2FB1366EA958p+0	// = sqrt(2)
#define M_SQRT2_2	0x1.6A09E667F3BCC908B2FB1366EA958p-1	// = sqrt(2) / 2 = 1 / sqrt(2)
#define M_1_SQRT2	0x1.6A09E667F3BCC908B2FB1366EA958p-1	// = 1 / sqrt(2) = sqrt(2) / 2
#define M_SQRT3		0x1.BB67AE8584CAA73B25742D7078B84p+0	// = sqrt(3)
#define M_1_SQRT3	0x1.279A74590331C4D218F81E4AFB258p-1	// = 1 / sqrt(3)
#define M_SQRT5		0x1.1E3779B97F4A7C15F39CC0605CEDDp+1	// = sqrt(5)
#define M_1_SQRT5	0x1.C9F25C5BFEDD93565294670094AFAp-2	// = 1 / sqrt(5)
#define M_SQRT7		0x1.52A7FA9D2F8E9B78E753F30FE1BD1p+1	// = sqrt(7)
#define M_1_SQRT7	0x1.83091E6A7F7E688A2CF23A5B4B213p-2	// = 1 / sqrt(7)
#define M_SQRT8		0x1.6A09E667F3BCC908B2FB1366EA958p+1	// = sqrt(8)
#define M_1_SQRT8	0x1.6A09E667F3BCC908B2FB1366EA958p-2	// = 1 / sqrt(8)
#define M_SQRT10	0x1.94C583ADA5B529204A2BC830CD9C0p+1	// = sqrt(10)
#define M_1_SQRT10	0x1.43D136248490EDB36E896CF3D7B00p-2	// = 1 / sqrt(10)
#endif // _MSC_VER
#endif // IEEE_754

#ifndef _countof
// https://msdn.microsoft.com/en-us/library/ms175773.aspx

#define _countof(array)			(sizeof(array) / sizeof(*array))
#endif

#ifndef offsetof
// https://msdn.microsoft.com/en-us/library/dz4y9b9a.aspx

#define offsetof(aggregate, member)	(size_t) &(((aggregate *) 0)->member)
#endif

#ifdef __cplusplus
extern	"C"
{
#endif

#if 0
void	__annotation(wchar_t *, ...);
void	__debugbreak(void);
void	__int2c(void);
void	__nop(void);

#pragma intrinsic(__annotation, __debugbreak, __int2c, __nop)
#endif

// https://msdn.microsoft.com/en-us/library/k64h826d.aspx

typedef	char	*va_list;

#ifdef _M_AMD64
// https://msdn.microsoft.com/en-us/library/0zz611kb.aspx

void	__va_start(va_list *, ...);
void	*__va_arg(va_list *, ...);
void	__va_end(va_list *);

#pragma intrinsic(__va_arg, __va_end, __va_start)

// https://msdn.microsoft.com/en-us/library/kb57fad8.aspx

#ifndef va_arg
#define va_arg(ptr, type)	((sizeof(type) > sizeof(long long) || (sizeof(type) & (sizeof(type) - 1)) != 0) ? **(type **) ((ptr += sizeof(long long)) - sizeof(long long)) : *(type *) ((ptr += sizeof(long long)) - sizeof(long long)))
#define va_copy(dst, src)	((dst) = (src))
#define va_end(ptr)		(ptr = (va_list) 0)
#define va_start(ptr, arg)	__va_start(&ptr, arg)
#endif
#elif _M_IX86
#ifndef va_arg
#define va_arg(ptr, type)	(*(type *) ((ptr += ((sizeof(type) + sizeof(int) - 1) & (0 - sizeof(int)))) - ((sizeof(type) + sizeof(int) - 1) & (0 - sizeof(int)))))
#define va_copy(dst, src)	((dst) = (src))
#define va_end(ptr)		(ptr = (va_list) 0)
#define va_start(ptr, arg)	(ptr = (va_list) &arg + ((sizeof(arg) + sizeof(int) - 1) & (0 - sizeof(int))))
#endif
#endif

typedef	enum	_stream
{
	STDIN = 0,
	STDOUT,
	STDERR
} STREAM;

BOOL	WINAPI	PrintCharA(STREAM enStream, CHAR cbChar);
BOOL	WINAPI	PrintCharW(STREAM enStream, WCHAR cwChar);

#define PrintString(STREAM, LITERAL)	PrintBuffer(STREAM, LITERAL, sizeof(LITERAL) / sizeof(*LITERAL) - 1)
#define PrintStringA(STREAM, LITERAL)	PrintBufferA(STREAM, LITERAL, sizeof(LITERAL) / sizeof("") - 1)
#define PrintStringW(STREAM, LITERAL)	PrintBufferW(STREAM, LITERAL, sizeof(LITERAL) / sizeof(L"") - 1)

BOOL	WINAPI	PrintBufferA(STREAM enStream, LPCSTR szBuffer, LONG nlBuffer);
BOOL	WINAPI	PrintBufferW(STREAM enStream, LPCWSTR szBuffer, LONG nlBuffer);

// https://msdn.microsoft.com/en-us/library/hh916382.aspx

BOOL	CDECL	PrintFormatA(STREAM enStream, [SA_FormatString(Style="printf")] LPCSTR szFormat, ...);
BOOL	CDECL	PrintFormatW(STREAM enStream, [SA_FormatString(Style="printf")] LPCWSTR szFormat, ...);

VOID	CDECL	DebugFormatA([SA_FormatString(Style="printf")] LPCSTR szFormat, ...);
VOID	CDECL	DebugFormatW([SA_FormatString(Style="printf")] LPCWSTR szFormat, ...);

#ifdef UNICODE
#define PrintChar	PrintCharW
#define PrintBuffer	PrintBufferW
#define PrintFormat	PrintFormatW
#define DebugFormat	DebugFormatW
#else
#define PrintChar	PrintCharA
#define PrintBuffer	PrintBufferA
#define PrintFormat	PrintFormatA
#define DebugFormat	DebugFormatA
#endif

// https://msdn.microsoft.com/en-us/library/hh916382.aspx

int	snprintf([SA_Pre(Access=SA_Write, Null=SA_Maybe)] [SA_Post(NullTerminated=SA_Yes)] char *buffer, size_t count,
	         [SA_FormatString(Style="printf")] char const *format, ...);
int	swprintf([SA_Pre(Access=SA_Write, Null=SA_Maybe)] [SA_Post(NullTerminated=SA_Yes)] wchar_t *buffer, size_t count,
	         [SA_FormatString(Style="printf")] wchar_t const *format, ...);
int	vsnprintf([SA_Pre(Access=SA_Write, Null=SA_Maybe)] [SA_Post(NullTerminated=SA_Yes)] char *buffer, size_t count,
	          [SA_FormatString(Style="printf")] char const *format, va_list input);
int	vswprintf([SA_Pre(Access=SA_Write, Null=SA_Maybe)] [SA_Post(NullTerminated=SA_Yes)] wchar_t *buffer, size_t count,
	          [SA_FormatString(Style="printf")] wchar_t const *format, va_list input);

int	sscanf([SA_Pre(Access=SA_Read, Deref=1, Valid=SA_Yes)] [SA_Pre(Null=SA_No, NullTerminated=SA_Yes)] char const *input,
	       [SA_FormatString(Style="scanf")] char const *format, ...);
int	swscanf([SA_Pre(Access=SA_Read, Deref=1, Valid=SA_Yes)] [SA_Pre(Null=SA_No, NullTerminated=SA_Yes)] wchar_t const *input,
	        [SA_FormatString(Style="scanf")] wchar_t const *format, ...);
int	vsscanf([SA_Pre(Access=SA_Read, Deref=1, Valid=SA_Yes)] [SA_Pre(Null=SA_No, NullTerminated=SA_Yes)] char const *input,
	        [SA_FormatString(Style="scanf")] char const *format, va_list output);
int	vswscanf([SA_Pre(Access=SA_Read, Deref=1, Valid=SA_Yes)] [SA_Pre(Null=SA_No, NullTerminated=SA_Yes)] wchar_t const *input,
	         [SA_FormatString(Style="scanf")] wchar_t const *format, va_list output);

char	*itoa(signed int value, char *string, int radix);
wchar_t	*itow(signed int value, wchar_t *string, int radix);
char	*ltoa(signed long value, char *string, int radix);
wchar_t	*ltow(signed long value, wchar_t *string, int radix);
char	*lltoa(signed long long value, char *string, int radix);
wchar_t	*lltow(signed long long value, wchar_t *string, int radix);
char	*ultoa(unsigned long value, char *string, int radix);
wchar_t	*ultow(unsigned long value, wchar_t *string, int radix);
char	*ulltoa(unsigned long long value, char *string, int radix);
wchar_t	*ulltow(unsigned long long value, wchar_t *string, int radix);

int	ltostr(signed long number, char string[12]);
int	lltostr(signed long long number, char string[21]);
int	ultostr(unsigned long number, char string[11]);
int	ulltostr(unsigned long long number, char string[21]);

#ifdef IEEE_754
int	dtostr(double number, char string[25]);
int	ilogb(double argument);
int	signbit(double argument);

long		_ftol(double argument);
long long	_ftol2(double argument);
long		lrint(double argument);
long long	llrint(double argument);
#ifdef _M_AMD64
long		lround(double argument);
long long	llround(double argument);
#endif

double	acos(double argument);
double	acosh(double argument);
double	acot(double argument);
double	acot2(double numerator, double denominator);
double	acoth(double argument);
double	acsc(double argument);
double	asec(double argument);
double	asin(double argument);
double	asinh(double argument);
double	atan(double argument);
double	atan2(double numerator, double denominator);
double	atanh(double argument);
double	cbrt(double argument);
double	ceil(double argument);
double	copysign(double to, double from);
double	cos(double argument);
double	cosh(double argument);
double	cot(double argument);
double	coth(double argument);
double	exp(double exponent);
double	exp2(double exponent);
double	exp2n(int exponent);
double	exp10(double exponent);
#ifdef _M_IX86
double	expm1(double exponent);
#endif
double	fabs(double argument);
double	fdim(double x, double y);
double	floor(double argument);
double	fmax(double x, double y);
double	fmin(double x, double y);
double	fmod(double dividend, double divisor);
double	frand(unsigned long long integer);
double	frexp(double argument, int *exponent);
double	hypot(double a, double b);
double	ldexp(double argument, int exponent);
double	ldexp10(double argument, int exponent);
double	log(double argument);
#ifdef _M_IX86
double	log1p(double argument);
#endif
double	log2(double argument);
double	log10(double argument);
double	logb(double argument);
double	modf(double argument, double *integer);
double	nextafter(double from, double to);
double	pow(double base, double exponent);
double	pown(double base, int exponent);
double	remainder(double numerator, double denominator);
double	remquo(double numerator, double denominator, int *quotient);
double	rint(double argument);
double	round(double argument);
double	scalbn(double argument, int exponent);
double	sin(double argument);
double	sinh(double argument);
double	sqrt(double radicand);
double	tan(double argument);
double	tanh(double argument);
double	trunc(double argument);
double	_scalb(double argument, int exponent);

double	strtod(char const *string, char **stop);
double	wcstod(wchar_t const *string, wchar_t **stop);

// https://msdn.microsoft.com/en-us/library/5704bbxw.aspx
// https://msdn.microsoft.com/en-us/library/tzkfha43.aspx

#pragma intrinsic(acos, asin, atan, atan2, cos, cosh, exp, fabs, floor, fmod, log, log10, pow, sin, sinh, sqrt, tan, tanh)

__inline
double	acsch(double x)
{
	return asinh(1.0 / x);
}

__inline
double	asech(double x)
{
	return acosh(1.0 / x);
}

#ifndef _M_IX86	// BUG: ceil() not available as intrinsic for _M_IX86
#pragma intrinsic(ceil)
#elif 0
#pragma function(ceil)

// https://msdn.microsoft.com/en-us/library/atdhw2dx.aspx

__declspec(deprecated("Don't use; included only for documentation!"))
__inline
double	ceil(double argument)
{
	double	integer;

	return modf(argument, &integer) > 0.0 ? integer + 1.0 : integer;
}
#elif 0
#pragma function(ceil)

// https://msdn.microsoft.com/en-us/library/atdhw2dx.aspx

__declspec(deprecated("Don't use; included only for documentation!"))
__inline
double	ceil(double argument)
{
	double	tmp = trunc(argument);

	return argument > tmp ? tmp + 1.0 : tmp;
}
#else
// https://msdn.microsoft.com/en-us/library/atdhw2dx.aspx

__inline
double	ceil(double argument)
{
	return -floor(-argument);
}
#endif // _M_IX86

#ifdef _M_IX86
// https://msdn.microsoft.com/en-us/library/0yafk1hc.aspx

__inline
double	copysign(double to, double from)
{
	__asm	// returns result in FPU st(0)
	{
		mov	eax, dword ptr from+4
		shld	dword ptr to+4, eax, 1
		ror	dword ptr to+4, 1
		fld	to
	}
}
#elif 0
// https://msdn.microsoft.com/en-us/library/0yafk1hc.aspx

__declspec(deprecated("Don't use; included only for documentation!"))
__inline
double	copysign(double to, double from)
{
#if 0
	return signbit(from) ? -fabs(to) : fabs(to);
#else
	return signbit(from) == signbit(to) ? to : -to;
#endif
}
#endif // _M_IX86

__inline
double	csc(double x)
{
	return 1.0 / sin(x);
}

__inline
double	csch(double x)
{
	return 1.0 / sinh(x);
}

#if 0
// https://msdn.microsoft.com/en-us/library/mt720714.aspx

__inline
double	fdim(double x, double y)
{
	return x < y ? 0.0 : x - y;
}
#endif

#if 0
#pragma function(floor)

// https://msdn.microsoft.com/en-us/library/x39715t6.aspx

__declspec(deprecated("Don't use; included only for documentation!"))
__inline
double	floor(double argument)
{
	double	integer;

	return modf(argument, &integer) < 0.0 ? integer - 1.0 : integer;
}
#elif 0
#pragma function(floor)

// https://msdn.microsoft.com/en-us/library/x39715t6.aspx

__declspec(deprecated("Don't use; included only for documentation!"))
__inline
double	floor(double argument)
{
	double	tmp = trunc(argument);

	return argument < tmp ? tmp - 1.0 : tmp;
}
#elif 0
#pragma function(floor)

// https://msdn.microsoft.com/en-us/library/x39715t6.aspx

__declspec(deprecated("Don't use; included only for documentation!"))
__inline
double	floor(double argument)
{
	return -ceil(-argument);
}
#endif

#if 0
// https://msdn.microsoft.com/en-us/library/mt720717.aspx

__declspec(deprecated("Don't use; included only for documentation!"))
__inline
double	fmax(double x, double y)
{
#if 0
	if (x != x)
		return y;

	if (x == y)
		return signbit(x) ? y : x;

	return x < y ? y : x;
#else
	return (x > y) || (y != y) ? x : y;
#endif
}
#endif

#if 0
// https://msdn.microsoft.com/en-us/library/mt720716.aspx

__declspec(deprecated("Don't use; included only for documentation!"))
__inline
double	fmin(double x, double y)
{
#if 0
	if (x != x)
		return y;

	if (x == y)
		return signbit(x) ? x : y;

	return x > y ? y : x;
#else
	return (x < y) || (y != y) ? x : y;
#endif
}
#endif

#if 0
// https://msdn.microsoft.com/en-us/library/20dckbeh.aspx

__declspec(deprecated("Don't use; included only for documentation!"))
__inline
double	fmod(double dividend, double divisor)
{
	return dividend - divisor * trunc(dividend / divisor);
}
#endif

#if 0
__declspec(deprecated("Don't use; included only for documentation!"))
__inline
double	frexp10(double argument, int *exponent)
{
	int	exponent2;
	double	fraction2 = frexp(argument, &exponent2);

	// convert exponent to base 2 into exponent to base 10,
	//  leaving a remainder
	double	exponent10;
#if 0
	double	remainder10 = modf(log10(2.0) * exponent2, &exponent10);
#else
	double	remainder10 = modf(M_LOG10TWO * exponent2, &exponent10);
#endif
	double	exponent10floor = floor(exponent10);

	*exponent = (int) exponent10floor;
#if 0
	return fraction2 * exp(log(10.0) * (exponent10 - exponent10floor + remainder10));
#else
	return fraction2 * exp(M_LN10 * (exponent10 - exponent10floor + remainder10));
#endif
}
#else
__inline
double	frexp10(double argument, int *exponent)
{
	*exponent = argument == 0.0 ? 0 : 1 + (int) floor(log10(fabs(argument)));
#if 0
	return argument / exp10(*exponent);
#else
	return argument * exp10(-*exponent);
#endif
}
#endif

#if 0
// https://msdn.microsoft.com/en-us/library/a9yb3dbt.aspx

__declspec(deprecated("Don't use; included only for documentation!"))
double	hypot(double x, double y)
{
	double	z;
	int	exponent;

	y = fabs(y);

	if (y == INFINITY)
		return y;

	x = fabs(x);

	if (x == INFINITY)
		return x;

	if (x < y)
		z = y, y = x, x = z;

	x = frexp(x, &exponent);
	y = ldexp(y, -exponent);

	return ldexp(sqrt(x * x + y * y), exponent);
}
#endif

__inline
int	isdenormal(double argument)
{
	return (argument != 0.0) && (fabs(argument) < DBL_MINIMUM);
}

// https://msdn.microsoft.com/en-us/library/sb8es7a8.aspx

__inline
int	isfinite(double argument)
{
#if 0	// CAVEAT: fails with '/fp:fast' compiler switch!

	return argument - argument == 0.0;
#else
	return fabs(argument) <= DBL_MAXIMUM;
#endif
}

__inline
int	isinfinite(double argument)
{
#if 0	// CAVEAT: fails with '/fp:fast' compiler switch!

	return argument - argument != 0.0;
#else
	return fabs(argument) > DBL_MAXIMUM;
#endif
}

// https://msdn.microsoft.com/en-us/library/tzthab44.aspx

__inline
int	isnan(double argument)
{
	// CAVEAT: fails with '/fp:fast' compiler switch!

	return argument != argument;
}

__inline
int	isnormal(double argument)
{
	return (fabs(argument) >= DBL_MINIMUM)
	    && (fabs(argument) <= DBL_MAXIMUM);
}

#ifdef _M_IX86
// https://msdn.microsoft.com/en-us/library/zx52ds7f.aspx

__inline
double	ldexp(double argument, int exponent)
{
	__asm	// returns result in FPU st(0)
	{
		fild	exponent
		fld	argument
		fscale
		fstp	st(1)
	}
}
#endif // _M_IX86

#ifdef _M_IX86
// https://msdn.microsoft.com/en-us/library/mt720723.aspx

__inline
long	lrint(double argument)
{
	__asm	// returns result in eax
	{
		fld	argument
		fistp	dword ptr argument
		mov	eax, dword ptr argument
	}
}
#endif // _M_IX86

#ifdef _M_IX86
// https://msdn.microsoft.com/en-us/library/mt720723.aspx

__inline
long long	llrint(double argument)
{
	__asm	// returns result in edx:eax
	{
		fld	argument
		fistp	argument
		mov	eax, dword ptr argument
		mov	edx, dword ptr argument+4
	}
}
#endif // _M_IX86

#if 0
// https://msdn.microsoft.com/en-us/library/t63833dz.aspx

__inline
double	log(double argument)
{
	__asm	// returns result in FPU st(0)
	{
		fldln2
		fld	argument
		fyl2x
	}
}
#endif

#ifdef _M_IX86
// https://msdn.microsoft.com/en-us/library/mt720721.aspx

__inline
double	log2(double argument)
{
	__asm	// returns result in FPU st(0)
	{
		fld1
		fld	argument
		fyl2x
	}
}
#elif 0
// https://msdn.microsoft.com/en-us/library/mt720721.aspx

__declspec(deprecated("Don't use; included only for documentation!"))
__inline
double	log2(double argument)
{
	double	mantissa, logarithm, power, tmp;
	int	exponent;

	if (argument != argument)
		return argument;

	if (argument < 0.0)
		return INDEFINITE;

	if (argument == 0.0)
		return -INFINITY;

	if (argument == 1.0)
		return 0.0;

	if (argument == INFINITY)
		return INFINITY;

	mantissa = frexp(argument, &exponent);

	if (mantissa == 0.5)
		return exponent - 1;
#if 1
	for (logarithm = 0.0, power = 0.5; power > 0.0; power *= 0.5)
	{
		mantissa *= mantissa;

		if (mantissa >= 0.5)
			continue;

		mantissa += mantissa;

		tmp = logarithm;
		logarithm -= power;

		if (logarithm == tmp)
			break;
	}
#else
	for (exponent -= 1, mantissa += mantissa, logarithm = 0.0,
	     power = 0.5; power > 0.0; power *= 0.5)
	{
		if (mantissa == 1.0)
			break;

		mantissa *= mantissa;

		if (mantissa < 2.0)
			continue;

		mantissa *= 0.5;

		tmp = logarithm;
		logarithm += power;

		if (logarithm == tmp)
			break;
	}
#endif
	return exponent + logarithm;
}
#endif // _M_IX86

#if 0
// https://msdn.microsoft.com/en-us/library/t63833dz.aspx

__inline
double	log10(double argument)
{
	__asm	// returns result in FPU st(0)
	{
		fldlg2
		fld	argument
		fyl2x
	}
}
#endif

#ifdef _M_IX86
// https://msdn.microsoft.com/en-us/library/e4x82d9s.aspx

__inline
double	logb(double argument)
{
	__asm	// returns result in FPU st(0)
	{
		fld	argument
		fxtract
		fstp	st(0)
	}
}
#endif // _M_IX86

#if 0
// https://msdn.microsoft.com/en-us/library/dn465170.aspx

__declspec(deprecated("Don't use; included only for documentation!"))
__inline
double	remainder(double dividend, double divisor)
{
	return dividend - divisor * rint(dividend / divisor);
}
#endif

#ifdef _M_IX86
// https://msdn.microsoft.com/en-us/library/dn465165.aspx

__inline
double	rint(double argument)
{
	__asm	// returns result in FPU st(0)
	{
		fld	argument
		frndint
	}
}
#elif 0
// https://msdn.microsoft.com/en-us/library/dn465165.aspx

__declspec(deprecated("Don't use; included only for documentation!"))
__inline
double	rint(double argument)
{
	if ((0.0 < argument) && (argument < DBL_YPSILON))
	{
		argument += DBL_YPSILON;
		argument -= DBL_YPSILON;
	}
	else if ((0.0 > argument) && (argument > -DBL_YPSILON))
	{
		argument -= DBL_YPSILON;
		argument += DBL_YPSILON;

		if (argument == 0.0)
			argument = -0.0;
	}

	return argument;
}
#endif // _M_IX86

#if 0
// https://msdn.microsoft.com/en-us/library/dn353646.aspx

__declspec(deprecated("Don't use; included only for documentation!"))
__inline
double	round(double argument)
{
	double	integer1, integer2;

	modf(modf(argument, &integer1) * 2.0, &integer2);

	return integer1 + integer2;
}
#elif 0
// https://msdn.microsoft.com/en-us/library/dn353646.aspx

__declspec(deprecated("Don't use; included only for documentation!"))
__inline
double	round(double argument)
{
	double	tmp = argument;

	if (argument > 0.0)
	{
		argument = floor(argument);

		if (tmp - argument >= 0.5)
			argument += 1.0;
	}
	else if (argument < 0.0)
	{
		argument = ceil(argument);

		if (tmp - argument <= -0.5)
			argument -= 1.0;
	}

	return argument;
}
#endif

__inline
double	sec(double x)
{
	return 1.0 / cos(x);
}

__inline
double	sech(double x)
{
	return 1.0 / cosh(x);
}

#ifdef _M_IX86
__inline
int	signbit(double argument)
{
	__asm	// returns result in eax
	{
		mov	eax, dword ptr argument+4
		shr	eax, 31
	}
}
#elif 0
__declspec(deprecated("Don't use; included only for documentation!"))
__inline
int	signbit(double argument)
{
	return (argument < 0.0) || ((argument == 0.0) && (1.0 / argument < 0.0));
}
#endif // _M_IX86

#if 0
// https://msdn.microsoft.com/en-us/library/mt720727.aspx

__declspec(deprecated("Don't use; included only for documentation!"))
__inline
double	trunc(double argument)
{
	double	integer;

	modf(argument, &integer);

	return integer;
}
#elif 0
// https://msdn.microsoft.com/en-us/library/mt720727.aspx

__declspec(deprecated("Don't use; included only for documentation!"))
__inline
double	trunc(double argument)
{
	if (argument < 0.0)
		return ceil(argument);

	if (argument > 0.0)
		return floor(argument);

	return argument;
}
#endif
#endif // IEEE_754

void	swab(void const *source, void const *destination, int count);

int	memcmp(void const *destination, void const *source, size_t count);
void	*memchr(void const *destination, int character, size_t count);
void	*memcpy(void *destination, void const *source, size_t count);
void	*memmem(void *haystack, size_t count, void *needle, size_t length);
void	*memmove(void *destination, void const *source, size_t count);
void	*memset(void *destination, int character, size_t count);
void	memswap(void *destination, void *source, size_t count);

#ifdef INTRINSIC
// https://msdn.microsoft.com/en-us/library/5704bbxw.aspx
// https://msdn.microsoft.com/en-us/library/tzkfha43.aspx
// https://msdn.microsoft.com/en-us/library/sctyh01s.aspx

#pragma intrinsic(memcmp, memcpy, memset)
#elif _M_IX86
#pragma function(memcmp)

// https://msdn.microsoft.com/en-us/library/zyaebf12.aspx

__declspec(naked)
int	memcmp(void const *source, void const *destination, size_t count)
{
	__asm
	{
		mov	eax, [esp+4]	; eax = source
		mov	edx, [esp+8]	; edx = destination
	;;	cmp	edx, eax
	;;	je	short equal
		xor	ecx, ecx	; CF = 0, ZF = 1 (required when count is 0)
		mov	ecx, [esp+12]	; ecx = count
	;;	jecxz	short equal
		xchg	esi, eax	; esi = source
		xchg	edi, edx	; edi = destination
		repe	cmpsb
		mov	edi, edx
		mov	esi, eax
;;	equal:
		seta	al
		movzx	eax, al		; eax = (source > destination)
		sbb	eax, 0		; eax = (source > destination) - (source < destination)
					;     = {1, 0, -1}
		ret
	}
}

// https://msdn.microsoft.com/en-us/library/d7zdhf37.aspx

__declspec(naked)
void	*memchr(void const *destination, int character, size_t count)
{
	__asm
	{
#if 1
		mov	edx, edi
		mov	edi, [esp+4]	; edi = destination
		mov	eax, [esp+8]	; eax = character
		mov	ecx, [esp+12]	; ecx = count
		repne	scasb
		dec	edi
		neg	ecx		; CF = (ecx <> 0)
					;    = (*edi = al)
		sbb	eax, eax	; eax = (ecx = 0) ? 0 : -1
		and	eax, edi	; eax = (ecx = 0) ? 0 : pointer
		mov	edi, edx
		ret
#else
		mov	ecx, [esp+12]	; ecx = count
		test	ecx, ecx
		jz	short zero
		mov	edx, [esp+8]	; edx = character
		mov	eax, [esp+4]	; eax = destination
	next:
		cmp	dl, [eax]
		je	short done
		inc	eax
		dec	ecx
		jnz	short next
	zero:
		mov	eax, ecx
	done:
		ret
#endif
	}
}

__declspec(naked)
void	*memrchr(void const *destination, int character, size_t count)
{
	__asm
	{
		mov	edx, edi
		mov	ecx, [esp+12]	; ecx = count
		mov	eax, [esp+8]	; eax = cb
		mov	edi, [esp+4]	; edi = destination
		lea	edi, [edi+ecx-1]
		std
		repne	scasb
		cld
		inc	edi
		neg	ecx		; CF = (ecx <> 0)
					;    = (*edi = al)
		sbb	eax, eax	; eax = (ecx = 0) ? 0 : -1
		and	eax, edi	; eax = (ecx = 0) ? 0 : pointer
		mov	edi, edx
		ret
	}
}

__declspec(naked)
void	*memmem(void const *haystack, size_t count, void const *needle, size_t length)
{
	__asm
	{
		xor	eax, eax	; eax = address of needle in haystack = 0
		mov	ecx, [esp+8]	; ecx = length of haystack
		test	ecx, ecx
		jz	short empty	; length of haystack = 0?

		mov	edx, [esp+16]	; edx = length of needle
		cmp	edx, ecx
		ja	short empty	; length of needle > length of haystack?

		mov	eax, [esp+4]	; eax = address of haystack
		test	edx, edx
		jz	short empty	; length of needle = 0?

		push	ebx
		push	edi
		mov	edi, eax	; edi = address of haystack
		push	esi
	search:
		mov	esi, [esp+24]	; esi = address of needle
		mov	al, [esi]	; al = first character of needle
					; edi = current address in haystack
		repne	scasb		; edi = next address in haystack,
					; ecx = current length of haystack
		jne	short break	; (first character of) needle not in haystack?

		dec	ecx		; ecx = next length of haystack

		mov	al, [esi+edx-1]	; al = last character of needle
		cmp	al, [edi+edx-2]
		jne	short continue	; last character of needle not in haystack?
	compare:
		mov	eax, edi	; eax = next address in haystack
		mov	ebx, ecx	; ebx = next length of haystack
#if 0
		dec	edi		; edi = address of matching character
					;     = current address in haystack
					; esi = address of needle
		mov	ecx, edx	; ecx = length of needle
#else
					; edi = next address in haystack
		inc	esi		; esi = address of needle + 1
		mov	ecx, edx
		dec	ecx		; ecx = length of needle - 1
	;;	jz	short match	; needle in haystack?
#endif
		repe	cmpsb
		je	short match	; needle in haystack?

		mov	edi, eax	; edi = current address in haystack
		mov	ecx, ebx	; ecx = current length of haystack
	continue:
		cmp	ecx, edx
		jae	short search	; length of haystack >= length of needle?
	break:
		xor	eax, eax
		pop	esi
		pop	edi
		pop	ebx
	empty:
		ret
	match:
		dec	eax		; eax = address of needle in haystack
		pop	esi
		pop	edi
		pop	ebx
		ret
	}
}

#pragma function(memcpy, memset)
#pragma intrinsic(__movsb, __stosb)

// https://msdn.microsoft.com/en-us/library/dswaw1wk.aspx

__inline
void	*memcpy(void *destination, void const *source, size_t count)
{
	// https://msdn.microsoft.com/en-us/library/hhta9830.aspx

	__movsb((unsigned char *) destination, (unsigned char const *) source, count);

	return destination;
}

// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

__inline
void	*memset(void *destination, int character, size_t count)
{
	// https://msdn.microsoft.com/en-us/library/ywwcatsa.aspx

	__stosb((unsigned char *) destination, (unsigned char) character, count);

	return destination;
}
#elif _M_AMD64
// https://msdn.microsoft.com/en-us/library/d7zdhf37.aspx

void	*memchr(void const *destination, int character, size_t count)
{
	while (count != 0)
	{
		if (*(char *) destination == character)
			return (void *) destination;

		destination = (char *) destination + 1;
		count--;
	}

	return NULL;
}

void	*memrchr(void const *destination, int character, size_t count)
{
	while (count != 0)
		if (((char *) destination)[--count] == character)
			return (char *) destination + count;
	return NULL;
}

#pragma function(memcpy, memset)
#pragma intrinsic(__movsb, __stosb)

// https://msdn.microsoft.com/en-us/library/dswaw1wk.aspx

__inline
void	*memcpy(void *destination, void const *source, size_t count)
{
	// https://msdn.microsoft.com/en-us/library/hhta9830.aspx

	__movsb((unsigned char *) destination, (unsigned char const *) source, count);

	return destination;
}

// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

__inline
void	*memset(void *destination, int character, size_t count)
{
	// https://msdn.microsoft.com/en-us/library/ywwcatsa.aspx

	__stosb((unsigned char *) destination, (unsigned char) character, count);

	return destination;
}

void	*memmem(void const *haystack, size_t m, void const *needle, size_t n)
{
	char const	*o;
	char const	*p = (char const *) haystack;
	char const	*q = (char const *) needle;

	if ((m == 0) || (n > m))
		return NULL;

	if (n == 0)
		return (void *) haystack;

	if (n == 1)
		return memchr(p, *q, m);

	m -= n - 1;		// maximum number of characters to scan

	while (o = p, p = (char const *) memchr(p, *q, m), p != NULL)
	{
		if ((p[n - 1] == q[n - 1])
		 && ((n - 2 == 0) || (memcmp(p + 1, q + 1, n - 2) == 0)))
			return (void *) p;

		m -= ++p - o;	// remaining number of characters to scan
		if (m == 0)
			break;
	}

	return NULL;
}
#endif // INTRINSIC

char	*strcat(char *destination, char const *source);
char	*strchr(char const *string, char cb);
int	strcmp(char const *left, char const *right);
int	strcoll(char const *left, char const *right);
char	*strcpy(char *destination, char const *source);
size_t	strcspn(char const *string, char const *delimiter);
size_t	strlen(char const *string);
char	*strncat(char *destination, char const *source, size_t count);
int	strncmp(char const *left, char const *right, size_t count);
size_t	strnlen(char const *string, size_t count);
char	*strnset(char *string, char character, size_t count);
char	*strpbrk(char const *string, char const *delimiter);
char	*strrchr(char const *string, char cb);
char	*strrev(char *string);
char	*strset(char *string, char cb);
char	*_strset(char *string, char cb);
size_t	strspn(char const *string, char const *delimiter);
char	*strstr(char const *string, char const *substring);
char	*strtok(char *string, char const *delimiter);
char	*strtok_r(char *string, char const *delimiter, char **next);
size_t	strxfrm(char *destination, char const *source, size_t count);

#ifdef INTRINSIC
// https://msdn.microsoft.com/en-us/library/5704bbxw.aspx
// https://msdn.microsoft.com/en-us/library/tzkfha43.aspx

#pragma intrinsic(strcat, strcmp, strcpy, strlen, strset, _strset)
#endif

// https://msdn.microsoft.com/en-us/library/b34ccac3.aspx

char	*strchr(char const *string, char cb)
{
	do
		if (*string == cb)
			return (char *) string;
	while (*string++ != '\0');

	return NULL;
}

// https://msdn.microsoft.com/en-us/library/tbyd7s1y.aspx

char	*strncat(char *destination, char const *source, size_t count)
{
	size_t	index = 0, offset = strlen(destination);

	while ((index < count) && (source[index] != '\0'))
		destination[index + offset] = source[index], index++;

	destination[index + offset] = '\0';

	return destination;
}

// https://msdn.microsoft.com/en-us/library/eywx8zcx.aspx

int	strncmp(char const *left, char const *right, size_t count)
{
	size_t	index = 0;

	if ((count != 0) && (left != right))
		do
			if ((left[index] == '\0')
		//	 || (right[index] == '\0')
			 || (left[index] - right[index]))
#if 1
				return left[index] - right[index];
#else
				return (left[index] > right[index])
				     - (left[index] < right[index]);
#endif
		while (++index < count);

	return 0;
}

// https://msdn.microsoft.com/en-us/library/xdsywd25.aspx

char	*strncpy(char *destination, char const *source, size_t count)
{
	size_t	index = 0;

	while ((index < count) && (source[index] != '\0'))
		destination[index] = source[index], index++;

	while (index < count)
		destination[index++] = '\0';

	return destination;
}

__inline
size_t	strnlen(char const *string, size_t count)
{
	char	*pointer = (char *) memchr(string, '\0', count);

	return pointer == NULL ? count : pointer - string;
}

// https://msdn.microsoft.com/en-us/library/ftw0heb9.aspx

char	*strrchr(char const *string, int character)
{
	char	*pointer = NULL;

	do
		if (*string == (char) character)
			pointer = (char *) string;
	while (*string++ != '\0');

	return pointer;
}

// https://msdn.microsoft.com/en-us/library/z9da80kz.aspx

__inline
char	*strstr(char const *string, char const *substring)
{
	if (*substring == '\0')
		return (char *) string;

	if (*string == '\0')
		return NULL;

	return (char *) memmem(string, strlen(string), substring, strlen(substring));
}

// https://msdn.microsoft.com/en-us/library/2c8d19sb.aspx

__inline
char	*strtok(char *string, char const *delimiter)
{
	static	char	*next = NULL;

	return strtok_r(string, delimiter, &next);
}

__inline
size_t	strxfrm(char *destination, char const *source, size_t count)
{
	size_t	length = strlen(source) + 1;

	if (count > length)
		count = length;

	memcpy(destination, source, count);

	return length;
}

signed long	strtol(char const *string, char **stop, int base);
unsigned long	strtoul(char const *string, char **stop, int base);

signed long long	strtoll(char const *string, char const **stop, int base);
unsigned long long	strtoull(char const *string, char const **stop, int base);

int	wmemcmp(wchar_t const *destination, wchar_t const *source, size_t count);
wchar_t	*wmemchr(wchar_t const *destination, wchar_t character, size_t count);
wchar_t	*wmemcpy(wchar_t *destination, wchar_t const *source, size_t count);
wchar_t	*wmemmem(wchar_t *haystack, size_t count, wchar_t *needle, size_t length);
wchar_t	*wmemmove(wchar_t *destination, wchar_t const *source, size_t count);
wchar_t	*wmemset(wchar_t *destination, wchar_t character, size_t count);
void	wmemswap(wchar_t *destination, wchar_t *source, size_t count);

#ifdef _M_IX86
// https://msdn.microsoft.com/en-us/library/zyaebf12.aspx

__declspec(naked)
int	wmemcmp(wchar_t const *source, wchar_t const *destination, size_t count)
{
	__asm
	{
		mov	eax, [esp+4]	; eax = source
		mov	edx, [esp+8]	; edx = destination
	;;	cmp	edx, eax
	;;	je	short equal
		xor	ecx, ecx	; CF = 0, ZF = 1 (required when count is 0)
		mov	ecx, [esp+12]	; ecx = count
	;;	jecxz	short equal
		xchg	esi, eax	; esi = source
		xchg	edi, edx	; edi = destination
		repe	cmpsw
		mov	edi, edx
		mov	esi, eax
;;	equal:
		seta	al
		movzx	eax, al		; eax = (source > destination)
		sbb	eax, 0		; eax = (source > destination) - (source < destination)
					;     = {1, 0, -1}
		ret

	}
}

// https://msdn.microsoft.com/en-us/library/d7zdhf37.aspx

__declspec(naked)
wchar_t	*wmemchr(wchar_t const *destination, wchar_t character, size_t count)
{
	__asm
	{
#if 1
		mov	edx, edi
		mov	edi, [esp+4]	; edi = destination
		mov	eax, [esp+8]	; eax = character
		mov	ecx, [esp+12]	; ecx = count
		repne	scasw
		dec	edi
		dec	edi
		neg	ecx		; CF = (ecx <> 0)
					;    = (*edi = ax)
		sbb	eax, eax	; eax = (ecx = 0) ? 0 : -1
		and	eax, edi	; eax = (ecx = 0) ? 0 : pointer
		mov	edi, edx
		ret
#else
		mov	ecx, [esp+12]	; ecx = count
		test	ecx, ecx
		jz	short zero
		mov	edx, [esp+8]	; edx = character
		mov	eax, [esp+4]	; eax = destination
	next:
		cmp	dx, [eax]
		je	short done
		add	eax, 2
		dec	ecx
		jnz	short next
	zero:
		mov	eax, ecx
	done:
		ret
#endif
	}
}

__declspec(naked)
wchar_t	*wmemrchr(wchar_t const *destination, wchar_t character, size_t count)
{
	__asm
	{
		mov	edx, edi
		mov	ecx, [esp+12]	; ecx = count
		mov	eax, [esp+8]	; eax = character
		mov	edi, [esp+4]	; edi = destination
		lea	edi, [edi+ecx*2-2]
		std
		repne	scasw
		cld
		inc	edi
		inc	edi
		neg	ecx		; CF = (ecx <> 0)
					;    = (*edi = ax)
		sbb	eax, eax	; eax = (ecx = 0) ? 0 : -1
		and	eax, edi	; eax = (ecx = 0) ? 0 : pointer
		mov	edi, edx
		ret
	}
}

__declspec(naked)
wchar_t	*wmemmem(wchar_t const *haystack, size_t count, wchar_t const *needle, size_t length)
{
	__asm
	{
		xor	eax, eax	; eax = address of needle in haystack = 0
		mov	ecx, [esp+8]	; ecx = length of haystack
		test	ecx, ecx
		jz	short empty	; length of haystack = 0?

		mov	edx, [esp+16]	; edx = length of needle
		cmp	edx, ecx
		ja	short empty	; length of needle > length of haystack?

		mov	eax, [esp+4]	; eax = address of haystack
		test	edx, edx
		jz	short empty	; length of needle = 0?

		push	ebx
		push	edi
		mov	edi, eax	; edi = address of haystack
		push	esi
	search:
		mov	esi, [esp+24]	; esi = address of needle
		mov	ax, [esi]	; ax = first wide character of needle
					; edi = current address in haystack
		repne	scasw		; edi = next address in haystack,
					; ecx = current length of haystack
		jne	short break	; (first wide character of) needle not in haystack?
	compare:
		mov	eax, edi	; eax = next address in haystack
		dec	ecx		; ecx = next length of haystack
		mov	ebx, ecx
		mov	cx, [esi+edx*2-2]
					; cx = last wide character of needle
		cmp	cx, [edi+edx*2-4]
		jne	short continue	; last wide character of needle not in haystack?
#if 0
		dec	edi
		dec	edi		; edi = address of matching wide character
					;     = current address in haystack
					; esi = address of needle
		mov	ecx, edx	; ecx = length of needle
#else
					; edi = next address in haystack
		inc	esi
		inc	esi		; esi = address of needle + 2
		mov	ecx, edx
		dec	ecx		; ecx = length of needle - 1
	;;	jz	short match	; needle in haystack?
#endif
		repe	cmpsw
		je	short match	; needle in haystack?
	continue:
		mov	edi, eax	; edi = current address in haystack
		mov	ecx, ebx	; ecx = current length of haystack
		cmp	ebx, edx
		jae	short search	; length of haystack >= length of needle?
	break:
		xor	eax, eax
		pop	esi
		pop	edi
		pop	ebx
	empty:
		ret
	match:
		dec	eax
		dec	eax		; eax = address of needle in haystack
		pop	esi
		pop	edi
		pop	ebx
		ret
	}
}

#pragma intrinsic(__movsw, __stosw)

// https://msdn.microsoft.com/en-us/library/dswaw1wk.aspx

__inline
wchar_t	*wmemcpy(wchar_t *destination, wchar_t const *source, size_t count)
{
	// https://msdn.microsoft.com/en-us/library/adf7db3x.aspx

	__movsw(destination, source, count);

	return destination;
}

// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

__inline
wchar_t	*wmemset(wchar_t *destination, wchar_t character, size_t count)
{
// https://msdn.microsoft.com/en-us/library/k99wy47b.aspx

	__stosw(destination, character, count);

	return destination;
}
#elif _M_AMD64
// https://msdn.microsoft.com/en-us/library/d7zdhf37.aspx

wchar_t	*wmemchr(wchar_t const *destination, wchar_t character, size_t count)
{
	while (count != 0)
	{
		if (*destination == character)
			return (wchar_t *) destination;

		destination++;
		count--;
	}

	return NULL;
}

wchar_t	*wmemrchr(wchar_t const *destination, wchar_t character, size_t count)
{
	while (count != 0)
		if (destination[--count] == character)
			return (wchar_t *) destination + count;
	return NULL;
}

#pragma intrinsic(__movsw, __stosw)

// https://msdn.microsoft.com/en-us/library/dswaw1wk.aspx

__inline
wchar_t	*wmemcpy(wchar_t *destination, wchar_t const *source, size_t count)
{
	// https://msdn.microsoft.com/en-us/library/adf7db3x.aspx

	__movsw(destination, source, count);

	return destination;
}

// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

__inline
wchar_t	*wmemset(wchar_t *destination, wchar_t character, size_t count)
{
// https://msdn.microsoft.com/en-us/library/k99wy47b.aspx

	__stosw(destination, character, count);

	return destination;
}

wchar_t	*wmemmem(wchar_t const *haystack, size_t m, wchar_t const *needle, size_t n)
{
	wchar_t	const	*o;
	wchar_t	const	*p = (wchar_t const *) haystack;
	wchar_t	const	*q = (wchar_t const *) needle;

	if ((m == 0) || (n > m))
		return NULL;

	if (n == 0)
		return (wchar_t *) haystack;

	if (n == 1)
		return wmemchr(p, *q, m);

	m -= n - 1;		// maximum number of characters to scan

	while (o = p, p = wmemchr(p, *q, m), p != NULL)
	{
		if ((p[n - 1] == q[n - 1])
		 && ((n - 2 == 0) || (wmemcmp(p + 1, q + 1, n - 2) == 0)))
			return (wchar_t *) p;

		m -= ++p - o;	// remaining number of characters to scan

		if (m == 0)
			break;
	}

	return NULL;
}
#endif // _M_IX86, _M_AMD64

wchar_t	*wcscat(wchar_t *destination, wchar_t const *source);
wchar_t	*wcschr(wchar_t const *string, wchar_t character);
int	wcscmp(wchar_t const *left, wchar_t const *right);
int	wcscoll(wchar_t const *left, wchar_t const *right);
wchar_t	*wcscpy(wchar_t *destination, wchar_t const *source);
size_t	wcscspn(wchar_t const *string, wchar_t const *delimiter);
size_t	wcslen(wchar_t const *string);
wchar_t	*wcsncat(wchar_t *destination, wchar_t const *source, size_t count);
int	wcsncmp(wchar_t const *left, wchar_t const *right, size_t count);
size_t	wcsnlen(wchar_t const *string, size_t count);
wchar_t	*wcsnset(wchar_t *string, wchar_t character, size_t count);
wchar_t	*wcspbrk(wchar_t const *string, wchar_t const *delimiter);
wchar_t	*wcsrchr(wchar_t const *string, wchar_t character);
wchar_t	*wcsrev(wchar_t *string);
wchar_t	*wcsset(wchar_t *string, wchar_t character);
wchar_t	*_wcsset(wchar_t *string, wchar_t character);
size_t	wcsspn(wchar_t const *string, wchar_t const *delimiter);
wchar_t	*wcsstr(wchar_t const *string, wchar_t const *substring);
wchar_t	*wcstok(wchar_t *string, wchar_t const *delimiter);
wchar_t	*wcstok_r(wchar_t *string, wchar_t const *delimiter, wchar_t **next);
size_t	wcsxfrm(wchar_t *destination, wchar_t const *source, size_t count);

#ifdef INTRINSIC
// https://msdn.microsoft.com/en-us/library/5704bbxw.aspx

#pragma intrinsic(wcscat, wcscmp, wcscpy, wcslen, _wcsset)
#endif

// https://msdn.microsoft.com/en-us/library/b34ccac3.aspx

wchar_t	*wcschr(wchar_t const *string, wchar_t character)
{
	do
		if (*string == character)
			return (wchar_t *) string;
	while (*string++ != L'\0');

	return NULL;
}

// https://msdn.microsoft.com/en-us/library/xe8sk0x7.aspx

size_t	wcscspn(wchar_t const *string, wchar_t const *delimiter)
{
	// number of leading characters in string which
	//  are NOT equal to any character in delimiter

	size_t	index;

	if (delimiter[0] == L'\0')
		return wcslen(string);

	for (index = 0; string[index] != L'\0'; index++)
		if (wcschr(delimiter, string[index]) != NULL)
			break;
	return index;
}

// https://msdn.microsoft.com/en-us/library/tbyd7s1y.aspx

wchar_t	*wcsncat(wchar_t *destination, wchar_t const *source, size_t count)
{
	size_t	index = 0, offset = wcslen(destination);

	while ((index < count) && (source[index] != L'\0'))
		destination[index + offset] = source[index], index++;

	destination[index + offset] = L'\0';

	return destination;
}

// https://msdn.microsoft.com/en-us/library/eywx8zcx.aspx

int	wcsncmp(wchar_t const *left, wchar_t const *right, size_t count)
{
	size_t	index = 0;

	if ((count != 0) && (left != right))
		do
			if ((left[index] == L'\0')
		//	 || (right[index] == L'\0')
			 || (left[index] - right[index]))
#if 1
				return left[index] - right[index];
#else
				return (left[index] > right[index])
				     - (left[index] < right[index]);
#endif
		while (++index < count);

	return 0;
}

// https://msdn.microsoft.com/en-us/library/xdsywd25.aspx

wchar_t	*wcsncpy(wchar_t *destination, wchar_t const *source, size_t count)
{
	size_t	index = 0;

	while ((index < count) && (source[index] != L'\0'))
		destination[index] = source[index], index++;

	while (index < count)
		destination[index++] = L'\0';

	return destination;
}

__inline
size_t	wcsnlen(wchar_t const *string, size_t count)
{
	wchar_t	*pointer = wmemchr(string, L'\0', count);

	return pointer != NULL ? pointer - string : count;
}

// https://msdn.microsoft.com/en-us/library/217yyhy9.aspx

wchar_t	*wcspbrk(wchar_t const *string, wchar_t const *delimiter)
{
	// pointer to first character in string which
	//  is equal to any character in delimiter
#if 0
	string += wcscspn(string, delimiter);

	return *string == L'\0' ? NULL : (wchar_t *) string;
#else
	if (*delimiter == L'\0')
		return NULL;

	for (; *string != L'\0'; string++)
		if (wcschr(delimiter, *string) != NULL)
			return (wchar_t *) string;
	return NULL;
#endif
}

// https://msdn.microsoft.com/en-us/library/ftw0heb9.aspx

wchar_t	*wcsrchr(wchar_t const *string, wchar_t character)
{
	wchar_t	*pointer = NULL;

	do
		if (*string == character)
			pointer = (wchar_t *) string;
	while (*string++ != L'\0');

	return pointer;
}

// https://msdn.microsoft.com/en-us/library/kthzzay9.aspx

size_t	wcsspn(wchar_t const *string, wchar_t const *delimiter)
{
	// number of leading characters in string which
	//  are equal to any character in delimiter

	size_t	index;

	if (delimiter[0] == L'\0')
		return 0;

	for (index = 0; string[index] != L'\0'; index++)
		if (wcschr(delimiter, string[index]) == NULL)
			break;
	return index;
}

// https://msdn.microsoft.com/en-us/library/z9da80kz.aspx

__inline
wchar_t	*wcsstr(wchar_t const *string, wchar_t const *substring)
{
	if (*substring == L'\0')
		return (wchar_t *) string;

	if (*string == L'\0')
		return NULL;

	return wmemmem(string, wcslen(string), substring, wcslen(substring));
}

// https://msdn.microsoft.com/en-us/library/2c8d19sb.aspx

__inline
wchar_t	*wcstok(wchar_t *string, wchar_t const *delimiter)
{
	static	wchar_t	*next = NULL;

	return wcstok_r(string, delimiter, &next);
}

// https://msdn.microsoft.com/en-us/library/ftsafwz3.aspx

wchar_t	*wcstok_r(wchar_t *string, wchar_t const *delimiter, wchar_t **next)
{
#if 0
	wchar_t	const	*token;

	if (string == NULL)
		string = *next;

	if ((string == NULL) || (*string == L'\0'))
		return *next = NULL;

	if (*delimiter == L'\0')
		return *next = NULL, string;

				// skip leading delimiters
	for (; *string != L'\0'; string++)
		if (wcschr(delimiter, *string) == NULL)
			break;
				// skip token, search delimiter
	for (token = string; *string != L'\0'; string++)
	{
		if (wcschr(delimiter, *string) == NULL)
			continue;
				// write terminating NUL after token
		*string++ = L'\0';
		break;
	}

	return (*next = string) != token ? (wchar_t *) token : NULL;
#else
	if (string == NULL)
		string = *next;

	if ((string == NULL) || (*string == L'\0'))
		return *next = NULL;

	string += wcsspn(string, delimiter);

	if (*string == L'\0')
		return *next = NULL;

	*next = string + wcscspn(string, delimiter);

	if (**next != L'\0')
		*(*next)++ = L'\0';
	else
		*next = NULL;

	return string;
#endif
}

__inline
size_t	wcsxfrm(wchar_t *destination, wchar_t const *source, size_t count)
{
	size_t	length = wcslen(source) + 1;

	if (count > length)
		count = length;

	wmemcpy(destination, source, count);

	return length;
}

signed long	wcstol(wchar_t const *string, wchar_t **stop, int base);
unsigned long	wcstoul(wchar_t const *string, wchar_t **stop, int base);

signed long long	wcstoll(wchar_t const *string, wchar_t const **stop, int base);
unsigned long long	wcstoull(wchar_t const *string, wchar_t const **stop, int base);

int		lmemcmp(unsigned long const *destination, unsigned long const *source, size_t count);
unsigned long	*lmemchr(unsigned long const *destination, unsigned long cl, size_t count);
unsigned long	*lmemcpy(unsigned long *destination, unsigned long const *source, size_t count);
unsigned long	*lmemmove(unsigned long *destination, unsigned long const *source, size_t count);
unsigned long	*lmemset(unsigned long *destination, unsigned long cl, size_t count);
void		lmemswap(unsigned long *destination, unsigned long *source, size_t count);

#ifdef _M_IX86
// https://msdn.microsoft.com/en-us/library/zyaebf12.aspx

__declspec(naked)
int	lmemcmp(unsigned long const *source, unsigned long const *destination, size_t count)
{
	__asm
	{
		mov	eax, [esp+4]	; eax = source
		mov	edx, [esp+8]	; edx = destination
	;;	cmp	edx, eax
	;;	je	short equal
		xor	ecx, ecx	; CF = 0, ZF = 1 (required when count is 0)
		mov	ecx, [esp+12]	; ecx = count
	;;	jecxz	short equal
		xchg	esi, eax	; esi = source
		xchg	edi, edx	; edi = destination
		repe	cmpsd
		mov	edi, edx
		mov	esi, eax
;;	equal:
		seta	al
		movzx	eax, al		; eax = (source > destination)
		sbb	eax, 0		; eax = (source > destination) - (source < destination)
					;     = {1, 0, -1}
		ret
	}
}

__declspec(naked)
unsigned long	*lmemchr(unsigned long const *destination, unsigned long ul, size_t count)
{
	__asm
	{
#if 1
		mov	edx, edi
		mov	ecx, [esp+12]		; ecx = count
		mov	eax, [esp+8]		; eax = ul
		mov	edi, [esp+4]		; edi = destination
		repne	scasd
		sub	edi, 4
		neg	ecx			; CF = (ecx <> 0)
						;    = (*edi = eax)
		sbb	eax, eax		; eax = (ecx = 0) ? 0 : -1
		and	eax, edi		; eax = (ecx = 0) ? 0 : pointer
		mov	edi, edx
		ret
#else
		mov	ecx, [esp+12]		; ecx = count
		test	ecx, ecx
		jz	short zero
		mov	edx, [esp+8]		; edx = ul
		mov	eax, [esp+4]		; eax = destination
	next:
		cmp	edx, [eax]
		je	short done
		add	eax, 4
		dec	ecx
		jnz	short next
	zero:
		mov	eax, ecx
	done:
		ret
#endif
	}
}

__declspec(naked)
unsigned long	*lmemrchr(unsigned long const *destination, unsigned long ul, size_t count)
{
	__asm
	{
		mov	edx, edi
		mov	ecx, [esp+12]		; ecx = count
		mov	eax, [esp+8]		; eax = ul
		mov	edi, [esp+4]		; edi = destination
		lea	edi, [edi+ecx*4-4]
		std
		repne	scasd
		cld
		add	edi, 4
		neg	ecx			; CF = (ecx <> 0)
						;    = (*edi = eax)
		sbb	eax, eax		; eax = (ecx = 0) ? 0 : -1
		and	eax, edi		; eax = (ecx = 0) ? 0 : pointer
		mov	edi, edx
		ret
	}
}

#pragma intrinsic(__movsd, __stosd)

__inline
unsigned long	*lmemcpy(unsigned long *destination, unsigned long const *source, size_t count)
{
	// https://msdn.microsoft.com/en-us/library/9d196b9h.aspx

	__movsd(destination, source, count);

	return destination;
}

__inline
unsigned long	*lmemset(unsigned long *destination, unsigned long ul, size_t count)
{
	// https://msdn.microsoft.com/en-us/library/d72sd2b6.aspx

	__stosd(destination, ul, count);

	return destination;
}
#elif _M_AMD64
unsigned long	*lmemchr(unsigned long const *destination, unsigned long ul, size_t count)
{
	while (count != 0)
	{
		if (*destination == ul)
			return (unsigned long *) destination;

		destination++;
		count--;
	}

	return NULL;
}

unsigned long	*lmemrchr(unsigned long const *destination, unsigned long ul, size_t count)
{
	while (count != 0)
		if (destination[--count] == ul)
			return (unsigned long *) destination + count;
	return NULL;
}

#pragma intrinsic(__movsd, __stosd)

__inline
unsigned long	*lmemcpy(unsigned long *destination, unsigned long const *source, size_t count)
{
	// https://msdn.microsoft.com/en-us/library/9d196b9h.aspx

	__movsd(destination, source, count);

	return destination;
}

__inline
unsigned long	*lmemset(unsigned long *destination, unsigned long ul, size_t count)
{
	// https://msdn.microsoft.com/en-us/library/d72sd2b6.aspx

	__stosd(destination, ul, count);

	return destination;
}
#endif // _M_IX86, _M_AMD64

#ifdef _WIN64
int			llmemcmp(unsigned long long const *destination, unsigned long long const *source, size_t count);
unsigned long long	*llmemchr(unsigned long long const *source, unsigned long long ull, size_t count);
unsigned long long	*llmemcpy(unsigned long long *destination, unsigned long long const *source, size_t count);
unsigned long long	*llmemmove(unsigned long long *destination, unsigned long long const *source, size_t count);
unsigned long long	*llmemset(unsigned long long *destination, unsigned long long ull, size_t count);
void			llmemswap(unsigned long long *destination, unsigned long long *source, size_t count);

#ifdef _M_AMD64
unsigned long long	*llmemchr(unsigned long long const *destination, unsigned long long ull, size_t count)
{
	while (count != 0)
	{
		if (*destination == ull)
			return (unsigned long long *) destination;

		destination++;
		count--;
	}

	return NULL;
}

unsigned long long	*llmemrchr(unsigned long long const *destination, unsigned long long ull, size_t count)
{
	while (count != 0)
		if (destination[--count] == ull)
			return (unsigned long long *) destination + count;
	return NULL;
}

#pragma intrinsic(__movsq, __stosq)

__inline
unsigned long long	*llmemcpy(unsigned long long *destination, unsigned long long const *source, size_t count)
{
	// https://msdn.microsoft.com/en-us/library/4tt6fykh.aspx

	__movsq(destination, source, count);

	return destination;
}

__inline
unsigned long long	*llmemset(unsigned long long *destination, unsigned long long ull, size_t count)
{
	// https://msdn.microsoft.com/en-us/library/3dsh30sa.aspx

	__stosq(destination, ull, count);

	return destination;
}
#endif // _M_AMD64
#endif // _WIN64

// companion routine for __emulu() intrinsic

__inline		// widening carry-less multiplication
unsigned long long	__eclmul(unsigned long p, unsigned long q)
{
#ifdef _M_IX86
	__asm	// returns carry-less product in edx:eax
	{
		xor	eax, eax
		xor	edx, edx	; edx:eax = carry-less product" = 0
		mov	edi, p		; edi = multiplicand
		mov	esi, q		; esi = multiplier

		bsr	ebx, esi	; ebx = index of most significant '1' bit
		jz	short done	; multiplier = 0?

		lea	ecx, [eax+31]	; ecx = index of most significant bit = 31
	next:
		sub	ecx, ebx	; ecx = length of '0' bits run
		shld	edx, eax, cl
		shl	eax, cl		; edx:eax = carry-less product"
					;         * 2**length of '0' bits run
		xor	eax, edi	;         ^ multiplicand
					;         = carry-less product'
		btr	esi, ebx	; esi = multiplier
					;     ^ 2**index of most significant '1' bit
					;     = multiplier'
		mov	ecx, ebx
		bsr	ebx, esi	; ebx = index of most significant '1' bit
		jnz	short next	; multiplier' <> 0?

		shld	edx, eax, cl
		shl	eax, cl		; edx:eax = carry-less product'
					;         * 2**length of last '0' bits run
					;         = carry-less product
	done:
	}
#elif 1
	unsigned long long	r = 0;
	unsigned long		s = 31, t;

	// https://msdn.microsoft.com/en-us/library/fbxyd7zd.aspx

	while (_BitScanReverse(&t, q))
	{
		// https://msdn.microsoft.com/en-us/library/8dt0hdsy.aspx

		r = __ll_lshift(r, s - t);
		(unsigned long) r ^= p;

		// https://msdn.microsoft.com/en-us/library/hd0hzyf8.aspx

		_bittestandreset(&q, s = t);
	}

	// https://msdn.microsoft.com/en-us/library/8dt0hdsy.aspx

	r = __ll_lshift(r, s);

	return r;
#elif 1
	unsigned long long	r = 0;
	unsigned long		s = 1UL << 31;

	do
	{
		r <<= 1;
		if (q & s)
			(unsigned long) r ^= p;
	} while (s >>= 1);

	return r;
#elif 1
	unsigned long long	r = 0;

	while (p)
	{	// https://msdn.microsoft.com/en-us/library/d2s81xt0.aspx

		r ^= __emulu(q, p & (0 - p));
#if 0
		p ^= p & (0 - p);
#else
		p &= p - 1;
#endif
	}

	return r;
#else
	// https://msdn.microsoft.com/en-us/library/d2s81xt0.aspx

	unsigned long long	r0 = __emulu(p & 0x11111111, q & 0x11111111)
				   ^ __emulu(p & 0x22222222, q & 0x88888888)
				   ^ __emulu(p & 0x44444444, q & 0x44444444)
				   ^ __emulu(p & 0x88888888, q & 0x22222222);
	unsigned long long	r1 = __emulu(p & 0x11111111, q & 0x22222222)
				   ^ __emulu(p & 0x22222222, q & 0x11111111)
				   ^ __emulu(p & 0x44444444, q & 0x88888888)
				   ^ __emulu(p & 0x88888888, q & 0x44444444);
	unsigned long long	r2 = __emulu(p & 0x11111111, q & 0x44444444)
				   ^ __emulu(p & 0x22222222, q & 0x22222222)
				   ^ __emulu(p & 0x44444444, q & 0x11111111)
				   ^ __emulu(p & 0x88888888, q & 0x88888888);
	unsigned long long	r3 = __emulu(p & 0x11111111, q & 0x88888888)
				   ^ __emulu(p & 0x22222222, q & 0x44444444)
				   ^ __emulu(p & 0x44444444, q & 0x22222222)
				   ^ __emulu(p & 0x88888888, q & 0x11111111);
	return r0 & 0x1111111111111111
	     | r1 & 0x2222222222222222
	     | r2 & 0x4444444444444444
	     | r3 & 0x8888888888888888;
#endif
}

#ifdef _M_IX86
#ifndef _DIV_T_DEFINED
// https://msdn.microsoft.com/en-us/library/984x0h58.aspx

typedef	struct	_div_t
{
	int	quot;
	int	rem;
} div_t;

__inline
div_t	div(int dividend, int divisor)
{
	__asm	// returns quotient in eax and remainder in edx
	{
		mov	eax, dividend
		cdq
		idiv	divisor
	}
}

typedef	struct	_ldiv_t
{
	long	quot;
	long	rem;
} ldiv_t;

__inline
ldiv_t	ldiv(long dividend, long divisor)
{
	__asm	// returns quotient in eax and remainder in edx
	{
		mov	eax, dividend
		cdq
		idiv	divisor
	}
}
#endif // _DIV_T_DEFINED

#undef Int32x32To64	// BUG: MSC generates call to _allmul() instead of inline code!

#ifdef INTRINSIC
// https://msdn.microsoft.com/en-us/library/d2s81xt0.aspx

long long	__emul(long multiplicand, long multiplier);

#pragma intrinsic(__emul)
#define Int32x32To64	__emul
#else // INTRINSIC
// https://msdn.microsoft.com/en-us/library/aa383703.aspx

__declspec(deprecated("Use __emul() intrinsic!"))
__inline
long long	Int32x32To64(long multiplicand,
		             long multiplier)
{
	__asm	// returns product in edx:eax
	{
		mov	eax, multiplicand
		imul	multiplier
	}
}
#endif // INTRINSIC

#undef UInt32x32To64	// BUG: MSC generates call to _allmul() instead of inline code!

#ifdef INTRINSIC
// https://msdn.microsoft.com/en-us/library/d2s81xt0.aspx

unsigned long long	__emulu(unsigned long multiplicand, unsigned long multiplier);

#pragma intrinsic(__emulu)
#define UInt32x32To64	__emulu
#else // INTRINSIC
// https://msdn.microsoft.com/en-us/library/aa383741.aspx

__declspec(deprecated("Use __emulu() intrinsic!"))
__inline
unsigned long long	UInt32x32To64(unsigned long multiplicand,
			              unsigned long multiplier)
{
	__asm	// returns product in edx:eax
	{
		mov	eax, multiplicand
		mul	multiplier
	}
}
#endif // INTRINSIC

#ifdef _WINNT_
#pragma deprecated(Int64ShllMod32)
#pragma deprecated(Int64ShraMod32)
#pragma deprecated(Int64ShrlMod32)
#else
#ifdef INTRINSIC
// https://msdn.microsoft.com/en-us/library/8dt0hdsy.aspx

unsigned long long	__ll_lshift(unsigned long long value, unsigned long count);

#pragma intrinsic(__ll_lshift)
#define Int64ShllMod32	__ll_lshift
#else // INTRINSIC
// https://msdn.microsoft.com/en-us/library/aa383704.aspx

__declspec(deprecated("Use __ll_lshift() intrinsic!"))
__inline
unsigned long long	Int64ShllMod32(unsigned long long value,
			               unsigned long      count)
{
	__asm	// returns result in edx:eax
	{
		mov	eax, dword ptr value
		mov	edx, dword ptr value+4
		mov	ecx, count
	;;	jecxz	short done
		shld	edx, eax, cl
		shl	eax, cl
;;	done:
	}
}
#endif // INTRINSIC

#ifdef INTRINSIC
// https://msdn.microsoft.com/en-us/library/6h9fz5f5.aspx

long long	__ll_rshift(long long value, unsigned long count);

#pragma intrinsic(__ll_rshift)
#define Int64ShraMod32	__ll_rshift
#else // INTRINSIC
// https://msdn.microsoft.com/en-us/library/aa383705.aspx

__declspec(deprecated("Use __ll_rshift() intrinsic!"))
__inline
long long	Int64ShraMod32(long long     value,
		               unsigned long count)
{
	__asm	// returns result in edx:eax
	{
		mov	eax, dword ptr value
		mov	edx, dword ptr value+4
		mov	ecx, count
	;;	jecxz	short done
		shrd	eax, edx, cl
		sar	edx, cl
;;	done:
	}
}
#endif // INTRINSIC

#ifdef INTRINSIC
// https://msdn.microsoft.com/en-us/library/wds0z27f.aspx

unsigned long long	__ull_rshift(unsigned long long value, unsigned long count);

#pragma intrinsic(__ull_rshift)
#define Int64ShrlMod32	__ull_rshift
#else // INTRINSIC
// https://msdn.microsoft.com/en-us/library/aa383706.aspx

__declspec(deprecated("Use __ull_rshift() intrinsic!"))
__inline
unsigned long long	Int64ShrlMod32(unsigned long long value,
			               unsigned long      count)
{
	__asm	// returns result in edx:eax
	{
		mov	eax, dword ptr value
		mov	edx, dword ptr value+4
		mov	ecx, count
	;;	jecxz	short done
		shrd	eax, edx, cl
		shr	edx, cl
;;	done:
	}
}
#endif // INTRINSIC
#endif // _WINNT_

// companion routine for __ull_rshift() intrinsic

__inline
unsigned long long	__ull_divrem(unsigned long long dividend,
			             unsigned long      divisor,
			             unsigned long      *remainder)
{
	__asm	// returns quotient in edx:eax
	{
		mov	ecx, divisor	; ecx = divisor
#if 0
		xor	edx, edx
		mov	eax, dword ptr dividend+4
		cmp	eax, edx
		je	short common	; high dword of dividend = 0?
#else
		xor	eax, eax
		mov	edx, dword ptr dividend+4
		cmp	edx, ecx
		jb	short common	; high dword of dividend < divisor?
					; quotient < 2**32?
		xchg	eax, edx
#endif
		div	ecx		; eax = high dword of quotient,
	common:				; edx = high dword of dividend'
		push	eax
		mov	eax, dword ptr dividend
					; edx:eax = dividend'
		div	ecx		; eax = low dword of quotient,
					; edx = remainder
		mov	ecx, remainder	; ecx = address of remainder
		mov	[ecx], edx
		pop	edx		; edx:eax = quotient
	}
}

// companion routine for __ll_rshift() intrinsic

__inline
signed long long	__ll_divrem(signed long long dividend,
			            signed long      divisor,
			            signed long      *remainder)
{
	__asm	// returns quotient in edx:eax
	{
		mov	ecx, divisor	; ecx = divisor
		mov	eax, dword ptr dividend
					; eax = (low dword of) dividend
		cdq
		cmp	edx, dword ptr dividend+4
		je	short trivial	; -2**32 <= dividend < 2**32?

		mov	eax, dword ptr dividend+4
					; eax = high dword of dividend
		cdq			; edx:eax = high dword of dividend
					;         = dividend'
		idiv	ecx		; eax = high dword of quotient',
					; edx = high dword of dividend"
		push	eax
		mov	eax, dword ptr dividend
					; edx:eax = dividend"
		idiv	ecx		; eax = low dword of quotient",
					; edx = remainder
		mov	ecx, remainder	; ecx = address of remainder
		mov	[ecx], edx
		cdq			; edx:eax = quotient"
		pop	ecx
		add	edx, ecx	; edx:eax = quotient
		jmp	short done
	trivial:
		idiv	ecx		; eax = (low dword of) quotient,
					; edx = remainder
		mov	ecx, remainder	; ecx = address of remainder
		mov	[ecx], edx
		cdq			; edx:eax = quotient
	done:
	}
}

#define __emuldiv	Int32x32To64Div32

#if _MSC_VER >= 1920 // MSC 19.20 alias 2019
__declspec(deprecated("Use __emul() and _div64() intrinsics!"))
#endif
__inline
long	Int32x32To64Div32(long multiplicand,
	                  long multiplier,
	                  long divisor)
{
	__asm	// returns quotient in eax
	{
		mov	eax, multiplicand
		imul	multiplier
		idiv	divisor
	}
}

#define __emulmod	Int32x32To64Rem32

#if _MSC_VER >= 1920 // MSC 19.20 alias 2019
__declspec(deprecated("Use __emul() and _div64() intrinsics!"))
#endif
__inline
long	Int32x32To64Rem32(long multiplicand,
	                  long multiplier,
	                  long divisor)
{
	__asm	// returns remainder in eax
	{
		mov	eax, multiplicand
		imul	multiplier
		idiv	divisor
		mov	eax, edx
	}
}

#define __ediv	Int64Div32

#if _MSC_VER >= 1920 // MSC 19.20 alias 2019
__declspec(deprecated("Use _div64() intrinsic!"))
#endif
__inline
long	Int64Div32(long long dividend,
	           long      divisor)
{
	__asm	// returns quotient in eax
	{
		mov	eax, dword ptr dividend
		mov	edx, dword ptr dividend+4
		idiv	divisor
	}
}

#define __emod	Int64Rem32

#if _MSC_VER >= 1920 // MSC 19.20 alias 2019
__declspec(deprecated("Use _div64() intrinsic!"))
#endif
__inline
long	Int64Rem32(long long dividend,
	           long      divisor)
{
	__asm	// returns remainder in eax
	{
		mov	eax, dword ptr dividend
		mov	edx, dword ptr dividend+4
		idiv	divisor
		mov	eax, edx
	}
}

#define __emuldivu	UInt32x32To64Div32

#if _MSC_VER >= 1920 // MSC 19.20 alias 2019
__declspec(deprecated("Use __emulu() and _udiv64() intrinsics!"))
#endif
__inline
unsigned long	UInt32x32To64Div32(unsigned long multiplicand,
		                   unsigned long multiplier,
		                   unsigned long divisor)
{
	__asm	// returns quotient in eax
	{
		mov	eax, multiplicand
		mul	multiplier
		div	divisor
	}
}

#define __emulmodu	UInt32x32To64Rem32

#if _MSC_VER >= 1920 // MSC 19.20 alias 2019
__declspec(deprecated("Use __emulu() and _udiv64() intrinsics!"))
#endif
__inline
unsigned long	UInt32x32To64Rem32(unsigned long multiplicand,
		                   unsigned long multiplier,
		                   unsigned long divisor)
{
	__asm	// returns remainder in eax
	{
		mov	eax, multiplicand
		mul	multiplier
		div	divisor
		mov	eax, edx
	}
}

#define __edivu	UInt64Div32

#if _MSC_VER >= 1920 // MSC 19.20 alias 2019
__declspec(deprecated("Use _udiv64() intrinsic!"))
#endif
__inline
unsigned long	UInt64Div32(unsigned long long dividend,
		            unsigned long      divisor)
{
	__asm	// returns quotient in eax
	{
		mov	eax, dword ptr dividend
		mov	edx, dword ptr dividend+4
		div	divisor
	}
}

#define __emodu	UInt64Rem32

#if _MSC_VER >= 1920 // MSC 19.20 alias 2019
__declspec(deprecated("Use _udiv64() intrinsic!"))
#endif
__inline
unsigned long	UInt64Rem32(unsigned long long dividend,
		            unsigned long      divisor)
{
	__asm	// returns remainder in eax
	{
		mov	eax, dword ptr dividend
		mov	edx, dword ptr dividend+4
		div	divisor
		mov	eax, edx
	}
}

#ifdef INTRINSIC
// https://msdn.microsoft.com/en-us/library/a3140177.aspx

unsigned long long	_byteswap_uint64(unsigned long long ull);

#pragma intrinsic(_byteswap_uint64)
#define htonll	_byteswap_uint64
#define ntohll	_byteswap_uint64
#else // INTRINSIC
__declspec(deprecated("Use _byteswap_uint64() intrinsic!"))
__inline
unsigned long long	htonll(unsigned long long ull)
{
	__asm
	{
#if _MSC_VER < 1900 // MSC 19.0 alias 2015
		mov	edx, dword ptr ull
		mov	eax, dword ptr ull+4
		bswap	edx
		bswap	eax
#else
		movbe	edx, dword ptr ull
		movbe	eax, dword ptr ull+4
#endif
	}
}

__declspec(deprecated("Use _byteswap_uint64() intrinsic!"))
__inline
unsigned long long	ntohll(unsigned long long ull)
{
	__asm
	{
#if _MSC_VER < 1900 // MSC 19.0 alias 2015
		mov	edx, dword ptr ull
		mov	eax, dword ptr ull+4
		bswap	edx
		bswap	eax
#else
		movbe	edx, dword ptr ull
		movbe	eax, dword ptr ull+4
#endif
	}
}
#endif // INTRINSIC

#ifndef _WINSOCKAPI_
#ifndef _WINSOCK2API_
#ifdef INTRINSIC
// https://msdn.microsoft.com/en-us/library/a3140177.aspx

unsigned short	_byteswap_ushort(unsigned short us);
unsigned long	_byteswap_ulong(unsigned long ul);

#pragma intrinsic(_byteswap_ushort, _byteswap_ulong)
#define htons	_byteswap_ushort
#define htonl	_byteswap_ulong
#define ntohs	_byteswap_ushort
#define ntohl	_byteswap_ulong
#else // INTRINSIC
__declspec(deprecated("Use _byteswap_ulong() intrinsic!"))
__inline
unsigned long	htonl(unsigned long ul)
{
	__asm
	{
#if _MSC_VER < 1900 // MSC 19.0 alias 2015
		mov	eax, ul
		bswap	eax
#else
		movbe	eax, ul
#endif
	}
}

__declspec(deprecated("Use _byteswap_ushort() intrinsic!"))
__inline
unsigned short	htons(unsigned short us)
{
	__asm
	{
#if _MSC_VER < 1900 // MSC 19.0 alias 2015
		mov	ax, us
		ror	ax, 8
#else
		movbe	ax, us
#endif
	}
}

__declspec(deprecated("Use _byteswap_ulong() intrinsic!"))
__inline
unsigned long	ntohl(unsigned long ul)
{
	__asm
	{
#if _MSC_VER < 1900 // MSC 19.0 alias 2015
		mov	eax, ul
		bswap	eax
#else
		movbe	eax, ul
#endif
	}
}

__declspec(deprecated("Use _byteswap_ushort() intrinsic!"))
__inline
unsigned short	ntohs(unsigned short us)
{
	__asm
	{
#if _MSC_VER < 1900 // MSC 19.0 alias 2015
		mov	ax, us
		ror	ax, 8
#else
		movbe	ax, us
#endif
	}
}
#endif // INTRINSIC
#endif // _WINSOCK2API_
#endif // _WINSOCKAPI_
#else // _M_IX86
// https://msdn.microsoft.com/en-us/library/a3140177.aspx

unsigned short		_byteswap_ushort(unsigned short us);
unsigned long		_byteswap_ulong(unsigned long ul);
unsigned long long	_byteswap_uint64(unsigned long long ull);

#pragma intrinsic(_byteswap_ushort, _byteswap_ulong, _byteswap_uint64)
#define htons	_byteswap_ushort
#define htonl	_byteswap_ulong
#define htonll	_byteswap_uint64
#define ntohs	_byteswap_ushort
#define ntohl	_byteswap_ulong
#define ntohll	_byteswap_uint64
#endif // _M_IX86

#if 0	// NOTE: prevent error with '/Oi' or '/O2' compiler switches
// https://msdn.microsoft.com/en-us/library/kw1ee768.aspx

__inline
int	abs(int x)
{
#ifndef _M_IX86
	return x < 0 ? -x : x;
#else
	__asm	// returns absolute value in eax
	{
		mov	eax, x
#if 0
		mov	edx, eax
		neg	eax
		cmovl	eax, edx
#elif 0
		cdq			; edx = (x < 0) ? -1 : 0
		xor	eax, edx	; eax = (x < 0) ? ~x : x
		sub	eax, edx	; eax = (x < 0) ? ~x + 1 : x
					;     = (x < 0) ? -x : x
					;     = |x|
#else
		cdq			; edx = (x < 0) ? -1 : 0
		add	eax, edx	; eax = (x < 0) ? x - 1 : x
		xor	eax, edx	; eax = (x < 0) ? -x : x
					;     = |x|
#endif
	}
#endif // _M_IX86
}

__inline
long	labs(long x)
{
	return x < 0 ? -x : x;
}

__inline
long long	llabs(long long x)
{
#ifndef _M_IX86
	return x < 0 ? -x : x;
#else
	__asm	// returns absolute value in edx:eax
	{
		mov	eax, dword ptr x+4
		mov	ecx, dword ptr x; eax:ecx = x
		cdq			; edx = (x < 0) ? -1 : 0
		add	ecx, edx
		adc	eax, edx	; eax:ecx = (x < 0) ? x - 1 : x
		xor	ecx, edx
		xor	edx, eax	; edx:ecx = (x < 0) ? -x : x
					;         = |x|
		mov	eax, ecx	; edx:eax = |x|
	}
#endif // _M_IX86
}
#else
#define llabs	_abs64
#endif

__inline
unsigned int	umax(unsigned int x, unsigned int y)
{
#ifndef _M_IX86
	return x > y ? x : y;
#else
	__asm	// returns maximum value in eax
	{
		mov	eax, x
		mov	ecx, y
#if 1
		cmp	eax, ecx
		cmovb	eax, ecx
#else
		sub	eax, ecx	; eax = x - y
		sbb	edx, edx	; edx = (x < y) ? ~0 : 0
		not	edx		; edx = (x < y) ? 0 : ~0
		and	eax, edx	; eax = (x < y) ? 0 : x - y
		add	eax, ecx	; eax = (x < y) ? y : x
					;     = max(x, y)
#endif
	}
#endif // _M_IX86
}

#ifndef max
__inline
int	max(int x, int y)
{
#ifndef _M_IX86
	return x > y ? x : y;
#else
	__asm	// returns maximum value in eax
	{
		mov	eax, x
		mov	ecx, y
#if 1
		cmp	eax, ecx
		cmovl	eax, ecx
#else
		sub	eax, ecx	; eax = x - y
		cdq			; edx = (x < y) ? -1 : 0
		not	edx		; edx = (x < y) ? 0 : -1
		and	eax, edx	; eax = (x < y) ? 0 : x - y
		add	eax, ecx	; eax = (x < y) ? y : x
					;     = max(x, y)
#endif
	}
#endif // _M_IX86
}
#endif // max

__inline
long	lmax(long x, long y)
{
	return x > y ? x : y;
}

__inline
unsigned long	ulmax(unsigned long x, unsigned long y)
{
	return x > y ? x : y;
}

__inline
long long	llmax(long long x, long long y)
{
#ifndef _M_IX86
	return x > y ? x : y;
#else
	__asm	// returns maximum value in edx:eax
	{
		mov	ecx, dword ptr x
		mov	eax, dword ptr x+4
					; eax:ecx = x
		sub	ecx, dword ptr y
		sbb	eax, dword ptr y+4
					; eax:ecx = x - y
		cdq			; edx = (x < y) ? -1 : 0
		not	edx		; edx = (x < y) ? 0 : -1
		and	ecx, edx
		and	edx, eax	; edx:ecx = (x < y) ? 0 : x - y
		add	ecx, dword ptr y
		adc	edx, dword ptr y+4
					; edx:ecx = (x < y) ? y : x
		mov	eax, ecx	; edx:eax = max(x, y)
	}
#endif // _M_IX86
}

__inline
unsigned int	umin(unsigned int x, unsigned int y)
{
#ifndef _M_IX86
	return x < y ? x : y;
#else
	__asm	// returns minimum value in eax
	{
		mov	eax, x
		mov	ecx, y
#if 1
		cmp	eax, ecx
		cmova	eax, ecx
#else
		sub	eax, ecx	; eax = x - y
		sbb	edx, edx	; edx = (x < y) ? ~0 : 0
		and	eax, edx	; eax = (x < y) ? x - y : 0
		add	eax, ecx	; eax = (x < y) ? x : y
					;     = min(x, y)
#endif
	}
#endif // _M_IX86
}

#ifndef min
__inline
int	min(int x, int y)
{
#ifndef _M_IX86
	return x < y ? x : y;
#else
	__asm	// returns minimum value in eax
	{
		mov	eax, x
		mov	ecx, y
#if 1
		cmp	eax, ecx
		cmovg	eax, ecx
#else
		sub	eax, ecx	; eax = x - y
		cdq			; edx = (x < y) ? -1 : 0
		and	eax, edx	; eax = (x < y) ? x - y : 0
		add	eax, ecx	; eax = (x < y) ? x : y
					;     = min(x, y)
#endif
	}
#endif // _M_IX86
}
#endif // min

__inline
long	lmin(long x, long y)
{
	return x < y ? x : y;
}

__inline
unsigned long	ulmin(unsigned long x, unsigned long y)
{
	return x < y ? x : y;
}

__inline
long long	llmin(long long x, long long y)
{
#ifndef _M_IX86
	return x < y ? x : y;
#else
	__asm	// returns maximum value in edx:eax
	{
		mov	ecx, dword ptr x
		mov	eax, dword ptr x+4
					; eax:ecx = x
		sub	ecx, dword ptr y
		sbb	eax, dword ptr y+4
					; eax:ecx = x - y
		cdq			; edx = (x < y) ? -1 : 0
		and	ecx, edx
		and	edx, eax	; edx:ecx = (x < y) ? x - y : 0
		add	ecx, dword ptr y
		adc	edx, dword ptr y+4
					; edx:ecx = (x < y) ? x : y
		mov	eax, ecx	; edx:eax = min(x, y)
	}
#endif // _M_IX86
}

__inline
int	sgn(int x)
{
#ifndef _M_IX86
#if 0
	if (x < 0)
		return -1;

	if (x > 0)
		return 1;

	return 0;
#elif 0
	return x < 0 ? -1 : x > 0;
#elif 1
	return (x > 0) - (x < 0);
#elif 1
	return (x >> 31) | (x != 0);
#else
	return (x >> 31) | ((0U - x) >> 31);
#endif
#else // _M_IX86
	__asm	// returns 1, 0, or -1 in eax
	{
#if 0
		xor	eax, eax
		xor	edx, edx
		cmp	eax, x
		setg	dl		; edx = (x < 0) ? 1 : 0
		setl	al		; eax = (x > 0) ? 1 : 0
		sub	eax, edx	; eax = (x > 0) - (x < 0)
#else
		mov	eax, x
		cdq			; edx = (x < 0) ? -1 : 0
		neg	eax		; CF = (x != 0) ? 1 : 0
		adc	edx, edx
		mov	eax, edx
#endif
	}
#endif // _M_IX86
}

__inline
long	lsgn(long x)
{
#ifndef _M_IX86
#if 0
	if (x < 0)
		return -1;

	if (x > 0)
		return 1;

	return 0;
#elif 0
	return x < 0 ? -1 : x > 0;
#elif 1
	return (x > 0) - (x < 0);
#elif 1
	return (x >> 31) | (x != 0);
#else
	return (x >> 31) | ((0UL - x) >> 31);
#endif
#else // _M_IX86
	__asm	// returns 1, 0, or -1 in eax
	{
#if 0
		xor	eax, eax
		xor	edx, edx
		cmp	eax, x
		setg	dl		; edx = (x < 0) ? 1 : 0
		setl	al		; eax = (x > 0) ? 1 : 0
		sub	eax, edx	; eax = (x > 0) - (x < 0)
#else
		mov	eax, x
		cdq			; edx = (x < 0) ? -1 : 0
		neg	eax		; CF = (x != 0) ? 1 : 0
		adc	edx, edx
		mov	eax, edx
#endif
	}
#endif // _M_IX86
}

__inline
int	llsgn(long long x)
{
#ifndef _M_IX86
#if 0
	if (x < 0)
		return -1;

	if (x > 0)
		return 1;

	return 0;
#elif 0
	return x < 0 ? -1 : x > 0;
#elif 1
	return (x > 0) - (x < 0);
#elif 1
	return (x >> 63) | (x != 0);
#else
	return (x >> 63) | ((0ULL - x) >> 63);
#endif
#else // _M_IX86
	__asm	// returns 1, 0, or -1 in eax
	{
		xor	edx, edx
		mov	eax, dword ptr x+4
					; eax = high dword of x
		cmp	edx, dword ptr x; CF = (low dword of x != 0) ? 1 : 0
		sbb	edx, eax	; edx = 0 - high dword of x - CF
		cdq			; edx = (x < 0) ? -1 : 0
		setl	al
		movzx	eax, al
		add	eax, edx
	}
#endif // _M_IX86
}

__inline
int	ullcmp(unsigned long long left, unsigned long long right)
{
#ifndef _M_IX86
#if 0
	if (left < right)
		return -1;

	if (left > right)
		return 1;

	return 0;
#elif 0
	return left < right ? -1 : left > right;
#else
	return (left > right) - (left < right);
#endif
#else // _M_IX86
	__asm	// returns -1, 0 or 1 in eax
	{
		xor	eax, eax	; eax = 0
		mov	ecx, dword ptr left
		mov	edx, dword ptr left+4
					; edx:ecx = left
		sub	ecx, dword ptr right
		sbb	edx, dword ptr right+4
					; edx:ecx = left - right
		seta	al		; eax = (left > right)
		sbb	eax, 0		; eax = (left > right)
					;     - (left < right)
					;     = {-1, 0, 1}
	}
#endif // _M_IX86
}

__inline
int	llcmp(long long left, long long right)
{
#ifndef _M_IX86
#if 0
	if (left < right)
		return -1;

	if (left > right)
		return 1;

	return 0;
#elif 0
	return left < right ? -1 : left > right;
#else
	return (left > right) - (left < right);
#endif
#else // _M_IX86
	__asm	// returns 1, 0 or -1 in eax
	{
		xor	eax, eax	; eax = 0
		mov	edx, dword ptr left
		cmp	edx, dword ptr right
					; CF = low dword of left < low dword of right
		mov	edx, dword ptr left+4
		sbb	edx, dword ptr right+4
					; edx:... = left - right,
					; eflags = (left - right)
		cdq			; edx = 0
		setg	al		; eax = (left > right)
		setl	dl		; edx = (left < right)
		sub	eax, edx	; eax = (left > right)
					;     - (left < right)
					;     = {1, 0, -1}
	}
#endif // _M_IX86
}

__inline	// average rounded down (floor)
unsigned long	avg32d(unsigned long x, unsigned long y)
{
#ifndef _M_IX86
	return (x & y) + ((x ^ y) >> 1);
#else
	__asm
	{
		mov	eax, x
		add	eax, y
		rcr	eax, 1
	}
#endif // _M_IX86
}

__inline	// average rounded up (ceiling)
unsigned long	avg32u(unsigned long x, unsigned long y)
{
#ifndef _M_IX86
	return (x | y) - ((x ^ y) >> 1);
#else
	__asm
	{
		mov	eax, x
		add	eax, y
		rcr	eax, 1
		adc	eax, 0
	}
#endif // _M_IX86
}

__inline		// average rounded down (floor)
unsigned long long	avg64d(unsigned long long x, unsigned long long y)
{
#ifndef _M_IX86
	return (x & y) + ((x ^ y) >> 1);
#else
	__asm
	{
		mov	eax, dword ptr x
		mov	edx, dword ptr x+4
		add	eax, dword ptr y
		adc	edx, dword ptr y+4
		rcr	edx, 1
		rcr	eax, 1
	}
#endif // _M_IX86
}

__inline		// average rounded up (ceiling)
unsigned long long	avg64u(unsigned long long x, unsigned long long y)
{
#ifndef _M_IX86
	return (x | y) - ((x ^ y) >> 1);
#else
	__asm
	{
		mov	eax, dword ptr x
		mov	edx, dword ptr x+4
		add	eax, dword ptr y
		adc	edx, dword ptr y+4
		rcr	edx, 1
		rcr	eax, 1
		adc	eax, 0
		adc	edx, 0
	}
#endif // _M_IX86
}

__inline	// modular multiplicative inverse
unsigned int	inverse32(unsigned int x)
{
	// NOTE: x must be co-prime (relative prime) to 2**32, i.e. odd!

	unsigned int	z, y = x;

	do	// Newton-Raphson iteration
	{
		z = y * x;
		y *= 2 - z;
	} while (z > 1);

	return y;
}

__inline		// modular multiplicative inverse
unsigned long long	inverse64(unsigned long long x)
{
	// NOTE: x must be co-prime (relative prime) to 2**64, i.e. odd!

	unsigned long long	z, y = x;

	do	// Newton-Raphson iteration
	{
		z = y * x;
		y *= 2 - z;
	} while (z > 1);

	return y;
}

__inline
int	parity8(unsigned char uc)
{
#ifndef _M_IX86
	uc ^= uc >> 4;

	return (0x69966996 >> uc) & 1;
#else
	__asm	// returns parity in eax
	{
		xor	eax, eax
		cmp	uc, al
		setpo	al
	}
#endif // _M_IX86
}

__inline
int	parity16(unsigned short us)
{
#ifndef _M_IX86
	us ^= us >> 8;
	us ^= us >> 4;

	return (0x69966996 >> us) & 1;
#else
	__asm	// returns parity in eax
	{
		mov	ax, us
		xor	al, ah
		setpo	al
		movzx	eax, al
	}
#endif // _M_IX86
}

__inline
int	parity32(unsigned long ul)
{
#ifndef _M_IX86
	ul ^= ul >> 16;
	ul ^= ul >> 8;
	ul ^= ul >> 4;

	return (0x69966996 >> ul) & 1;
#else
	__asm	// returns parity in eax
	{
#if 0
		mov	ax, word ptr ul
		xor	ax, word ptr ul+2
		xor	al, ah
		setpo	al
		movzx	eax, al
#else
		mov	eax, ul
		mov	ecx, eax
		bswap	eax
		xor	ecx, eax
		xor	eax, eax
		xor	cl, ch
		setpo	al
#endif
	}
#endif // _M_IX86
}

__inline
int	parity64(unsigned long long ull)
{
#ifndef _M_IX86
	unsigned long ul = (unsigned long) ull;

	ul ^= ull >> 32;
	ul ^= ul >> 16;
	ul ^= ul >> 8;
	ul ^= ul >> 4;

	return (0x69966996 >> ul) & 1;
#else
	__asm	// returns parity in eax
	{
#if 0
		mov	ax, word ptr ull
		xor	ax, word ptr ull+2
		xor	ax, word ptr ull+4
		xor	ax, word ptr ull+6
		xor	al, ah
		setpo	al
		movzx	eax, al
#else
		mov	eax, dword ptr ull
		xor	eax, dword ptr ull+4
		mov	ecx, eax
		bswap	eax
		xor	ecx, eax
		xor	eax, eax
		xor	cl, ch
		setpo	al
#endif
	}
#endif // _M_IX86
}

unsigned char	rev8(unsigned char uc);
#if 0
{
	uc = ((uc << 1) & '\xAA') | ((uc >> 1) & '\x55');
	uc = ((uc << 2) & '\xCC') | ((uc >> 2) & '\x33');

	return (uc << 4) | (uc >> 4);
}
#endif

unsigned short	rev16(unsigned short us);
#if 0
{
	us = ((us << 1) & L'\xAAAA') | ((us >> 1) & L'\x5555');
	us = ((us << 2) & L'\xCCCC') | ((us >> 2) & L'\x3333');
	us = ((us << 4) & L'\xF0F0') | ((us >> 4) & L'\x0F0F');

	return (us << 8) | (us >> 8);
}
#endif

unsigned long	rev32(unsigned long ul);
#if 0
{
	unsigned long	index, reverse = 0;

	// https://msdn.microsoft.com/en-us/library/wfd9z0bb.aspx

	while (_BitScanForward(&index, ul))

		// https://msdn.microsoft.com/en-us/library/hd0hzyf8.aspx
		// https://msdn.microsoft.com/en-us/library/z56sc6y4.aspx

		_bittestandreset(&ul, index), _bittestandset(&reverse, ~index);

	return reverse;
}
#elif 0
{
	ul = ((ul << 1) & 0xAAAAAAAA) | ((ul >> 1) & 0x55555555);
	ul = ((ul << 2) & 0xCCCCCCCC) | ((ul >> 2) & 0x33333333);
	ul = ((ul << 4) & 0xF0F0F0F0) | ((ul >> 4) & 0x0F0F0F0F);
	ul = ((ul << 8) & 0xFF00FF00) | ((ul >> 8) & 0x00FF00FF);

	return (ul << 16) | (ul >> 16);
}
#endif

unsigned long long	rev64(unsigned long long ull);
#if 0
{
	unsigned long		index;
	unsigned long long	reverse = 0;

	// https://msdn.microsoft.com/en-us/library/wfd9z0bb.aspx

	while (_BitScanForward64(&index, ull))

		// https://msdn.microsoft.com/en-us/library/hd0hzyf8.aspx
		// https://msdn.microsoft.com/en-us/library/z56sc6y4.aspx

		_bittestandreset64(&ull, index), _bittestandset64(&reverse, ~index);

	return reverse;
}
#elif 0
{
	ull = ((ull <<  1) & 0xAAAAAAAAAAAAAAAA) | ((ull >>  1) & 0x5555555555555555);
	ull = ((ull <<  2) & 0xCCCCCCCCCCCCCCCC) | ((ull >>  2) & 0x3333333333333333);
	ull = ((ull <<  4) & 0xF0F0F0F0F0F0F0F0) | ((ull >>  4) & 0x0F0F0F0F0F0F0F0F);
	ull = ((ull <<  8) & 0xFF00FF00FF00FF00) | ((ull >>  8) & 0x00FF00FF00FF00FF);
	ull = ((ull << 16) & 0xFFFF0000FFFF0000) | ((ull >> 16) & 0x0000FFFF0000FFFF);

	return (ull << 32) | (ull >> 32);
}
#endif

#define _bitswap_uchar	rev8
#define _bitswap_ushort	rev16
#define _bitswap_ulong	rev32
#define _bitswap_uint64	rev64

__declspec(deprecated("Don't use; included only for documentation!"))
__inline	// greatest common divisor
int		gcd(int a, int b)
{
	if (a < 0)
		a = -a;

	if (b < 0)
		b = -b;

	if (b != 0)
		while (((a %= (unsigned) b) != 0) && ((b %= (unsigned) a) != 0))
			continue;

	return a | b;
}

__inline	// greatest common divisor
unsigned long	gcd32(unsigned long p, unsigned long q)
{
#if 0 // Euclid's (modern) algorithm
	unsigned long	r;

	if (p == q)
		return p;

	if (p == 0)
		return q;

	if (q == 0)
		return p;

	do
		r = p % q, p = q, q = r;
	while (q != 0);

	return p;
#else // Stein's (binary) algorithm
#ifndef _M_IX86
	unsigned long	s = 0, t;

	if (p == q)
		return p;

	if (p == 0)
		return q;

	if (q == 0)
		return p;

	while (((p | q) & 1) == 0)
		p >>= 1, q >>= 1, s++;

	while ((p & 1) == 0)
		p >>= 1;

	do
	{
		while ((q & 1) == 0)
			q >>= 1;

		if (p > q)
			t = q, q = p, p = t;

		q -= p;
	} while (q != 0);

	return p << s;
#else
	__asm
	{
		mov	edx, p		; edx = p
		mov	eax, q		; eax = q
		bsf	ecx, edx	; ecx = exponent of greatest power of 2 dividing p
		jz	short done

		cmp	eax, edx
		je	short done

		mov	ebx, eax	; ebx = q
		mov	eax, edx	; eax = p
		bsf	edx, ebx	; edx = exponent of greatest power of 2 dividing q
		jz	short done

		cmp	edx, ecx
		cmova	edx, ecx	; edx = minimum of both exponents
					;     = exponent of greatest power of 2 dividing p and q
		shr	eax, cl		; eax = p / greatest power of 2 dividing p
					;     = p'
	next:
		bsf	ecx, ebx	; ecx = exponent of greatest power of 2 dividing q
		shr	ebx, cl		; ebx = q / greatest power of 2 dividing q
					;     = q'
		mov	ecx, eax	; ecx = p'
		cmp	ebx, eax
		cmovb	eax, ebx	; eax = (q' < p') ? q' : p'
					;     = p"
		cmovb	ebx, ecx	; ebx = (q' < p') ? p' : q'
					;     = q"
		sub	ebx, eax	; ebx = (q' < p') ? p' - q' : q' - p'
					;     = q" - p"
		jnz	short next

		mov	ecx, edx	; ecx = minimum of both exponents
		shl	eax, cl		; eax = p" * greatest power of 2 dividing p and q
					;     = gcd(p, q)
	done:
	}
#endif // _M_IX86
#endif
}

__inline		// greatest common divisor
unsigned long long	gcd64(unsigned long long p,
			      unsigned long long q)
{
	unsigned long long	r;
	unsigned long		s, t;

	if (p == q)
		return p;

	if (p == 0)
		return q;

	if (q == 0)
		return p;
#ifdef _M_AMD64
	_BitScanForward64(&s, p | q);
	p >>= s;
	q >>= s;

	_BitScanForward64(&t, p);
	p >>= t;

	do	// https://msdn.microsoft.com/en-us/library/wfd9z0bb.aspx
	{
		_BitScanForward64(&t, q);
		q >>= t;

		if (p > q)
			r = q, q = p, p = r;
	} while (q -= p);

	return p << s;
#elif _M_IX86
	if (((unsigned long) p | (unsigned long) q) == 0)
		s = 32, p >>= 32, q >>= 32;
	else
		s = 0;

	// https://msdn.microsoft.com/en-us/library/wfd9z0bb.aspx

	_BitScanForward(&t, (unsigned long) p | (unsigned long) q);

	// https://msdn.microsoft.com/en-us/library/wds0z27f.aspx

	p = __ull_rshift(p, t);
	q = __ull_rshift(q, t);
	s += t;

	if ((unsigned long) p == 0)
		p >>= 32;

	_BitScanForward(&t, (unsigned long) p);
	p = __ull_rshift(p, t);

	do
	{
		if ((unsigned long) q == 0)
			q >>= 32;

		_BitScanForward(&t, (unsigned long) q);
		q = __ull_rshift(q, t);

		if (p > q)
			r = q, q = p, p = r;
	} while (q -= p);

	// https://msdn.microsoft.com/en-us/library/8dt0hdsy.aspx

	return __ll_lshift(p, s);
#else
	while (((p | q) & 1) == 0)
		p >>= 1, q >>= 1, s++;

	while ((p & 1) == 0)
		p >>= 1;

	do
	{
		while ((q & 1) == 0)
			q >>= 1;

		if (p > q)
			r = q, q = p, p = r;
	} while (q -= p);

	return p << s;
#endif
}

// https://msdn.microsoft.com/en-us/library/kw1ee768.aspx

int	abs(int x);

#pragma intrinsic(abs)

__inline	// least common multiple
int		lcm(int a, int b)
{
	if ((a == 0) || (b == 0))
		return 0;
#ifdef _M_IX86
	return __emuldiv(abs(a), abs(b), gcd(a, b));
#else
	return a / gcd(a, b) * b;
#endif
}

__inline	// least common multiple
unsigned long	lcm32(unsigned long a, unsigned long b)
{
	if ((a == 0) || (b == 0))
		return 0;
#ifdef _M_IX86
	return __emuldivu(a, b, gcd32(a, b));
#else
	return a / gcd32(a, b) * b;
#endif
}

__inline		// least common multiple
unsigned long long	lcm64(unsigned long long a, unsigned long long b)
{
	if ((a == 0) || (b == 0))
		return 0;
#if 0
	unsigned long long	c = gcd64(a, b);

	// https://msdn.microsoft.com/en-us/library/3dayytw9.aspx

	b = __umul128(a, b, &a);

	// https://docs.microsoft.com/en-us/cpp/intrinsics/udiv128

	return _udiv128(a, b, c, NULL);
#else
	return a / gcd64(a, b) * b;
#endif
}

__inline	// integer square root
unsigned long	sqrt32(unsigned long p)
{
#if 0
	unsigned long	q = 1 << 16, r;

	if (p <= 1)
		return p;

	do	// Newton-Raphson iteration
	{
		r = q;
		q += p / q >> 1;
	} while (q < r);
#else
	unsigned long	q = 1 << 30, r = 0;

	if (p <= 1)
		return p;

	while (q > p)
		q >>= 2;

	while (q != 0)
	{
		if (p >= q + r)
		{
			p -= q + r;
			r += q << 1;
		}

		r >>= 1;
		q >>= 2;
	}
#endif
	return r;
}

__inline		// integer square root
unsigned long long	sqrt64(unsigned long long p)
{
#if 0
	unsigned long long	q = 1ULL << 32, r;

	if (p <= 1)
		return p;

	do	// Newton-Raphson iteration
	{
		r = q;
		q += p / q >> 1;
	} while (q < r);
#else
	unsigned long long	q = 1ULL << 62, r = 0;

	if (p <= 1)
		return p;

	while (q > p)		// greatest power of 4 in p
		q >>= 2;

	while (q != 0)
	{
		if (p >= q + r)
		{
			p -= q + r;
			r += q << 1;
		}

		r >>= 1;
		q >>= 2;
	}
#endif
	return r;
}

__inline	// integer cube root
unsigned long	cbrt32(unsigned long q)
{
	unsigned long	r = 0, s = 33, t;

	do
	{
		r += r;
		s -= 3;
		t = (3 * r * (r + 1) + 1) << s;

		if (q < t)
			continue;

		q -= t;
		r += 1;
	} while (s > 0);

	return r;
}

__inline	// integer cube root
unsigned long	cbrt64(unsigned long long q)
{
	unsigned long		r = 0, s = 66;
	unsigned long long	t;

	do
	{
		r += r;
		s -= 3;

		// https://msdn.microsoft.com/en-us/library/d2s81xt0.aspx

		t = __emulu(r, r + 1) * 3 + 1;

		if ((q >> s) < t)
			continue;

		q -= t << s;
		r += 1;
	} while (s > 0);

	return r;
}

#ifndef _WINNT_
// https://msdn.microsoft.com/en-us/library/wfd9z0bb.aspx

unsigned char	_BitScanForward(unsigned long *Index, unsigned long Mask);

// https://msdn.microsoft.com/en-us/library/fbxyd7zd.aspx

unsigned char	_BitScanReverse(unsigned long *Index, unsigned long Mask);

#pragma intrinsic(_BitScanForward, _BitScanReverse)
#endif // _WINNT_

__inline	// leading zero count
unsigned long	lzcnt32(unsigned long ul)
{
#ifndef _M_IX86
	unsigned long	index;

	// https://msdn.microsoft.com/en-us/library/fbxyd7zd.aspx
#if 0
	return _BitScanReverse(&index, ul) ? 31 - index : 32;
#else
	return _BitScanReverse(&index, ul) ? 31 ^ index : 32;
#endif
#else
	__asm	// returns number of leading '0' bits in eax
	{
		mov	eax, 32
#if 1
		mov	edx, ul
		bsr	ecx, edx	; ecx = index of most significant '1' bit
		cmovz	ecx, edx	; ecx = (ul == 0) ? 0 : ecx
		neg	edx		; CF = (ul == 0) ? 0 : 1
		sbb	eax, ecx	; eax = 32 - ecx - CF
#else
		bsr	ecx, ul		; ecx = index of most significant '1' bit
		jz	short done

		stc
		sbb	eax, ecx	; eax = 31 - index of most significant '1' bit
	done:
#endif
	}
#endif // _M_IX86
}

__inline	// trailing zero count
unsigned long	tzcnt32(unsigned long ul)
{
#ifndef _M_IX86
	unsigned long	index;

	// https://msdn.microsoft.com/en-us/library/wfd9z0bb.aspx

	return _BitScanForward(&index, ul) ? index : 32;
#else
	__asm	// returns number of trailing '0' bits in eax
	{
		mov	eax, 32
		bsf	ecx, ul		; ecx = index of least significant '1' bit
		cmovnz	eax, ecx	; eax = (ul == 0) ? 32 : ecx
	}
#endif // _M_IX86
}

#ifdef _WIN64
#ifndef _WINNT_
// https://msdn.microsoft.com/en-us/library/wfd9z0bb.aspx

unsigned char	_BitScanForward64(unsigned long *Index, unsigned long long Mask);

// https://msdn.microsoft.com/en-us/library/fbxyd7zd.aspx

unsigned char	_BitScanReverse64(unsigned long *Index, unsigned long long Mask);

#pragma intrinsic(_BitScanForward64, _BitScanReverse64)
#endif // _WINNT_

__inline	// leading zero count
unsigned long	lzcnt64(unsigned long long ull)
{
	unsigned long	index;

	// https://msdn.microsoft.com/en-us/library/fbxyd7zd.aspx
#if 0
	return _BitScanReverse64(&index, ull) ? 63 - index : 64;
#else
	return _BitScanReverse64(&index, ull) ? 63 ^ index : 64;
#endif
}

__inline	// trailing zero count
unsigned long	tzcnt64(unsigned long long ull)
{
	unsigned long	index;

	// https://msdn.microsoft.com/en-us/library/wfd9z0bb.aspx

	return _BitScanForward64(&index, ull) ? index : 64;
}
#elif 0
__inline	// leading zero count
unsigned long	lzcnt64(unsigned long long ull)
{
	unsigned long	count = 1;

	if (ull == 0)
		return 63;

	if ((signed long long) ull < 0)
		return 0;

	if (((unsigned long) (ull >> 32) == 0)
		count += 32;
	else
		ull >>= 32;

	if (((unsigned long) ull & 0xFFFF0000) == 0)
		(unsigned long) ull <<= 16, count += 16;

	if (((unsigned long) ull & 0xFF000000) == 0)
		(unsigned long) ull <<= 8, count += 8;

	if (((unsigned long) ull & 0xF0000000) == 0)
		(unsigned long) ull <<= 4, count += 4;

	if (((unsigned long) ull & 0xC0000000) == 0)
		(unsigned long) ull <<= 2, count += 2;

	return count - ((signed long) ull < 0);
}

__inline	// trailing zero count
unsigned long	tzcnt64(unsigned long long ull)
{
	unsigned long	count = 1;

	if (ull == 0)
		return 63;

	if (ull & 1)
		return 0;

	if (((unsigned long) ull == 0)
		ull >>= 32, count += 32;

	if (((unsigned long) ull & 65535) == 0)
		(unsigned long) ull >>= 16, count += 16;

	if (((unsigned long) ull & 255) == 0)
		(unsigned long) ull >>= 8, count += 8;

	if (((unsigned long) ull & 15) == 0)
		(unsigned long) ull >>= 4, count += 4;

	if (((unsigned long) ull & 3) == 0)
		(unsigned long) ull >>= 2, count += 2;

	return count - ((unsigned long) ull & 1);
}
#endif // _WIN64

__inline	// population count
unsigned long	popcnt32(unsigned long ul)
{
#if 0
	unsigned long	count = 0;

	while (ul != 0)
		ul &= ul - 1, count++;

	return count;
#elif 0
	unsigned long	index, count = 0;

	// https://msdn.microsoft.com/en-us/library/wfd9z0bb.aspx

	while (_BitScanForward(&index, ul))

		// https://msdn.microsoft.com/en-us/library/hd0hzyf8.aspx

		_bittestandreset(&ul, index), count++;

	return count;
#else
	ul -= (ul >> 1) & 0x55555555;
	ul = ((ul >> 2) & 0x33333333) + (ul & 0x33333333);
	ul = ((ul >> 4) + ul) & 0x0F0F0F0F;
#if 0
	ul += ul >> 8;
	ul += ul >> 16;

	return ul & 63;
#else
	return (ul * 0x01010101) >> 24;
#endif
#endif
}

__inline	// population count
unsigned long	popcnt64(unsigned long long ull)
{
#if 0
	unsigned long	count = 0;

	while (ull != 0)
		ull &= ull - 1, count++;

	return count;
#else
	ull -= (ull >> 1) & (~0ULL / 3);
	ull = ((ull >> 2) & (~0ULL / 5)) + (ull & (~0ULL / 5));
	ull = ((ull >> 4) + ull) & (~0ULL / 17);
#ifndef _WIN64
	ull += ull >> 8;
	ull += ull >> 16;
	ull += ull >> 32;

	return (unsigned long) ull & 127;
#else
	return (unsigned long) ((ull * (~0ULL / 255)) >> 56);
#endif // _WIN64
#endif
}

#ifdef _M_IX86
__inline	// leading zero count
unsigned long	lzcnt64(unsigned long long ull)
{
	__asm	// returns number of leading '0' bits in eax
	{
		mov	eax, 32
		bsr	ecx, dword ptr ull+4
					; ecx = index of most significant '1' bit in high dword
		jnz	short skip	; high dword <> 0?

		add	eax, eax
		bsr	ecx, dword ptr ull
					; ecx = index of most significant '1' bit in low dword
		jz	short done	; low dword = 0?
	skip:
		stc
		sbb	eax, ecx	; eax = 63 - index of most significant '1' bit in qword
	done:
	}
}

__inline	// trailing zero count
unsigned long	tzcnt64(unsigned long long ull)
{
	__asm	// returns number of trailing '0' bits in eax
	{
		bsf	eax, dword ptr ull
					; eax = index of least significant '1' bit in low dword
		jnz	short done	; low dword <> 0?

		mov	eax, 64
		bsf	ecx, dword ptr ull+4
					; ecx = index of least significant '1' bit in high dword
		jz	short done	; high dword = 0?

		shr	eax, 1
		add	eax, ecx	; eax = index of least significant '1' bit in qword
	done:
	}
}
#endif // _M_IX86

#ifdef INTRINSIC
// https://msdn.microsoft.com/en-us/library/t5e2f3sc.aspx

unsigned char	_rotl8(unsigned char value, unsigned char shift);
unsigned short	_rotl16(unsigned short value, unsigned char shift);

// https://msdn.microsoft.com/en-us/library/yy0728bz.aspx

unsigned char	_rotr8(unsigned char value, unsigned char shift);
unsigned short	_rotr16(unsigned short value, unsigned char shift);

// https://msdn.microsoft.com/en-us/library/5cc576c4.aspx

unsigned int		_rotl(unsigned int value, int shift);
unsigned int		_rotr(unsigned int value, int shift);
unsigned long long	_rotl64(unsigned long long value, int shift);
unsigned long long	_rotr64(unsigned long long value, int shift);

// https://msdn.microsoft.com/en-us/library/a0w705h5.aspx

unsigned long	_lrotl(unsigned long value, int shift);
unsigned long	_lrotr(unsigned long value, int shift);

// https://msdn.microsoft.com/en-us/library/5704bbxw.aspx
// https://msdn.microsoft.com/en-us/library/tzkfha43.aspx

#pragma intrinsic(_rotl8, _rotl16)
#pragma intrinsic(_rotr8, _rotr16)
#pragma intrinsic(_rotl, _rotr, _rotl64, _rotr64)
#pragma intrinsic(_lrotl, _lrotr)
#endif // INTRINSIC

#ifdef _WIN64
#pragma intrinsic(__shiftleft128, __shiftright128, __umulh, _umul128, _mul128)

// companion routine for __shiftleft128() intrinsic
// https://msdn.microsoft.com/en-us/library/szzkhewe.aspx

__inline		// returns low part!
unsigned long long	__rotateleft128(unsigned long long LowPart,
			                unsigned long long HighPart,
			                unsigned char      Shift)
{
	return __shiftleft128(HighPart, LowPart, Shift);
}

// companion routine for __shiftright128() intrinsic
// https://msdn.microsoft.com/en-us/library/5az7sk38.aspx

__inline		// returns high part!
unsigned long long	__rotateright128(unsigned long long LowPart,
			                 unsigned long long HighPart,
			                 unsigned char      Shift)
{
	return __shiftright128(HighPart, LowPart, Shift);
}
#else // _WIN64
__inline		// returns low part!
unsigned long long	__rotateleft128(unsigned long long LowPart,
			                unsigned long long HighPart,
			                unsigned char      Shift)
{
	return (LowPart << Shift) | (HighPart >> (64 - Shift));
}

__inline		// returns high part!
unsigned long long	__rotateright128(unsigned long long LowPart,
			                 unsigned long long HighPart,
			                 unsigned char      Shift)
{
	return (HighPart >> Shift) | (LowPart << (64 - Shift));
}

// https://msdn.microsoft.com/en-us/library/szzkhewe.aspx

__inline		// returns high part!
unsigned long long	__shiftleft128(unsigned long long LowPart,
			               unsigned long long HighPart,
			               unsigned char      Shift)
{
	return (HighPart << Shift) | (LowPart >> (64 - Shift));
}

// https://msdn.microsoft.com/en-us/library/5az7sk38.aspx

__inline		// returns low part!
unsigned long long	__shiftright128(unsigned long long LowPart,
			                unsigned long long HighPart,
			                unsigned char      Shift)
{
	return (LowPart >> Shift) | (HighPart << (64 - Shift));
}
#endif // _WIN64

#ifdef _M_IX86
#if _MSC_VER >= 1920 // MSC 19.20 alias 2019
// https://docs.microsoft.com/en-us/cpp/intrinsics/div64

long	_div64(long long dividend,
	       long      divisor,
	       long      *remainder);

// https://docs.microsoft.com/en-us/cpp/intrinsics/udiv64

unsigned long	_udiv64(unsigned long long dividend,
		        unsigned long      divisor,
		        unsigned long      *remainder);

#pragma intrinsic(_div64, _udiv64)
#else
// https://docs.microsoft.com/en-us/cpp/intrinsics/div64

__inline
long	_div64(long long dividend,
	       long      divisor,
	       long      *remainder)
{
	__asm // returns quotient in eax
	{
		mov	eax, dword ptr dividend
		mov	edx, dword ptr dividend+4
		idiv	divisor

		mov	ecx, remainder
		test	ecx, ecx
		jz	short done

		mov	[ecx], edx
	done:
	}
}

// https://docs.microsoft.com/en-us/cpp/intrinsics/udiv64

__inline
unsigned long	_udiv64(unsigned long long dividend,
		        unsigned long      divisor,
		        unsigned long      *remainder)
{
	__asm // returns quotient in eax
	{
		mov	eax, dword ptr dividend
		mov	edx, dword ptr dividend+4
		div	divisor

		mov	ecx, remainder
		test	ecx, ecx
		jz	short done

		mov	[ecx], edx
	done:
	}
}
#endif // _MSC_VER

// https://msdn.microsoft.com/en-us/library/jj635748.aspx

#define UnsignedMultiplyHigh	__umulh

// https://msdn.microsoft.com/en-us/library/67xf5cy5.aspx

__inline
unsigned long long	__umulh(unsigned long long multiplicand,
			        unsigned long long multiplier)
{
	__asm	// returns high qword of product in edx:eax
	{
		mov	ebx, dword ptr multiplier
		mov	eax, dword ptr multiplicand
		mul	ebx
		mov	ecx, edx
		mov	eax, dword ptr multiplicand+4
		mul	ebx
		xor	ebx, ebx
		add	ecx, eax
		adc	ebx, edx
		mov	eax, dword ptr multiplicand
		mul	dword ptr multiplier+4
		add	ecx, eax
		adc	ebx, edx
		sbb	ecx, ecx
		mov	eax, dword ptr multiplicand+4
		mul	dword ptr multiplier+4
		neg	ecx
		add	eax, ebx
		adc	edx, ecx
	}
}

// https://msdn.microsoft.com/en-us/library/jj635744.aspx

#define MultiplyHigh	__mulh

// https://msdn.microsoft.com/en-us/library/17wa28cw.aspx

__inline
long long	__mulh(long long multiplicand,
		       long long multiplier)
{
	__asm	// returns high qword of product in edx:eax
	{
		mov	ebx, dword ptr multiplier
		mov	eax, dword ptr multiplicand
		mul	ebx
		mov	ecx, edx
		mov	eax, dword ptr multiplicand+4
		mul	ebx
		xor	ebx, ebx
		add	ecx, eax
		adc	ebx, edx
		mov	eax, dword ptr multiplicand
		mul	dword ptr multiplier+4
		add	ecx, eax
		adc	ebx, edx
		sbb	ecx, ecx
		mov	eax, dword ptr multiplicand+4
		mul	dword ptr multiplier+4
		neg	ecx
		add	eax, ebx
		adc	edx, ecx
#if 0
		cmp	dword ptr multiplier+4, 0
		jge	short nosign1
		sub	eax, dword ptr multiplicand
		sbb	edx, dword ptr multiplicand+4
	nosign1:
		cmp	dword ptr multiplicand+4, 0
		jge	short nosign2
		sub	eax, dword ptr multiplier
		sbb	edx, dword ptr multiplier+4
	nosign2:
#else
		mov	ecx, dword ptr multiplier+4
		sar	ecx, 31
		mov	ebx, ecx
		and	ecx, dword ptr multiplicand
		and	ebx, dword ptr multiplicand+4
		sub	eax, ecx
		sbb	edx, ebx
		mov	ecx, dword ptr multiplicand+4
		sar	ecx, 31
		mov	ebx, ecx
		and	ecx, dword ptr multiplier
		and	ebx, dword ptr multiplier+4
		sub	eax, ecx
		sbb	edx, ebx
#endif
	}
}

// https://msdn.microsoft.com/en-us/library/hh802933.aspx

#define UnsignedMultiply128	_umul128

// https://msdn.microsoft.com/en-us/library/3dayytw9.aspx

__inline
unsigned long long	_umul128(unsigned long long multiplicand,
			         unsigned long long multiplier,
			         unsigned long long *highproduct)
{
	__asm	// returns low qword of product in edx:eax
	{
		mov	ebx, dword ptr multiplier
		mov	eax, dword ptr multiplicand
		mul	ebx
		push	eax
		mov	ecx, edx
		mov	eax, dword ptr multiplicand+4
		mul	ebx
		xor	ebx, ebx
		add	ecx, eax
		adc	ebx, edx
		mov	eax, dword ptr multiplicand
		mul	dword ptr multiplier+4
		add	ecx, eax
		adc	ebx, edx
		push	ecx
		sbb	ecx, ecx
		mov	eax, dword ptr multiplicand+4
		mul	dword ptr multiplier+4
		neg	ecx
		add	eax, ebx
		adc	edx, ecx
		mov	ebx, highproduct
		mov	[ebx+4], edx
		mov	[ebx], eax
		pop	edx
		pop	eax
	}
}

// https://msdn.microsoft.com/en-us/library/hh802931.aspx

#define Multiply128	_mul128

// https://msdn.microsoft.com/en-us/library/82cxdw50.aspx

#if 0
__declspec(deprecated("Don't use; included only for documentation!"))
long long	_mul128(long long Multiplicand,
		        long long Multiplier,
		        long long *HighProduct)
{
	long long const	LowProduct = _umul128(Multiplicand, Multiplier, HighProduct);

	if (Multiplier < 0)
		*HighProduct -= Multiplicand;

	if (Multiplicand < 0)
		*HighProduct -= Multiplier;

	return LowProduct;
}
#else
__inline
long long	_mul128(long long multiplicand,
		        long long multiplier,
		        long long *highproduct)
{
	__asm	// returns low qword of product in edx:eax
	{
		mov	ebx, dword ptr multiplier
		mov	eax, dword ptr multiplicand
		mul	ebx
		push	eax
		mov	ecx, edx
		mov	eax, dword ptr multiplicand+4
		mul	ebx
		xor	ebx, ebx
		add	ecx, eax
		adc	ebx, edx
		mov	eax, dword ptr multiplicand
		mul	dword ptr multiplier+4
		add	ecx, eax
		adc	ebx, edx
		push	ecx
		sbb	ecx, ecx
		mov	eax, dword ptr multiplicand+4
		mul	dword ptr multiplier+4
		neg	ecx
		add	eax, ebx
		adc	edx, ecx
#if 0
		cmp	dword ptr multiplier+4, 0
		jge	short nosign1
		sub	eax, dword ptr multiplicand
		sbb	edx, dword ptr multiplicand+4
	nosign1:
		cmp	dword ptr multiplicand+4, 0
		jge	short nosign2
		sub	eax, dword ptr multiplier
		sbb	edx, dword ptr multiplier+4
	nosign2:
#else
		mov	ecx, dword ptr multiplier+4
		sar	ecx, 31
		mov	ebx, ecx
		and	ecx, dword ptr multiplicand
		and	ebx, dword ptr multiplicand+4
		sub	eax, ecx
		sbb	edx, ebx
		mov	ecx, dword ptr multiplicand+4
		sar	ecx, 31
		mov	ebx, ecx
		and	ecx, dword ptr multiplier
		and	ebx, dword ptr multiplier+4
		sub	eax, ecx
		sbb	edx, ebx
#endif
		mov	ecx, highproduct
		mov	[ecx+4], edx
		mov	[ecx], eax
		pop	edx
		pop	eax
	}
}
#endif

// companion routine for _umul128() intrinsic

__inline		// widening carry-less multiplication
unsigned long long	_clmul128(unsigned long long multiplicand,
			          unsigned long long multiplier,
			          unsigned long long *highproduct)
{
#if 0
	unsigned long long	ull, low = 0, high = 0;

	while (multiplier)
	{
		low ^= _umul128(multiplicand, multiplier & (0 - multiplier), &ull);
		high ^= ull;
#if 0
		multiplier ^= multiplier & (0 - multiplier);
#else
		multiplier &= multiplier - 1;
#endif
	}

	*highproduct = high;

	return low;
#else
	__asm	// returns low qword of carry-less product in edx:eax
	{
		xor	ebx, ebx
		xor	edx, edx
		xor	eax, eax	; ebx:edx:eax = carry-less product" = 0

		mov	esi, dword ptr multiplier+4
					; esi = high dword of multiplier
		bsr	edi, esi	; edi = index of most significant '1' bit
		jz	short zero_high	; high dword of multiplier = 0?

		lea	ecx, [eax+31]	; ecx = index of most significant bit = 31
	loop_high:
		sub	ecx, edi	; ecx = length of '0' bits run
		shld	ebx, edx, cl
		shld	edx, eax, cl
		shl	eax, cl
		xor	edx, dword ptr multiplicand+4
		xor	eax, dword ptr multiplicand
					; ebx:edx:eax = carry-less product"
					;             * 2**length of '0' bits run
					;             ^ multiplicand
					;             = carry-less product'
		btr	esi, edi	; esi = multiplier
					;     ^ 2**index of most significant '1' bit
					;     = multiplier'
		mov	ecx, edi
		bsr	edi, esi	; edi = index of most significant '1' bit
		jnz	short loop_high	; multiplier' <> 0?

		shld	ebx, edx, cl
		shld	edx, eax, cl
		shl	eax, cl		; ebx:edx:eax = carry-less product'
					;             * 2**length of last '0' bits run
					;             = carry-less product of multiplicand
					;               and high dword of multiplier
	zero_high:
		push	ebx
		push	edx
		push	eax		; [esp] = carry-less product of multiplicand
					;         and high dword of multiplier
		xor	ebx, ebx
		xor	edx, edx
		xor	eax, eax	; ebx:edx:eax = carry-less product" = 0

		mov	esi, dword ptr multiplier
					; esi = low dword of multiplier
		bsr	edi, esi	; edi = index of most significant '1' bit
		jz	short zero_low	; low dword of multiplier = 0?

		lea	ecx, [eax+31]	; ecx = index of most significant bit = 31
	loop_low:
		sub	ecx, edi	; ecx = length of '0' bits run
		shld	ebx, edx, cl
		shld	edx, eax, cl
		shl	eax, cl
		xor	edx, dword ptr multiplicand+4
		xor	eax, dword ptr multiplicand
					; ebx:edx:eax = carry-less product"
					;             * 2**length of '0' bits run
					;             ^ multiplicand
					;             = carry-less product
		btr	esi, edi	; esi = multiplier
					;     ^ 2**index of most significant '1' bit
					;     = multiplier'
		mov	ecx, edi
		bsr	edi, esi	; edi = index of most significant '1' bit
		jnz	short loop_low	; multiplier' <> 0?

		shld	ebx, edx, cl
		shld	edx, eax, cl
		shl	eax, cl		; ebx:edx:eax = carry-less product'
					;             * 2**length of last '0' bits run
					;             = carry-less product of multiplicand
					;               and low dword of multiplier
	zero_low:
		pop	ecx
		pop	edi
		pop	esi		; esi:edi:ecx = carry-less product of multiplicand
					;               and high dword of multiplier
		xor	edx, ecx	; edx:eax = low qword of carry-less product
		xor	edi, ebx	; esi:edi = high qword of carry-less product
		mov	ecx, highproduct
					; ecx = address of high qword of carry-less product
		mov	[ecx+4], esi
		mov	[ecx], edi	; [ecx] = high qword of carry-less product
	}
#endif
}

// https://msdn.microsoft.com/en-us/library/d2s81xt0.aspx

unsigned long long	__emulu(unsigned long multiplicand, unsigned long multiplier);

#pragma intrinsic(__emulu)

// companion routine for _umul128() intrinsic
// https://msdn.microsoft.com/en-us/library/3dayytw9.aspx
// https://msdn.microsoft.com/en-us/library/hh802933.aspx
// https://docs.microsoft.com/en-us/cpp/intrinsics/udiv128

#if 1
unsigned long long	_udiv128(unsigned long long DividendHigh,
			         unsigned long long DividendLow,
			         unsigned long long Divisor,
			         unsigned long long *Remainder);
#else
__declspec(deprecated("Don't use; included only for documentation!"))
unsigned long long	_udiv128(unsigned long long DividendHigh,
			         unsigned long long DividendLow,
			         unsigned long long Divisor,
			         unsigned long long *Remainder)
{
	unsigned long long	Quotient, Modulus, DividendMid;
	unsigned long		High, Low, Shift;
#if 0
	if (Divisor == 0)
	{				// divisor equal 0:
		__asm	_emit	0D4h	//  raise "divide exception",
		__asm	_emit	000h	//   set remainder to invalid value
					//    and return largest integer
		if (Remainder != NULL)
			*Remainder = ~0ULL /* % Divisor */;

		return ~0ULL /* / Divisor */;
	}
#endif
	if (DividendHigh >= Divisor)
	{				// quotient overflow or divisor equal 0:
		if (Divisor == 0)	//  raise "integer overflow exception"
		{			//   or "divide by 0 exception"
			__asm	_emit	0D4h
			__asm	_emit	000h
		}
		else
			__asm	int	4
					// set remainder to invalid value
		if (Remainder != NULL)	//  and return largest integer
			*Remainder = ~0ULL;

		return ~0ULL;
	}

	if (DividendHigh == 0)
	{				// dividend less 2**64
		if (DividendLow == 0)
		{			// dividend equal 0
			if (Remainder != NULL)
				*Remainder = 0;

			return 0;
		}

		if (DividendLow == Divisor)
		{			// dividend equal divisor
			if (Remainder != NULL)
				*Remainder = 0;

			return 1;
		}
#if 0
		if (Divisor == 1)
		{			// divisor equal 1 (2**0)
			if (Remainder != NULL)
				*Remainder = 0;

			return DividendLow;
		}
#endif
		Modulus = Divisor - 1;

		if ((Divisor & Modulus) == 0)
		{			// divisor equal 2**n
			if (Remainder != NULL)
				*Remainder = DividendLow & Modulus;

			if (Modulus != 0)
			{		// divisor greater 1
				Shift = tzcnt64(Divisor);

				DividendLow >>= Shift;
			}

			return DividendLow;
		}
#if 0	// NOTE: MSC generates separate calls to _aullrem() and _aulldiv()
	//       instead of a single call to _aulldvrm()!

		if (Remainder != NULL)
			*Remainder = DividendLow % Divisor;

		return DividendLow / Divisor;
#else
		Quotient = DividendLow / Divisor;
		Modulus = DividendLow % Divisor;

		if (Remainder != NULL)
			*Remainder = Modulus;

		return Quotient;
#endif
	}

	Modulus = Divisor - 1;

	if ((Divisor & Modulus) == 0)
	{				// divisor equal 2**n
		if (Remainder != NULL)
			*Remainder = DividendLow & Modulus;

		if (Modulus != 0)
		{			// divisor greater 1
			Shift = tzcnt64(Divisor);

			DividendLow >>= Shift;
			DividendLow |= DividendHigh << (64 - Shift);
		}

		return DividendLow;
	}

#define _HIGH(ULL)	(unsigned long) (ULL >> 32)
#define _LOW(ULL)	(unsigned long) ULL

	if (_HIGH(Divisor) == 0)
	{				// dividend less 2**96, divisor less 2**32:
					//  "long" division
		Quotient = _udiv64((DividendHigh << 32) | _HIGH(DividendLow),
		                   _LOW(Divisor),
		                   &High);
		DividendMid = High;
		DividendMid <<= 32;
		DividendMid |= _LOW(DividendLow);
		Quotient <<= 32;
		Quotient |= _udiv64(DividendMid, _LOW(Divisor), &Low);

		if (Remainder != NULL)
			*Remainder = Low;

		return Quotient;
	}
#if 0 // binary "long" division
	Shift = 64;

	do
	{
		DividendMid = (long long) DividendHigh >> 63;
		DividendHigh <<= 1;
		DividendHigh |= DividendLow >> 63;
		DividendLow <<= 1;

		if (Divisor > (DividendMid | DividendHigh))
			continue;

		DividendHigh -= Divisor;
		DividendLow |= 1;
	} while (--Shift != 0);

	if (Remainder != NULL)
		*Remainder = DividendHigh;

	return DividendLow;
#else // Algorithm D (Donald Ervin Knuth)
#if 0
	Shift = lzcnt64(Divisor);
#else
	Shift = lzcnt32(_HIGH(Divisor));
#endif
	if (Shift != 0)
	{				// align divisor and dividend to the left
		Divisor <<= Shift;
		DividendHigh <<= Shift;
		DividendHigh |= DividendLow >> (64 - Shift);
		DividendLow <<= Shift;
	}
					// most significant (upper) "digit" of quotient
	Quotient = DividendHigh / _HIGH(Divisor);
	Modulus = DividendHigh % _HIGH(Divisor);

	while ((_HIGH(Quotient) != 0)
	    || (__emulu(_LOW(Quotient), _LOW(Divisor)) > ((Modulus << 32) | _HIGH(DividendLow))))
	{
		Quotient -= 1;
		Modulus += _HIGH(Divisor);

		if (_HIGH(Modulus) != 0)
			break;
	}

	High = _LOW(Quotient);

	DividendMid = ((DividendHigh << 32) | _HIGH(DividendLow))
#if 0	// NOTE: MSC generates call to _aullmul() instead of inline code!
	            - High * Divisor;
#else
	            - __emulu(High, _LOW(Divisor))
	            - (((unsigned long long) (High * _HIGH(Divisor))) << 32);
#endif
					// least significant (lower) "digit" of quotient
	Quotient = DividendMid / _HIGH(Divisor);
	Modulus = DividendMid % _HIGH(Divisor);

	while ((_HIGH(Quotient) != 0)
	    || (__emulu(_LOW(Quotient), _LOW(Divisor)) > ((Modulus << 32) | _LOW(DividendLow))))
	{
		Quotient -= 1;
		Modulus += _HIGH(Divisor);

		if (_HIGH(Modulus) != 0)
			break;
	}

	Low = _LOW(Quotient);

	if (Remainder != NULL)
	{
		Modulus = ((DividendMid << 32) | _LOW(DividendLow))
#if 0	// NOTE: MSC generates call to _aullmul() instead of inline code!
		        - Low * Divisor;
#else
		        - __emulu(Low, _LOW(Divisor))
		        - (((unsigned long long) (Low * _HIGH(Divisor))) << 32);
#endif
		if (Shift != 0)
			Modulus >>= Shift;

		*Remainder = Modulus;
	}

	Quotient = High;
	Quotient <<= 32;
	Quotient |= Low;

	return Quotient;
#endif // Algorithm D
#undef _LOW
#undef _HIGH
}
#endif

long long	_abs64(long long x);

#pragma intrinsic(_abs64)

// companion routine for _mul128() intrinsic
// https://msdn.microsoft.com/en-us/library/82cxdw50.aspx
// https://msdn.microsoft.com/en-us/library/hh802931.aspx
// https://docs.microsoft.com/en-us/cpp/intrinsics/div128

__inline
long long	_div128(  signed long long DividendHigh,
		        unsigned long long DividendLow,
		          signed long long Divisor,
		          signed long long *Remainder)
{
	unsigned long long	low, high, quotient, remainder;
#if 0
	if (DividendHigh < 0)
#else
	if ((long) (DividendHigh >> 32) < 0)
#endif
	{
		high = 0 - DividendHigh;
		low = 0 - DividendLow;

		if (low > DividendLow)
			high -= 1;
	}
	else
	{
		low = DividendLow;
		high = DividendHigh;
	}

	quotient = _udiv128(low, high, _abs64(Divisor), &remainder);

	if (Remainder != NULL)

		// sign of remainder is same as sign of dividend
#if 0
		if (DividendHigh < 0)
#else
		if ((long) (DividendHigh >> 32) < 0)
#endif
			*Remainder = 0 - remainder;
		else
			*Remainder = remainder;

	// sign of quotient is negative if signs of dividend and divisor differ
#if 0
	if ((DividendHigh < 0) != (Divisor < 0))
#elif 0
	if ((DividendHigh >> 63) != (Divisor >> 63))
#elif 0
	if (((long) (DividendHigh >> 32) < 0) != ((long) (Divisor >> 32) < 0))
#else
	if (((long) (DividendHigh >> 32) ^ (long) (Divisor >> 32)) < 0)
#endif
		return 0 - quotient;

	return quotient;
}
#endif // _M_IX86

#ifdef _M_AMD64
// https://docs.microsoft.com/en-us/cpp/intrinsics/div64

long	_div64(long long dividend,
	       long      divisor,
	       long      *remainder);

// https://docs.microsoft.com/en-us/cpp/intrinsics/udiv64

unsigned long	_udiv64(unsigned long long dividend,
		        unsigned long      divisor,
		        unsigned long      *remainder);

// companion routine for _mul128() intrinsic
// https://msdn.microsoft.com/en-us/library/82cxdw50.aspx
// https://msdn.microsoft.com/en-us/library/hh802931.aspx
// https://docs.microsoft.com/en-us/cpp/intrinsics/div128

long long	_div128(signed long long DividendHigh,
		        signed long long DividendLow,
		        signed long long Divisor,
		        signed long long *Remainder);

// companion routine for _umul128() intrinsic
// https://msdn.microsoft.com/en-us/library/3dayytw9.aspx
// https://msdn.microsoft.com/en-us/library/hh802933.aspx
// https://docs.microsoft.com/en-us/cpp/intrinsics/udiv128

unsigned long long	_udiv128(unsigned long long DividendHigh,
			         unsigned long long DividendLow,
			         unsigned long long Divisor,
			         unsigned long long *Remainder);

#if _MSC_VER >= 1920 // MSC 19.20 alias 2019
#pragma intrinsic(_div64, _udiv64, _div128, _udiv128)
#endif
#endif // _M_AMD64

#ifdef _WIN64
// companion routine for _umul128() intrinsic

__inline		// widening carry-less multiplication
unsigned long long	_clmul128(unsigned long long p,
			          unsigned long long q,
			          unsigned long long *r)
{
	unsigned long long	low = 0, high = 0;
	unsigned long		s = 63, t;

	while (_BitScanReverse64(&t, q))
	{
		high = __shiftleft128(low, high, s - t);
		low <<= s - t;
		low ^= p;
		_bittestandreset64(&q, s = t);
	}

	*r = __shiftleft128(low, high, s);

	return low << s;
}
#endif // _WIN64

__inline		// combined multiplication and division: p * q / s
unsigned long long	_umuldiv64(unsigned long long p,
			           unsigned long long q,
			           unsigned long long s)
{
#ifdef _M_AMD64
	unsigned long long	r, high, low = _umul128(p, q, &high);

	return _udiv128(low, high, s, &r);
#else
	unsigned long long	r = 0;
	unsigned long long	t;

	while (p)
	{
		if (p & 1)
		{
			r += q;
			if ((r < q) || (r >= s))
				r -= s;
		}

		t = q;
		q += q;
		if ((q < t) || (q >= s))
			q -= s;

		p >>= 1;
	}

	return q;
#endif // _M_AMD64
}

__inline		// combined multiplication and modulo: p * q % s
unsigned long long	_umulmod64(unsigned long long p,
			           unsigned long long q,
			           unsigned long long s)
{
#ifdef _M_AMD64
	unsigned long long	r, high, low = _umul128(p, q, &high);

	_udiv128(low, high, s, &r);

	return r;
#else
	unsigned long long	r = 0;
	unsigned long long	t;

	while (p)
	{
		if (p & 1)
		{
			r += q;
			if ((r < q) || (r >= s))
				r -= s;
		}

		t = q;
		q += q;
		if ((q < t) || (q >= s))
			q -= s;

		p >>= 1;
	}

	return r;
#endif // _M_AMD64
}

__declspec(deprecated("Don't use; included only for documentation!"))
		// multiplicative linear congruential generator
__inline	// ("minimal standard", Stephen K. Park & Keith W. Miller)
unsigned long	minstd(void)
{
	static	unsigned long	seed = 65537UL;	// = 2**16+1 (Fermat prime F4)
#ifndef _M_IX86
	return seed = ((unsigned long long) seed * 48271UL) % 2147483647UL;
#else
	__asm	// returns pseudo-random number in eax
	{
		mov	eax, 48271	; eax = multiplier
#if 0
		mov	ecx, 2147483647	; ecx = divisor
					;     = 2**31-1 (Mersenne prime M8)
		mul	seed		; edx:eax = seed * 48271
		div	ecx		; eax = (seed * 48271) / 2147483647,
					; edx = (seed * 48271) % 2147483647
					;     = seed'
		mov	seed, edx
		mov	eax, edx	; eax = seed'
#else
		mul	seed		; edx:eax = seed * 48271
		add	eax, eax
		adc	edx, edx
		add	edx, edx
		add	eax, edx
		sbb	edx, edx
		shr	eax, 1
		sub	eax, edx	; eax = seed'
#endif
	}
#endif // _M_IX86
}

__inline	// (mixed) linear congruential generator
unsigned long	mlcg32(void)
{
	static	unsigned long	seed = 0x9E3779B9;
#if 0 // multipliers proposed by Pierre L'Ecuyer
	return seed = seed * 29943829 + 3037000493;
	return seed = seed * 32310901 + 3037000493;
	return seed = seed * 2891336453 + 3037000493;
#else // multiplier proposed by Melissa O'Neill
	return seed = seed * 1019135901 + 3037000493;
#endif
}

__inline		// (mixed) linear congruential generator
unsigned long long	mlcg64(void)
{
	static	unsigned long long	seed = 0x9E3779B97F4A7C15;
#ifndef _M_IX86
#if 0 // multipliers proposed by Pierre L'Ecuyer
	return seed = seed * 2862933555777941757 + 3037000493;
	return seed = seed * 3202034522624059733 + 3037000493;
	return seed = seed * 3935559000370003845 + 3037000493;
#else // multipliers proposed by Melissa O'Neill
	return seed = seed * 9199940308585234877 + 3037000493;
//	return seed = seed * 14647171131086947261 + 3037000493;
#endif
#else
	__asm	// returns pseudo-random number in edx:eax
	{
#if 0
		mov	ecx, dword ptr seed
					; ecx = low dword of seed
		mov	eax, 0A00541BDh	; eax = low dword of 9199940308585234877
		mul	ecx		; edx:eax = low dword of 9199940308585234877
					;         * low dword of seed
		imul	ecx, 07FACC0F7h	; ecx = low dword of seed
					;     * high dword of 9199940308585234877
		add	edx, ecx	; edx:eax = low dword of seed
					;         * 9199940308585234877
		mov	ecx, dword ptr seed+4
					; ecx = high dword of seed
		imul	ecx, 0A00541BDh	; ecx = high dword of seed
					;     * low dword of 9199940308585234877
#else
		mov	eax, 0A00541BDh	; eax = low dword of 9199940308585234877
		mul	dword ptr seed	; edx:eax = low dword of 9199940308585234877
					;         * low dword of seed
		imul	ecx, dword ptr seed, 07FACC0F7h
					; ecx = low dword of seed
					;     * high dword of 9199940308585234877
		add	edx, ecx	; edx:eax = low dword of seed
					;         * 9199940308585234877
		imul	ecx, dword ptr seed+4, 0A00541BDh
					; ecx = high dword of seed
					;     * low dword of 9199940308585234877
#endif
		add	eax, 0B504F32Dh
		adc	edx, ecx	; edx:eax = 9199940308585234877 * seed
					;         + 3037000493
					;         = seed'
		mov	dword ptr seed, eax
		mov	dword ptr seed+4, edx
					; seed = edx:eax
	}
#endif // _M_IX86
}

// https://msdn.microsoft.com/en-us/library/d2s81xt0.aspx

unsigned long long	__emulu(unsigned long multiplicand, unsigned long multiplier);

#pragma intrinsic(__emulu)

__inline	// multiply-with-carry generator
unsigned long	mwcg32(void)
{
	// preconditions for maximum period length:
	// prime" = multiplier * 2**32 - 1 is a "safe" prime and
	// prime' = multiplier * 2**31 - 1 is a Sophie Germain prime, i.e.
	// prime" = prime' * 2 + 1 holds for a multiplier less than 2**32
	//
	// multiplier = 2**32 - 178 = 4294967118

	static	unsigned long		carry = 0xC13FA9A9;
					//    = 2**32 / plastic number
	static	unsigned long		seed = 0x9E3779B9;
					//   = 2**32 / golden ratio
	const	unsigned long long	ull = __emulu(0xFFFFFF4E, seed) + carry;

	carry = ull >> 32;
	seed = (unsigned long) ull;

	return seed;
}

__inline		// multiply-with-carry generator
unsigned long long	mwcg64(void)
{
	// preconditions for maximum period length:
	// prime" = multiplier * 2**64 - 1 is a "safe" prime and
	// prime' = multiplier * 2**63 - 1 is a Sophie Germain prime, i.e.
	// prime" = prime' * 2 + 1 holds for a multiplier less than 2**64
	//
	// multiplier = 2**64 - 742 = 18446744073709550874

	static	unsigned long long	carry = 0xC13FA9A902A6328F;
					//    = 2**64 / plastic number
	static	unsigned long long	seed = 0x9E3779B97F4A7C15;
					//   = 2**64 / golden ratio
#ifndef _M_IX86
	auto	unsigned long long	high;

	seed = _umul128(0xFFFFFFFFFFFFFD1A, seed, &high) + carry;
	high += seed < carry;
	carry = high;

	return seed;
#else
	__asm	// returns pseudo-random number in edx:eax
	{
		mov	ebx, dword ptr seed
					; ebx = low dword of seed
		mov	eax, 0FFFFFD1Ah	; eax = low dword of multiplier
		mul	ebx		; edx:eax = low dword of seed
					;         * low dword of multiplier
		push	eax
		mov	ecx, edx
		mov	eax, 0FFFFFFFFh	; eax = high dword of multiplier
		mul	ebx		; edx:eax = low dword of seed
					;         * high dword of multiplier
		xor	ebx, ebx	; ebx:ecx = initial "inner" product
		add	ecx, eax
		adc	ebx, edx	; ebx:ecx = intermediate "inner" product
		mov	eax, 0FFFFFD1Ah	; eax = low dword of multiplier
		mul	dword ptr seed+4; edx:eax = high dword of seed
					;         * low dword of multiplier
		add	ecx, eax
		adc	ebx, edx	; ebx:ecx = final "inner" product
		push	ecx
		sbb	ecx, ecx	; ecx = 0 - carry from "inner" product
		mov	eax, 0FFFFFFFFh	; eax = high dword of multiplier
		mul	dword ptr seed+4; edx:eax = high dword of seed
					;         * high dword of multiplier
		neg	ecx		; ecx = carry from "inner" product
		add	ebx, eax
		adc	ecx, edx	; ecx:ebx = high dword of seed
					;         * high dword of multiplier
					;         + high dword of "inner" product
					;         + carry from "inner" product
					;         = high qword of product
		pop	edx
		pop	eax		; edx:eax = low qword of product
		add	eax, dword ptr carry
		adc	edx, dword ptr carry+4
					; edx:eax = low qword of (product + carry)
					;         = seed'
		mov	dword ptr seed, eax
		mov	dword ptr seed+4, edx
					; seed = edx:eax
		adc	ebx, 0
		adc	ecx, 0		; ecx:ebx = high qword of (product + carry)
					;         = carry'
		mov	dword ptr carry, ebx
		mov	dword ptr carry+4, ecx
					; carry = ecx:ebx
	}
#endif // _M_IX86
}

// https://msdn.microsoft.com/en-us/library/d2s81xt0.aspx

unsigned long long	__emulu(unsigned long multiplicand, unsigned long multiplier);

__inline	// square-with-carry generator
unsigned long	swcg32(void)
{
	static	unsigned long		carry = 0xC13FA9A9;
					//    = 2**32 / plastic number
	static	unsigned long		seed = 0x9E3779B9;
					//   = 2**32 / golden ratio
	const	unsigned long long	ull = __emulu(seed, seed) + carry;

	carry = ull >> 32;
	seed = (unsigned long) ull;

	return seed;
}

__inline		// square-with-carry generator
unsigned long long	swcg64(void)
{
	static	unsigned long long	carry = 0xC13FA9A902A6328F;
					//    = 2**64 / plastic number
	static	unsigned long long	seed = 0x9E3779B97F4A7C15;
					//   = 2**64 / golden ratio
#ifndef _M_IX86
	auto	unsigned long long	high;

	seed = _umul128(seed, seed, &high) + carry;
	high += seed < carry;
	carry = high;

	return seed;
#else
	__asm	// returns pseudo-random number in edx:eax
	{
		mov	ebx, dword ptr seed
					; ebx = low dword of seed
		mov	eax, ebx	; eax = low dword of seed
		mul	ebx		; edx:eax = (low dword of seed)**2
		push	eax
		mov	ecx, edx
		mov	eax, dword ptr seed+4
					; eax = high dword of seed
		mul	ebx		; edx:eax = high dword of seed
					;         * low dword of seed
		xor	ebx, ebx	; ebx:ecx = initial "inner" product
		add	ecx, eax
		adc	ebx, edx	; ebx:ecx = intermediate "inner" product
		add	ecx, eax
		adc	ebx, edx	; ebx:ecx = final "inner" product
		push	ecx
		sbb	ecx, ecx	; ecx = 0 - carry from "inner" product
		mov	eax, dword ptr seed+4
					; eax = high dword of seed
		mul	eax		; edx:eax = (high dword of seed)**2
		neg	ecx		; ecx = carry from "inner" product
		add	ebx, eax
		adc	ecx, edx	; ecx:ebx = (high dword of seed)**2
					;         + high dword of "inner" product
					;         + carry from "inner" product
					;         = high qword of seed**2
		pop	edx
		pop	eax		; edx:eax = low qword of seed**2
		add	eax, dword ptr carry
		adc	edx, dword ptr carry+4
					; edx:eax = low qword of (seed**2 + carry)
					;         = seed'
		mov	dword ptr seed, eax
		mov	dword ptr seed+4, edx
					; seed = edx:eax
		adc	ebx, 0
		adc	ecx, 0		; ecx:ebx = high qword of (seed**2 + carry)
					;         = carry'
		mov	dword ptr carry, ebx
		mov	dword ptr carry+4, ecx
					; carry = ecx:ebx
	}
#endif // _M_IX86
}

// https://msdn.microsoft.com/en-us/library/d2s81xt0.aspx

unsigned long long	__emulu(unsigned long multiplicand, unsigned long multiplier);

// https://msdn.microsoft.com/en-us/library/a0w705h5.aspx

unsigned long	_lrotr(unsigned long value, int shift);

#pragma intrinsic(__emulu, _lrotr)

__inline	// multiplicative (linear) congruential generator, scrambled
unsigned long	smcg32(void)
{
	static	unsigned long		seed = 0x9E3779B9;
#ifndef _M_IX86
#if 1 // multiplier proposed by Melissa O'Neill
	const	unsigned long long	ull = __emulu(2739110765, seed);
#else // multipliers proposed by Pierre L'Ecuyer
	const	unsigned long long	ull = __emulu(1597334677, seed);
	const	unsigned long long	ull = __emulu(3425435293, seed);
	const	unsigned long long	ull = __emulu(4109213157, seed);
#endif
	const	unsigned long		low = (unsigned long) (ull >> 11);
	const	unsigned long		high = (unsigned long) (ull >> (32 + 11));

	seed = (unsigned long) ull;

	return _lrotr(low ^ (low >> 17), high);
#else
	__asm	// returns pseudo-random number in eax
	{
		mov	eax, 2739110765	; eax = multiplier
		mul	seed		; edx:eax = multiplier * seed
					;         = seed'
		mov	seed, eax
		mov	ecx, edx
		shrd	eax, edx, 11	; eax = seed' >> 11
					;     = value"
		shr	ecx, 11
		mov	edx, eax
		shr	eax, 17
		xor	eax, edx	; eax = value" ^ (value" >> 17)
					;     = value'
		ror	eax, cl		; eax = rotated value'
					;     = value
	}
#endif // _M_IX86
}

// https://msdn.microsoft.com/en-us/library/5cc576c4.aspx

unsigned long long	_rotr64(unsigned long long value, int shift);

#pragma intrinsic(_rotr64)

__inline		// multiplicative (linear) congruential generator, scrambled
unsigned long long	smcg64(void)
{
	static	unsigned long long	seed = 0x9E3779B97F4A7C15;
#ifndef _M_IX86
	auto	unsigned long long	low, high;
#if 0 // multiplier proposed by Richard Simard
	seed = _umul128(8372773778140471301, seed, &high);
#elif 1 // multipliers proposed by Pierre L'Ecuyer
//	seed = _umul128(1181783497276652981, seed, &high);
//	seed = _umul128(1865811235122147685, seed, &high);
//	seed = _umul128(2685821657736338717, seed, &high);
//	seed = _umul128(6415128727920758069, seed, &high);
//	seed = _umul128(7664345821815920749, seed, &high);
	seed = _umul128(13515856136758413469, seed, &high);
#else // multipliers proposed by Melissa O'Neill
	seed = _umul128(12605985483714917081, seed, &high);
	seed = _umul128(14647171131086947261, seed, &high);
#endif
	low = __shiftright128(seed, high, 13);

	return _rotr64(low ^ (low >> 31), ((int) high >> 13) & 31);
#else
	__asm	// returns pseudo-random number in edx:eax
	{
		mov	ecx, dword ptr seed
					; ecx = low dword of seed
		mov	eax, 0DAC4C89Dh	; eax = low dword of 13515856136758413469
		mul	ecx		; edx:eax = low dword of seed
					;         * low dword of 13515856136758413469
		mov	dword ptr seed, eax
		mov	ebx, edx
		mov	eax, 0BB91F78Bh	; eax = high dword of 13515856136758413469
		mul	ecx		; edx:eax = low dword of seed
					;         * high dword of 13515856136758413469
		xor	ecx, ecx	; ecx:ebx = initial "inner" product
		add	ebx, eax
		adc	ecx, edx	; ecx:ebx = first intermediate "inner" product
		mov	eax, 0DAC4C89Dh	; eax = low dword of 13515856136758413469
		mul	dword ptr seed+4; edx:eax = high dword of seed
					;         * low dword of 13515856136758413469
		add	ebx, eax
		adc	ecx, edx	; ecx:ebx = second intermediate "inner" product
		imul	eax, dword ptr seed+4, 0BB91F78Bh
					; eax = high dword of seed
					;     * high dword of 13515856136758413469
		add	ecx, eax	; ecx:ebx = final "inner" product
		mov	eax, dword ptr seed
					; eax = low dword of seed'
		mov	dword ptr seed+4, ebx
		shrd	eax, ebx, 13
		shrd	ebx, ecx, 13	; ebx:eax = seed' >> 13
		shr	ecx, 13		;         = value"
		mov	edx, eax
		shrd	eax, ebx, 31
		xor	eax, edx
		mov	edx, ebx
		shr	ebx, 31
		xor	edx, ebx	; edx:eax = value" ^ (value" >> 31)
					;         = value'
		mov	ebx, eax
		shrd	eax, edx, cl
		shrd	edx, ebx, cl	; edx:eax = rotated value'
					;         = value
	}
#endif // _M_IX86
}

// https://msdn.microsoft.com/en-us/library/d2s81xt0.aspx

unsigned long long	__emulu(unsigned long multiplicand, unsigned long multiplier);

// https://msdn.microsoft.com/en-us/library/wds0z27f.aspx

unsigned long long	__ull_rshift(unsigned long long value, unsigned long count);

// https://msdn.microsoft.com/en-us/library/a3140177.aspx

unsigned long		_byteswap_ulong(unsigned long ul);

#pragma intrinsic(__emulu, __ull_rshift, _byteswap_ulong)

__inline	// enhanced middle-square pseudo-random number generator
unsigned long	msqr32(void)
{
	static	unsigned long	seed = 0;
	static	unsigned long	weyl = 0;
#ifndef _M_IX86
	weyl += 0x9E3779B9;			// advance Weyl sequence:
						//  add 2**32 / golden ratio
	seed = (unsigned long) __ull_rshift(__emulu(seed, seed), 16);
	seed += weyl;
	seed = _byteswap_ulong(seed);

	return seed;
#else
	__asm	// returns pseudo-random number in eax
	{
		mov	eax, seed	; eax = seed
		mul	eax		; edx:eax = seed**2
#if 1
		shr	eax, 16
		shl	edx, 16
		or	eax, edx	; eax = (seed**2 >> 16) % 2**32
					;     = seed'
#else
		shrd	eax, edx, 16	; eax = (seed**2 >> 16) % 2**32
					;     = seed'
#endif
		mov	edx, weyl	; edx = weyl
		add	edx, 09E3779B9h	; edx = weyl + 2**32 / golden ratio
					;     = weyl'
		mov	weyl, edx	; weyl = edx
		add	eax, edx	; eax = seed' + weyl'
					;     = seed"
		bswap	eax
		mov	seed, eax	; seed = eax
	}
#endif // _M_IX86
}

// https://msdn.microsoft.com/en-us/library/a3140177.aspx

unsigned long long	_byteswap_uint64(unsigned long long ull);

#pragma intrinsic(__emulu, _byteswap_uint64)

__inline		// enhanced middle-square pseudo-random number generator
unsigned long long	msqr64(void)
{
	static	unsigned long long	seed = 0;
	static	unsigned long long	weyl = 0;
#ifdef _WIN64
	auto	unsigned long long	high,
					// square current value, extract middle
	low = _umul128(seed, seed, &high);
	seed = __shiftright128(low, high, 32);
					// advance Weyl sequence:
	weyl += 0x9E3779B97F4A7C15;	//  add 2**64 / golden ratio
	seed += weyl;
	seed = _byteswap_uint64(seed);

	return seed;
#else
#ifndef _M_IX86
					// advance Weyl sequence:
	weyl += 0x9E3779B97F4A7C15;	//  add 2**64 / golden ratio
	seed = weyl
	     + (__emulu((unsigned long) seed, (unsigned long) seed) >> 32)
	     + (__emulu((unsigned long) seed, (unsigned long) (seed >> 32)) << 1)
	     + (__emulu((unsigned long) (seed >> 32), (unsigned long) (seed >> 32)) << 32);
	seed = _byteswap_uint64(seed);

	return seed;
#else
	__asm	// returns pseudo-random number in edx:eax
	{
		mov	ecx, dword ptr seed
					; ecx = low dword of seed
		mov	eax, ecx
		mul	eax		; edx:eax = low dword of seed
					;         * low dword of seed
		mov	ebx, edx	; ebx = high dword of (low dword of seed)**2
		mov	eax, ecx	; eax = low dword of seed
		mov	ecx, dword ptr seed+4
					; ecx = high dword of seed
		mul	ecx		; edx:eax = low dword of seed
					;         * high dword of seed
		imul	ecx, ecx	; ecx = low dword of (high dword of seed)**2
					;     = low dword of (high qword of seed**2)
		add	eax, eax
		adc	edx, edx	; edx:eax = low dword of seed
					;         * high dword of seed
					;         * 2
		add	ebx, eax
		adc	ecx, edx	; ecx:ebx = (seed**2 >> 32) % 2**64
					;         = seed'
		mov	edx, dword ptr weyl
		mov	eax, dword ptr weyl+4
					; eax:edx = weyl
		add	edx, 07F4A7C15h
		adc	eax, 09E3779B9h	; eax:edx = weyl + 2**64 / golden ratio
					;         = weyl'
		mov	dword ptr weyl, edx
		mov	dword ptr weyl+4, eax
					; weyl = eax:edx
		add	edx, ebx
		adc	eax, ecx	; eax:edx = weyl' + seed'
					;         = seed"
		bswap	edx
		bswap	eax		; edx:eax = byte-swapped seed"

		mov	dword ptr seed+4, edx
		mov	dword ptr seed, eax
					; seed = edx:eax
	}
#endif // _M_IX86
#endif // _WIN64
}

__inline	// hashed modular arithmetic progression deterministic random bit generator
unsigned long	drbg32()
{
	static	unsigned long		seed = 0;
	auto	unsigned long long	ull;

	seed += 0x927B458DUL;		// prime numbers with equal count of '0' and '1' bits,
					//  plus their most and least significant bits set
	ull = __emulu(0xAD4C2177, seed);
	ull = __emulu(0xBA9A4A53, (unsigned long) ull ^ (unsigned long) (ull >> 32));

	return (unsigned long) ull ^ (unsigned long) (ull >> 32);
}

__inline		// hashed modular arithmetic progression deterministic random bit generator
unsigned long long	drbg64()
{
	static	unsigned long long	seed = 0;
	auto	unsigned long long	low, high;

	seed += 0x9365D8C2239F41BD;	// prime numbers with equal count of '0' and '1' bits,
					//  plus their most and least significant bits set
	low = _umul128(0xA418D9B62FAB4563, seed, &high);
	low = _umul128(0xBCAE53A3978642E1, low ^ high, &high);

	return low ^ high;
}

__inline	// linear feedback shift register (Galois form)
unsigned long	cntr32(void)
{
	// primitive polynomials for maximum period length 2**32-1,
	//  grouped by number of coefficients:
	//
	// 0x000000C5, 0x46000001 = x**32 + x**7  + x**6  + x**2 + x**0
	// 0x00010085, 0x42010001 = x**32 + x**16 + x**7  + x**2 + x**0
	// 0x00400007, 0xC0000401 = x**32 + x**22 + x**2  + x**1 + x**0
	// 0x18000003, 0x80000031 = x**32 + x**28 + x**27 + x**1 + x**0
	//
	// 0x000000AF, 0xEA000001 = x**32 + x**7 + x**5 + x**3 + x**2 + x**1 + x**0
	// 0x000000F5, 0x5E000001 = x**32 + x**7 + x**6 + x**5 + x**4 + x**2 + x**0
	//
	// 0x4002100B, 0xA0108005 = x**32 + x**30 + x**17 + x**12 + x**3 + x**1 + x**0
	//
	// 0x04C11DB7, 0xDB710641 (CRC-32 IEEE)
	//
	// 0x3FF0C3FF, 0xFF861FF9

	static	unsigned long	lfsr = ~0UL;
#ifndef _M_IX86
#if 0
	return lfsr = (lfsr << 1) ^ ((0 - ((long) lfsr < 0)) & 0xDB710641);
#else
	return lfsr = (lfsr << 1) ^ (((long) lfsr >> 31) & 0xDB710641);
#endif
#else
	__asm	// returns result in eax
	{
		mov	eax, lfsr	; eax = lfsr
#if 0
		add	eax, eax	; eax = lfsr << 1
					;     = lfsr'
		sbb	edx, edx	; edx = (lfsr < 0) ? -1 : 0
#else
		cdq			; edx = (lfsr < 0) ? -1 : 0
		add	eax, eax	; eax = lfsr << 1
					;     = lfsr'
#endif
		and	edx, 0DB710641h	; edx = (lfsr < 0) ? polynomial : 0
					;     = polynomial'
		xor	eax, edx	; eax = lfsr' ^ polynomial'
					;     = lfsr"
		mov	lfsr, eax	; lfsr = eax
	}
#endif // _M_IX86
}

__inline		// linear feedback shift register (Galois form)
unsigned long long	cntr64(void)
{
	// primitive polynomials for maximum period length 2**64-1:
	//
	// 0x000000000000001B, 0xB000000000000001 (CRC-64 ISO-3309)
	// 0x000000000000001D, 0x7000000000000001
	// 0x00000000000000F5, 0x5E00000000000001
	// 0x2B5926535897936B, 0xAD93D23594C935A9 (CRC-64 "Jones")
	//
	// 0x000000005141A33D, 0x798B051400000001
	// 0x00000000FE190681, 0x02C130FE00000001
	//
	// 0x06A2F03CD73BA813, 0x902BB9D6781E8AC1
	// 0x0B9BA8DF12777545, 0x455DDC91F62BB3A1
	// 0x101E46D04EE6E989, 0x232ECEE416C4F011
	// 0x13B5066C86DC2BB1, 0x1BA876C26CC15B91
	// 0x2EC23767AAE89BF5, 0x5FB22EABCDD886E9
	// 0x3D1945EABD0ACCDD, 0x7666A17AAF453179
	// 0x5043443F1755DF4D, 0x65F755D1F8458415
	// 0x54E6B179A3691FF5, 0x5FF12D8B3D1ACE55
	// 0x65A3653F84B1296D, 0x6D291A43F94D8B4D
	// 0x681B30E94C845ACB, 0xA6B442652E19B02D
	// 0x830F7DB354117363, 0x8D9D10559B7DE183

	static	unsigned long long	lfsr = ~0ULL;
#ifndef _M_IX86
#if 0
	return lfsr = (lfsr << 1) ^ ((0 - ((long long) lfsr < 0)) & 0xAD93D23594C935A9);
#else
	return lfsr = (lfsr << 1) ^ (((long long) lfsr >> 63) & 0xAD93D23594C935A9);
#endif
#else
	__asm	// returns pseudo-random number in edx:eax
	{
#if 0
		mov	eax, dword ptr lfsr
		mov	edx, dword ptr lfsr+4
					; edx:eax = lfsr
		add	eax, eax
		adc	edx, edx	; edx:eax = lfsr << 1
					;         = lfsr'
		sbb	ecx, ecx	; ecx = (lfsr < 0) ? -1 : 0
		and	ecx, 0FE190681h	; ecx = (lfsr < 0) ? polynomial : 0
					;     = polynomial'
		xor	eax, ecx	; edx:eax = lfsr' ^ polynomial'
					;         = lfsr"
		mov	dword ptr lfsr+4, edx
		mov	dword ptr lfsr, eax
					; lfsr = edx:eax
#else
		mov	eax, dword ptr lfsr+4
		mov	ecx, dword ptr lfsr
		mov	ebx, eax	; ebx:ecx = lfsr
		cdq			; edx = (lfsr < 0) ? -1 : 0
		add	ecx, ecx
		adc	ebx, ebx	; ebx:ecx = lfsr << 1
					;         = lfsr'
		mov	eax, edx	; edx:eax = (lfsr < 0) ? -1 : 0
		and	edx, 0AD93D235h
		and	eax, 094C935A9h	; edx:eax = (lfsr < 0) ? polynomial : 0
					;         = polynomial'
		xor	edx, ebx
		xor	eax, ecx	; edx:eax = polynomial' ^ lfsr'
					;         = lfsr"
		mov	dword ptr lfsr+4, edx
		mov	dword ptr lfsr, eax
					; lfsr = edx:eax
#endif
	}
#endif // _M_IX86
}

// https://msdn.microsoft.com/en-us/library/a0w705h5.aspx

unsigned long	_lrotr(unsigned long value, int shift);

// https://msdn.microsoft.com/en-us/library/t5e2f3sc.aspx

unsigned char	_rotl8(unsigned char value, unsigned char shift);

#pragma intrinsic(_lrotr, _rotl8)

__inline	// linear feedback shift register (Fibonacci form)
unsigned long	lfsr32(void)
{
	// primitive polynomials for maximum period length 2**32-1:
	//
	// 0x000000C5 = x**32 + x**7 + x**6 + x**2 + x**0
	// 0x000000AF = x**32 + x**7 + x**5 + x**3 + x**2 + x**1 + x**0
	// 0x000000F5 = x**32 + x**7 + x**6 + x**5 + x**4 + x**2 + x**0

	static	unsigned long	lfsr = ~0UL;
#ifndef _M_IX86
#if 0
	const	unsigned long	tmp = lfsr & 0xC5UL;
	const	unsigned long	msb = 0x69966996UL >> (tmp ^ (tmp >> 4));
#else
	const	unsigned long	msb = 0x96696996UL >> _rotl8((unsigned char) lfsr & 0xC5U, 2);
#endif
	return lfsr = _lrotr((lfsr & ~1UL) | (msb & 1UL), 1);
#else
	__asm	// returns pseudo-random number in eax
	{
		mov	eax, lfsr	; eax = lfsr
		test	al, 0C5h
		setpo	cl		; cl = parity(lfsr & polynomial) ? 0 : 1
					;    = most significant bit of lfsr'
#if 0
		shr	cl, 1
		rcr	eax, 1		; eax = lfsr'
#else
		shrd	eax, ecx, 1	; eax = lfsr'
#endif
		mov	lfsr, eax	; lfsr = eax
	}
#endif // _M_IX86
}

// https://msdn.microsoft.com/en-us/library/5cc576c4.aspx

unsigned long long	_rotr64(unsigned long long value, int shift);

#pragma intrinsic(_rotr64)

__inline		// linear feedback shift register (Fibonacci form)
unsigned long long	lfsr64(void)
{
	// primitive polynomials for maximum period length 2**64-1:
	//
	// 0x000000000000001B = x**64 + x**4 + x**3 + x**1 + x**0
	// 0x000000000000001D = x**64 + x**4 + x**3 + x**2 + x**0
	// 0x00000000000000F5 = x**64 + x**7 + x**6 + x**5 + x**4 + x**2 + x**0

	static	unsigned long long	lfsr = ~0ULL;
#ifndef _M_IX86
#ifndef _WIN64
	const	unsigned		msb = 0x96696996U >> ((unsigned) lfsr & 0x1DU);

	return lfsr = _rotr64((lfsr & ~1ULL) | (msb & 1ULL), 1);
#else
	return lfsr = __shiftright128(lfsr, 0x96696996U >> ((unsigned) lfsr & 0x1DU), 1);
#endif
#else
	__asm	// returns pseudo-random number in edx:eax
	{
		mov	eax, dword ptr lfsr
		mov	edx, dword ptr lfsr+4
					; edx:eax = lfsr
		test	al, 01Dh
		setpo	cl		; cl = parity(lfsr & polynomial) ? 0 : 1
					;    = most significant bit of lfsr'
		shrd	eax, edx, 1
		shrd	edx, ecx, 1	; edx:eax = lfsr'
		mov	dword ptr lfsr, eax
		mov	dword ptr lfsr+4, edx
					; lfsr = edx:eax
	}
#endif // _M_IX86
}

__inline	// linear feedback shift register (XorShift form, 32-bit)
unsigned long	full32(void)
{
	static	unsigned long	lfsr = ~0UL;
#if 1 // shift constants from George Marsaglia
	lfsr ^= lfsr << 7;
	lfsr ^= lfsr >> 13;
	lfsr ^= lfsr << 25;
#else // shift constants from Richard Peirce Brent
	lfsr ^= lfsr << 10;
	lfsr ^= lfsr >> 15;
	lfsr ^= lfsr << 4;
	lfsr ^= lfsr >> 13;
#endif
	return lfsr;
}

#ifdef MMX
#include <mmintrin.h>

__inline	// convert to scalar quadword
__int64		_mm_cvtsi64(__m64 _M)
{
	return _M.m64_i64;
}

__inline	// negate (arithmetic/two's complement) packed bytes
__m64		_mm_neg_pi8(__m64 _M)
{
	return _mm_sub_pi8(_mm_setzero_si64(), _M);
}

__inline	// negate (arithmetic/two's complement) packed words
__m64		_mm_neg_pi16(__m64 _M)
{
	return _mm_sub_pi16(_mm_setzero_si64(), _M);
}

__inline	// negate (arithmetic/two's complement) packed doublewords
__m64		_mm_neg_pi32(__m64 _M)
{
	return _mm_sub_pi32(_mm_setzero_si64(), _M);
}

__inline	// not (logical/one's complement) scalar quadword
__m64		_mm_not_si64(__m64 _M)
{
#if 0
	static	const	__m64	_N = {~0ULL};
#if 0
	return _mm_andnot_si64(_M, _N);
#else
	return _mm_xor_si64(_M, _N);
#endif
#else
#if 0
	return _mm_andnot_si64(_M, _mm_cmpeq_pi32(_M, _M));
#else
	return _mm_xor_si64(_M, _mm_cmpeq_pi32(_M, _M));
#endif
#endif
}

__inline	// rotate packed words left by immediate
__m64		_mm_rotli_pi16(__m64 _M, int _Count)
{
	return _mm_or_si64(_mm_slli_pi16(_M, _Count), _mm_srli_pi16(_M, 16 - _Count));
}

__inline	// rotate packed doublewords left by immediate
__m64		_mm_rotli_pi32(__m64 _M, int _Count)
{
	return _mm_or_si64(_mm_slli_pi32(_M, _Count), _mm_srli_pi32(_M, 32 - _Count));
}

__inline	// rotate scalar quadword left by immediate
__m64		_mm_rotli_si64(__m64 _M, int _Count)
{
	return _mm_or_si64(_mm_slli_si64(_M, _Count), _mm_srli_si64(_M, 64 - _Count));
}

__inline	// rotate packed words right by immediate
__m64		_mm_rotri_pi16(__m64 _M, int _Count)
{
	return _mm_or_si64(_mm_srli_pi16(_M, _Count), _mm_slli_pi16(_M, 16 - _Count));
}

__inline	// rotate packed doublewords right by immediate
__m64		_mm_rotri_pi32(__m64 _M, int _Count)
{
	return _mm_or_si64(_mm_srli_pi32(_M, _Count), _mm_slli_pi32(_M, 32 - _Count));
}

__inline	// rotate scalar quadword right by immediate
__m64		_mm_rotri_si64(__m64 _M, int _Count)
{
	return _mm_or_si64(_mm_srli_si64(_M, _Count), _mm_slli_si64(_M, 64 - _Count));
}

__inline	// set scalar quadword
__m64		_mm_set_si64(unsigned __int64 _I)
{
#if 0
	const	__m64	_M = {_I};

	return _M;
#else
	return *((__m64 *) &_I);
#endif
}

__inline	// set packed bytes to 1
__m64		_mm_setone_pi8()
{
	const	__m64	_N = _mm_setzero_si64();

	return _mm_sub_pi8(_N, _mm_cmpeq_pi8(_N, _N));
}

__inline	// set packed words to 1
__m64		_mm_setone_pi16()
{
	const	__m64	_N = _mm_setzero_si64();

	return _mm_sub_pi16(_N, _mm_cmpeq_pi16(_N, _N));
}

__inline	// set packed doublewords to 1
__m64		_mm_setone_pi32()
{
	const	__m64	_N = _mm_setzero_si64();

	return _mm_sub_pi32(_N, _mm_cmpeq_pi32(_N, _N));
}

#include <emmintrin.h>

__inline	// negate (arithmetic/two's complement) scalar quadword
__m64		_mm_neg_si64(__m64 _M)
{
	return _mm_sub_si64(_mm_setzero_si64(), _M);
}

__inline	// rotate packed words left
__m64		_mm_rotl_pi16(__m64 _M, __m64 _Count)
{
#if 0
	static	const	__m64	_N = {16ULL};

	return _mm_or_si64(_mm_sll_pi16(_M, _Count), _mm_srl_pi16(_M, _mm_sub_si64(_N, _Count)));
#else
	return _mm_or_si64(_mm_sll_pi16(_M, _Count), _mm_srl_pi16(_M, _mm_sub_si64(_mm_cvtsi32_si64(16L), _Count)));
#endif
}

__inline	// rotate packed doublewords left
__m64		_mm_rotl_pi32(__m64 _M, __m64 _Count)
{
#if 0
	static	const	__m64	_N = {32ULL};

	return _mm_or_si64(_mm_sll_pi32(_M, _Count), _mm_srl_pi32(_M, _mm_sub_si64(_N, _Count)));
#else
	return _mm_or_si64(_mm_sll_pi32(_M, _Count), _mm_srl_pi32(_M, _mm_sub_si64(_mm_cvtsi32_si64(32L), _Count)));
#endif
}

__inline	// rotate scalar quadword left
__m64		_mm_rotl_si64(__m64 _M, __m64 _Count)
{
#if 0
	static	const	__m64	_N = {64ULL};

	return _mm_or_si64(_mm_sll_si64(_M, _Count), _mm_srl_si64(_M, _mm_sub_si64(_N, _Count)));
#else
	return _mm_or_si64(_mm_sll_si64(_M, _Count), _mm_srl_si64(_M, _mm_sub_si64(_mm_cvtsi32_si64(64L), _Count)));
#endif
}

__inline	// rotate packed words right
__m64		_mm_rotr_pi16(__m64 _M, __m64 _Count)
{
#if 0
	static	const	__m64	_N = {16ULL};

	return _mm_or_si64(_mm_srl_pi16(_M, _Count), _mm_sll_pi16(_M, _mm_sub_si64(_N, _Count)));
#else
	return _mm_or_si64(_mm_srl_pi16(_M, _Count), _mm_sll_pi16(_M, _mm_sub_si64(_mm_cvtsi32_si64(16L), _Count)));
#endif
}

__inline	// rotate packed doublewords right
__m64		_mm_rotr_pi32(__m64 _M, __m64 _Count)
{
#if 0
	static	const	__m64	_N = {32ULL};

	return _mm_or_si64(_mm_srl_pi32(_M, _Count), _mm_sll_pi32(_M, _mm_sub_si64(_N, _Count)));
#else
	return _mm_or_si64(_mm_srl_pi32(_M, _Count), _mm_sll_pi32(_M, _mm_sub_si64(_mm_cvtsi32_si64(32L), _Count)));
#endif
}

__inline	// rotate scalar quadword right
__m64		_mm_rotr_si64(__m64 _M, __m64 _Count)
{
#if 0
	static	const	__m64	_N = {64ULL};

	return _mm_or_si64(_mm_srl_si64(_M, _Count), _mm_sll_si64(_M, _mm_sub_si64(_N, _Count)));
#else
	return _mm_or_si64(_mm_srl_si64(_M, _Count), _mm_sll_si64(_M, _mm_sub_si64(_mm_cvtsi32_si64(64L), _Count)));
#endif
}

__inline	// set scalar quadword to 1
__m64		_mm_setone_si64()
{
	const	__m64	_N = _mm_setzero_si64();

	return _mm_sub_si64(_N, _mm_cmpeq_pi32(_N, _N));
}
#endif // MMX

__inline		// linear feedback shift register (XorShift form, 64-bit)
unsigned long long	full64(void)
{
#ifndef MMX
	static	unsigned long long	lfsr = ~0ULL;
#if 1 // shift constants from George Marsaglia
	lfsr ^= lfsr << 11;
	lfsr ^= lfsr >> 31;
	lfsr ^= lfsr << 18;
#else // shift constants from Richard Peirce Brent
	lfsr ^= lfsr << 10;
	lfsr ^= lfsr >> 15;
	lfsr ^= lfsr << 4;
	lfsr ^= lfsr >> 13;
#endif
	return lfsr;
#else
	static	__m64	lfsr = {~0ULL};
#if 1 // shift constants from George Marsaglia
	lfsr = _mm_xor_si64(lfsr, _mm_slli_si64(lfsr, 11));
	lfsr = _mm_xor_si64(lfsr, _mm_srli_si64(lfsr, 31));
	lfsr = _mm_xor_si64(lfsr, _mm_slli_si64(lfsr, 18));
#else // shift constants from Richard Peirce Brent
	lfsr = _mm_xor_si64(lfsr, _mm_slli_si64(lfsr, 10));
	lfsr = _mm_xor_si64(lfsr, _mm_srli_si64(lfsr, 15));
	lfsr = _mm_xor_si64(lfsr, _mm_slli_si64(lfsr, 4));
	lfsr = _mm_xor_si64(lfsr, _mm_srli_si64(lfsr, 13));
#endif
	return lfsr.m64_u64;
#endif // MMX
}

__inline	// linear feedback shift register (XorShift form, 64-bit)
unsigned long	half32(void)
{
	static	unsigned long	low = ~0UL, high = ~0UL;
	auto	unsigned long	tmp = high;

	high = low;
#if 1 // shift constants from George Marsaglia
	tmp ^= tmp << 2; //  8, 10, 23
	low ^= low >> 3; // 22, 10, 24
	tmp ^= tmp >> 7; //  9, 13,  3
#else // shift constants from Richard Peirce Brent
	low ^= low << 17;
	tmp ^= tmp << 12;
	low ^= low >> 14;
	tmp ^= tmp >> 19;
#endif
	low ^= tmp;

	return low;
}

__inline		// linear feedback shift register (XorShift form, 128-bit)
unsigned long long	half64(void)
{
#ifndef MMX
	static	unsigned long long	low = ~0ULL, high = ~0ULL;
	auto	unsigned long long	tmp = high;

	high = low;
#if 0 // shift constants from Richard Peirce Brent
	low ^= low << 33;
	tmp ^= tmp << 28;
	low ^= low >> 31;
	tmp ^= tmp >> 29;
#else // shift constants from Sebastiano Vigna
	tmp ^= tmp << 23;
#if 1
	low ^= low >> 26;
	tmp ^= tmp >> 17;
#else
	low ^= low >> 5;
	tmp ^= tmp >> 18;
#endif
#endif
	low ^= tmp;

	return low;
#else
	static	__m64	low = {~0ULL}, high = {~0ULL};
	auto	__m64	tmp = high;

	high = low;
#if 1 // shift constants from Sebastiano Vigna
	tmp = _mm_xor_si64(tmp, _mm_slli_si64(tmp, 23));
#if 1
	low = _mm_xor_si64(low, _mm_srli_si64(low, 26));
	tmp = _mm_xor_si64(tmp, _mm_srli_si64(tmp, 17));
#else
	low = _mm_xor_si64(low, _mm_srli_si64(low, 5));
	tmp = _mm_xor_si64(tmp, _mm_srli_si64(tmp, 18));
#endif
#else // shift constants from Richard Peirce Brent
	low = _mm_xor_si64(low, _mm_slli_si64(low, 33));
	tmp = _mm_xor_si64(tmp, _mm_slli_si64(tmp, 28));
	low = _mm_xor_si64(low, _mm_srli_si64(low, 31));
	tmp = _mm_xor_si64(tmp, _mm_srli_si64(tmp, 29));
#endif
	low = _mm_xor_si64(low, tmp);

	return low.m64_u64;
#endif // MMX
}

__inline		// linear feedback shift register (XorShift form, 128-bit)
unsigned long long	full128(void)
{
#ifndef MMX
	static	unsigned long long	low = ~0ULL, high = ~0ULL;
#if 1 // shift constants from Melissa O'Neill
	high ^= __shiftleft128(low, high, 26);
	low ^= low << 26;
	low ^= __shiftright128(low, high, 61);
	high ^= high >> 61;
	high ^= __shiftleft128(low, high, 37);
	low ^= low << 37;
#else
	high ^= __shiftleft128(low, high, 11);
	low ^= low << 11;
	low ^= __shiftright128(low, high, 61);
	high ^= high >> 61;
	high ^= __shiftleft128(low, high, 45);
	low ^= low << 45;
#endif
	return low;
#else
	static	__m64	low = {~0ULL}, high = {~0ULL};
#if 1 // shift constants from Melissa O'Neill
	high = _mm_xor_si64(high, _mm_or_si64(_mm_slli_si64(high, 26), _mm_srli_si64(low, 64 - 26)));
	low = _mm_xor_si64(low, _mm_slli_si64(low, 26));
	low = _mm_xor_si64(low, _mm_or_si64(_mm_srli_si64(low, 61), _mm_slli_si64(high, 64 - 61)));
	high = _mm_xor_si64(high, _mm_srli_si64(high, 61));
	high = _mm_xor_si64(high, _mm_or_si64(_mm_slli_si64(high, 37), _mm_srli_si64(low, 64 - 37)));
	low = _mm_xor_si64(low, _mm_slli_si64(low, 37));
#else
	high = _mm_xor_si64(high, _mm_or_si64(_mm_slli_si64(high, 11), _mm_srli_si64(low, 64 - 11)));
	low = _mm_xor_si64(low, _mm_slli_si64(low, 11));
	low = _mm_xor_si64(low, _mm_or_si64(_mm_srli_si64(low, 61), _mm_slli_si64(high, 64 - 61)));
	high = _mm_xor_si64(high, _mm_srli_si64(high, 61));
	high = _mm_xor_si64(high, _mm_or_si64(_mm_slli_si64(high, 45), _mm_srli_si64(low, 64 - 45)));
	low = _mm_xor_si64(low, _mm_slli_si64(low, 45));
#endif
	return low.m64_u64;
#endif // MMX
}

__inline		// linear feedback shift register (Galois form, 128-bit)
unsigned long long	left128(void)
{
	// primitive polynomial for maximum period length 2**128-1:
	//
	// 0x5DB2B62B0C5F8E1B:D8CCE715FCB2726D
#ifndef MMX
	static	unsigned long long	low = ~0ULL, high = ~0ULL;
#if 0
	const	unsigned long long	mask = 0 - ((long long) high < 0);
#else
	const	unsigned long long	mask = (long long) high >> 63;
#endif
	high = __shiftleft128(low, high, 1) ^ (mask & 0x5DB2B62B0C5F8E1B);
	low = (low << 1) ^ (mask & 0xD8CCE715FCB2726D);

	return low;
#else
	static		__m64	low = {~0ULL}, high = {~0ULL};
	static	const	__m64	LOW = {0xD8CCE715FCB2726DULL};
	static	const	__m64	HIGH = {0x5DB2B62B0C5F8E1BULL};
	auto	const	__m64	mask = _mm_unpackhi_pi32(_mm_srai_pi32(high, 31), _mm_srai_pi32(high, 31));

	high = _mm_or_si64(_mm_slli_si64(high, 1), _mm_srli_si64(low, 63));
	low = _mm_slli_si64(low, 1);
	high = _mm_xor_si64(high, _mm_and_si64(HIGH, mask));
	low = _mm_xor_si64(low, _mm_and_si64(LOW, mask));

	return low.m64_u64;
#endif // MMX
}

__inline		// linear feedback shift register (Galois form, 128-bit)
unsigned long long	right128(void)
{
	// primitive polynomial for maximum period length 2**128-1:
	//
	// 0xB64E4D3FA8E7331B:D871FA30D46D4DBA
#ifndef MMX
	static	unsigned long long	low = ~0ULL, high = ~0ULL;
	const	unsigned long long	mask = 0 - (low & 1);

	low = __shiftright128(low, high, 1) ^ (mask & 0xD871FA30D46D4DBA);
	high = (high >> 1) ^ (mask & 0xB64E4D3FA8E7331B);

	return low;
#else
	static		__m64	low = {~0ULL}, high = {~0ULL};
	static	const	__m64	LOW = {0xD871FA30D46D4DBAULL};
	static	const	__m64	HIGH = {0xB64E4D3FA8E7331BULL};
#if 0
	static	const	__m64	MASK = {1ULL};
	auto	const	__m64	mask = _mm_sub_si64(_mm_setzero_si64(), _mm_and_si64(MASK, low));
#elif 0
	auto	const	__m64	mask = _mm_sub_si64(_mm_setzero_si64(), _mm_and_si64(_mm_sub_si64(_mm_setzero_si64(), _mm_cmpeq_pi32(low, low)), low));
#else
	auto	const	__m64	mask = _mm_unpacklo_pi32(_mm_srai_pi32(_mm_slli_pi32(low, 31), 31), _mm_srai_pi32(_mm_slli_pi32(low, 31), 31));
#endif
	low = _mm_or_si64(_mm_srli_si64(low, 1), _mm_slli_si64(high, 63));
	high = _mm_srli_si64(high, 1);
	low = _mm_xor_si64(low, _mm_and_si64(LOW, mask));
	high = _mm_xor_si64(high, _mm_and_si64(HIGH, mask));

	return low.m64_u64;
#endif // MMX
}

// https://msdn.microsoft.com/en-us/library/t5e2f3sc.aspx

unsigned char	_rotl8(unsigned char value, unsigned char shift);

#pragma intrinsic(_rotl8)

__inline		// linear feedback shift register (Fibonacci form, 128-bit)
unsigned long long	lfsr128(void)
{
	// primitive polynomial for maximum period length 2**128-1:
	//
	// 0x0000000000000000:0000000000000087 = x**128 + x**7 + x**2 + x**1 + x**0
#ifndef MMX
	static	unsigned long long	low = ~0ULL, high = ~0ULL;
#if 0
	const	unsigned		tmp = (unsigned) low & 0x87U;
	const	unsigned		msb = 0x69966996U >> (tmp ^ (tmp >> 4));
#else
	const	unsigned		msb = 0x96696996U >> _rotl8((unsigned char) low & 0x87U, 1);
#endif
	low >>= 1;
	low |= high << 63;
	high >>= 1;
	high |= (unsigned long long) msb << 63;

	return low;
#else
	static	__m64	low = {~0ULL}, high = {~0ULL};
	const	__m64	tmp = _mm_and_si64(_mm_cvtsi32_si64(0x87U), low);
	const	__m64	msb = _mm_srl_si64(_mm_cvtsi32_si64(0x69966996U), _mm_and_si64(_mm_cvtsi32_si64(63U), _mm_xor_si64(tmp, _mm_srli_si64(tmp, 4))));

	low = _mm_or_si64(_mm_srli_si64(low, 1), _mm_slli_si64(high, 63));
	high = _mm_or_si64(_mm_srli_si64(high, 1), _mm_slli_si64(msb, 63));

	return low.m64_u64;
#endif // MMX
}

// https://msdn.microsoft.com/en-us/library/a0w705h5.aspx

unsigned long	_lrotl(unsigned long value, int shift);
unsigned long	_lrotr(unsigned long value, int shift);

#pragma intrinsic(_lrotl, _lrotr)

__inline	// small fast pseudo-random number generator
unsigned long	prng32(void)	//  (Robert John "Bob" Jenkins)
{
#if 1
	static	unsigned long	p = 0x9E3779B9UL;
#else
	static	unsigned long	p = 0xF1EA5EEDUL;
#endif
	static	unsigned long	q = ~0UL;
	static	unsigned long	r = ~0UL;
	static	unsigned long	s = ~0UL;
	const	unsigned long
	t = p - _lrotr(q, 9);
#if 0
	p = q ^ _lrotr(r, 17);
	q = r + _lrotr(s, 21);
#else
	p = q ^ _lrotl(r, 15);
	q = r + _lrotl(s, 11);
#endif
	r = s + t;
	s = t + p;

	return s;
}

// https://msdn.microsoft.com/en-us/library/5cc576c4.aspx

unsigned long long	_rotl64(unsigned long long value, int shift);
unsigned long long	_rotr64(unsigned long long value, int shift);

#pragma intrinsic(_rotl64, _rotr64)

__inline		// small fast pseudo-random number generator
unsigned long long	prng64(void)	//  (Robert John "Bob" Jenkins)
{
#ifndef MMX
#if 1
	static	unsigned long long	p = 0x9E3779B97F4A7C15ULL;
#else
	static	unsigned long long	p = 0xF1EA5EEDULL;
#endif
	static	unsigned long long	q = ~0ULL;
	static	unsigned long long	r = ~0ULL;
	static	unsigned long long	s = ~0ULL;
	const	unsigned long long
	t = p - _rotl64(q, 7);
	p = q ^ _rotl64(r, 13);
#ifdef _WIN64
	q = r + _rotr64(s, 27);
#else
	q = r + _rotl64(s, 37);
#endif
	r = s + t;
	s = t + p;

	return s;
#else
#if 1
	static	__m64	p = {0x9E3779B97F4A7C15ULL};
#else
	static	__m64	p = {0xF1EA5EEDULL};
#endif
	static	__m64	q = {~0ULL};
	static	__m64	r = {~0ULL};
	static	__m64	s = {~0ULL};
	const	__m64
	t = _mm_sub_si64(p, _mm_or_si64(_mm_slli_si64(q, 7), _mm_srli_si64(q, 64 - 7)));
	p = _mm_xor_si64(q, _mm_or_si64(_mm_slli_si64(r, 13), _mm_srli_si64(r, 64 - 13)));
	q = _mm_add_si64(r, _mm_or_si64(_mm_slli_si64(s, 37), _mm_srli_si64(s, 64 - 37)));
	r = _mm_add_si64(s, t);
	s = _mm_add_si64(t, p);

	return s.m64_u64;
#endif // MMX
}

// https://msdn.microsoft.com/en-us/library/d2s81xt0.aspx

unsigned long long	__emulu(unsigned long multiplicand, unsigned long multiplier);

#pragma intrinsic(__emulu)

__inline	// returns value in interval [0, range)
unsigned long	range32(unsigned long range)
{
#if 0
	unsigned long	limit = (0 - range) % range;
	unsigned long	value;

	do
		value = prng32();
	while (value < limit);

	return value % range;
#elif 0
	unsigned long		limit;
	unsigned long long	multi = __emulu(prng32(), range);

	if (range > (unsigned long) multi)
		for (limit = (0 - range) % range;
		     limit > (unsigned long) multi;
		     multi = __emulu(prng32(), range))
			continue;

	return multi >> 32;
#else
	unsigned long		limit = 0 - range;
	unsigned long long	multi = __emulu(prng32(), range);

	if ((unsigned long) multi < range)
	{
		if (limit >= range)
		{
			limit -= range;

			if (limit >= range)
				limit %= range;
		}

		while ((unsigned long) multi < limit)
			multi = __emulu(prng32(), range);
	}

	return multi >> 32;
#endif
}

__inline		// returns value in interval [0, range)
unsigned long long	range64(unsigned long long range)
{
#ifndef _WIN64
	unsigned long long	limit = (0 - range) % range;
	unsigned long long	value;

	do
		value = prng64();
	while (value < limit);

	return value % range;
#elif 0
	unsigned long long	limit;
	unsigned long long	high;
	unsigned long long	low = _umul128(prng64(), range, &high);

	if (low < range)
		for (limit = (0 - range) % range;
		     low < limit;
		     low = _umul128(prng64(), range, &high))
			continue;

	return high;
#else
	unsigned long long	limit = 0 - range;
	unsigned long long	high;
	unsigned long long	low = _umul128(prng64(), range, &high);

	if (low < range)
	{
		if (limit >= range)
		{
			limit -= range;

			if (limit >= range)
				limit %= range;
		}

		while (low < limit)
			low = _umul128(prng64(), range, &high);
	}

	return high;
#endif // _WIN64
}

// https://msdn.microsoft.com/en-us/library/t9zea13t.aspx

// https://msdn.microsoft.com/en-us/library/5wcd48xw.aspx

__inline
int	isasciiA(char cb)
{
#if 0
	return (cb & '\x80') == '\0';
#else
	return (signed char) cb < '\0';
#endif
}

#undef isascii
#define isascii		isasciiA

__inline
int	isasciiW(wchar_t cw)
{
	return (cw & L'\xFF80') == L'\0';
}

#undef iswascii
#define iswascii	isasciiW

// https://msdn.microsoft.com/en-us/library/k84c0490.aspx

__inline
int	isalnumA(char cb)
{
	unsigned short	us;

	// https://msdn.microsoft.com/en-us/library/dd318118.aspx

	return GetStringTypeExA(LOCALE_USER_DEFAULT, CT_CTYPE1, &cb, 1, &us)
	    && (us & (C1_ALPHA | C1_DIGIT));
}

#define isalnum		isalnumA

__inline
int	isalnumW(wchar_t cw)
{
	unsigned short	us;

	// https://msdn.microsoft.com/en-us/library/dd318118.aspx

	return GetStringTypeExW(LOCALE_USER_DEFAULT, CT_CTYPE1, &cw, 1, &us)
	    && (us & (C1_ALPHA | C1_DIGIT));
}

#undef iswalnum
#define iswalnum	isalnumW

// https://msdn.microsoft.com/en-us/library/xt82b8z8.aspx

__inline
int	isalphaA(char cb)
{
	unsigned short	us;

	// https://msdn.microsoft.com/en-us/library/dd318118.aspx

	return GetStringTypeExA(LOCALE_USER_DEFAULT, CT_CTYPE1, &cb, 1, &us)
	    && (us & C1_ALPHA);
}

#define isalpha		isalphaA

__inline
int	isalphaW(wchar_t cw)
{
	unsigned short	us;

	// https://msdn.microsoft.com/en-us/library/dd318118.aspx

	return GetStringTypeExW(LOCALE_USER_DEFAULT, CT_CTYPE1, &cw, 1, &us)
	    && (us & C1_ALPHA);
}

#undef iswalpha
#define iswalpha	isalphaW

// https://msdn.microsoft.com/en-us/library/dn237255.aspx

__inline
int	isblankA(char cb)
{
	unsigned short	us;

	// https://msdn.microsoft.com/en-us/library/dd318118.aspx

	return GetStringTypeExA(LOCALE_USER_DEFAULT, CT_CTYPE1, &cb, 1, &us)
	    && (us & C1_BLANK);
}

#define isblank		isblankA

__inline
int	isblankW(wchar_t cw)
{
	unsigned short	us;

	// https://msdn.microsoft.com/en-us/library/dd318118.aspx

	return GetStringTypeExW(LOCALE_USER_DEFAULT, CT_CTYPE1, &cw, 1, &us)
	    && (us & C1_BLANK);
}

#undef iswblank
#define iswblank	isblankW

// https://msdn.microsoft.com/en-us/library/1ad6a790.aspx

__inline
int	iscntrlA(char cb)
{
	unsigned short	us;

	// https://msdn.microsoft.com/en-us/library/dd318118.aspx

	return GetStringTypeExA(LOCALE_USER_DEFAULT, CT_CTYPE1, &cb, 1, &us)
	    && (us & C1_CNTRL);
}

#define iscntrl		iscntrlA

__inline
int	iscntrlW(wchar_t cw)
{
	unsigned short	us;

	// https://msdn.microsoft.com/en-us/library/dd318118.aspx

	return GetStringTypeExW(LOCALE_USER_DEFAULT, CT_CTYPE1, &cw, 1, &us)
	    && (us & C1_CNTRL);
}

#undef iswcntrl
#define iswcntrl	iscntrlW

// https://msdn.microsoft.com/en-us/library/fcc4ksh8.aspx

__inline
int	isdigitA(char cb)
{
	unsigned short	us;

	// https://msdn.microsoft.com/en-us/library/dd318118.aspx

	return GetStringTypeExA(LOCALE_USER_DEFAULT, CT_CTYPE1, &cb, 1, &us)
	    && (us & C1_DIGIT);
}

#define isdigit		isdigitA

__inline
int	isdigitW(wchar_t cw)
{
	unsigned short	us;

	// https://msdn.microsoft.com/en-us/library/dd318118.aspx

	return GetStringTypeExW(LOCALE_USER_DEFAULT, CT_CTYPE1, &cw, 1, &us)
	    && (us & C1_DIGIT);
}

#undef iswdigit
#define iswdigit	isdigitW

// https://msdn.microsoft.com/en-us/library/wsfaff19.aspx

__inline
int	isgraphA(char cb)
{
	unsigned short	us;

	// https://msdn.microsoft.com/en-us/library/dd318118.aspx

	return GetStringTypeExA(LOCALE_USER_DEFAULT, CT_CTYPE1, &cb, 1, &us)
	    && (us & (C1_UPPER | C1_LOWER | C1_DIGIT | C1_PUNCT | C1_XDIGIT | C1_ALPHA | C1_DEFINED));
}

#define isgraph		isgraphA

__inline
int	isgraphW(wchar_t cw)
{
	unsigned short	us;

	// https://msdn.microsoft.com/en-us/library/dd318118.aspx

	return GetStringTypeExW(LOCALE_USER_DEFAULT, CT_CTYPE1, &cw, 1, &us)
	    && (us & (C1_UPPER | C1_LOWER | C1_DIGIT | C1_PUNCT | C1_XDIGIT | C1_ALPHA | C1_DEFINED));
}

#undef iswgraph
#define iswgraph	isgraphW

// https://msdn.microsoft.com/en-us/library/1z2s6by9.aspx

__inline
int	islowerA(char cb)
{
	unsigned short	us;

	// https://msdn.microsoft.com/en-us/library/dd318118.aspx

	return GetStringTypeExA(LOCALE_USER_DEFAULT, CT_CTYPE1, &cb, 1, &us)
	    && (us & C1_LOWER);
}

#define islower		islowerA

__inline
int	islowerW(wchar_t cw)
{
	unsigned short	us;

	// https://msdn.microsoft.com/en-us/library/dd318118.aspx

	return GetStringTypeExW(LOCALE_USER_DEFAULT, CT_CTYPE1, &cw, 1, &us)
	    && (us & C1_LOWER);
}

#undef iswlower
#define iswlower	islowerW

// https://msdn.microsoft.com/en-us/library/ewx8s4kw.aspx

__inline
int	isprintA(char cb)
{
	unsigned short	us;

	// https://msdn.microsoft.com/en-us/library/dd318118.aspx

	return GetStringTypeExA(LOCALE_USER_DEFAULT, CT_CTYPE1, &cb, 1, &us)
	    && (us & (C1_UPPER | C1_LOWER | C1_DIGIT | C1_SPACE | C1_PUNCT | C1_BLANK | C1_XDIGIT | C1_ALPHA | C1_DEFINED));
}

#define isprint		isprintA

__inline
int	isprintW(wchar_t cw)
{
	unsigned short	us;

	// https://msdn.microsoft.com/en-us/library/dd318118.aspx

	return GetStringTypeExW(LOCALE_USER_DEFAULT, CT_CTYPE1, &cw, 1, &us)
	    && (us & (C1_UPPER | C1_LOWER | C1_DIGIT | C1_SPACE | C1_PUNCT | C1_BLANK | C1_XDIGIT | C1_ALPHA | C1_DEFINED));
}

#undef iswprint
#define iswprint	isprintW

// https://msdn.microsoft.com/en-us/library/0ce6f2zc.aspx

__inline
int	ispunctA(char cb)
{
	unsigned short	us;

	// https://msdn.microsoft.com/en-us/library/dd318118.aspx

	return GetStringTypeExA(LOCALE_USER_DEFAULT, CT_CTYPE1, &cb, 1, &us)
	    && (us & C1_PUNCT);
}

#define ispunct		ispunctA

__inline
int	ispunctW(wchar_t cw)
{
	unsigned short	us;

	// https://msdn.microsoft.com/en-us/library/dd318118.aspx

	return GetStringTypeExW(LOCALE_USER_DEFAULT, CT_CTYPE1, &cw, 1, &us)
	    && (us & C1_PUNCT);
}

#undef iswpunct
#define iswpunct	ispunctW

// https://msdn.microsoft.com/en-us/library/y13z34da.aspx

__inline
int	isspaceA(char cb)
{
	unsigned short	us;

	// https://msdn.microsoft.com/en-us/library/dd318118.aspx

	return GetStringTypeExA(LOCALE_USER_DEFAULT, CT_CTYPE1, &cb, 1, &us)
	    && (us & C1_SPACE);
}

#define isspace		isspaceA

__inline
int	isspaceW(wchar_t cw)
{
	unsigned short	us;

	// https://msdn.microsoft.com/en-us/library/dd318118.aspx

	return GetStringTypeExW(LOCALE_USER_DEFAULT, CT_CTYPE1, &cw, 1, &us)
	    && (us & C1_SPACE);
}

#undef iswspace
#define iswspace	isspaceW

// https://msdn.microsoft.com/en-us/library/htd1kwta.aspx

__inline
int	isupperA(char cb)
{
	unsigned short	us;

	// https://msdn.microsoft.com/en-us/library/dd318118.aspx

	return GetStringTypeExA(LOCALE_USER_DEFAULT, CT_CTYPE1, &cb, 1, &us)
	    && (us & C1_UPPER);
}

#define isupper		isupperA

__inline
int	isupperW(wchar_t cw)
{
	unsigned short	us;

	// https://msdn.microsoft.com/en-us/library/dd318118.aspx

	return GetStringTypeExW(LOCALE_USER_DEFAULT, CT_CTYPE1, &cw, 1, &us)
	    && (us & C1_UPPER);
}

#undef iswupper
#define iswupper	isupperW

// https://msdn.microsoft.com/en-us/library/t6h81hwd.aspx

__inline
int	isxdigitA(char cb)
{
	unsigned short	us;

	// https://msdn.microsoft.com/en-us/library/dd318118.aspx

	return GetStringTypeExA(LOCALE_USER_DEFAULT, CT_CTYPE1, &cb, 1, &us)
	    && (us & C1_XDIGIT);
}

#define isxdigit	isxdigitA

__inline
int	isxdigitW(wchar_t cw)
{
	unsigned short	us;

	// https://msdn.microsoft.com/en-us/library/dd318118.aspx

	return GetStringTypeExW(LOCALE_USER_DEFAULT, CT_CTYPE1, &cw, 1, &us)
	    && (us & C1_XDIGIT);
}

#undef iswxdigit
#define iswxdigit	isxdigitW

__inline
int	issurrogateW(wchar_t cw)
{
#if 0
	unsigned short	us;

	// https://msdn.microsoft.com/en-us/library/dd318118.aspx

	return GetStringTypeExW(LOCALE_USER_DEFAULT, CT_CTYPE3, &cw, 1, &us)
	    && (us & (C3_HIGHSURROGATE | C3_LOWSURROGATE));
#elif 0
	return (cw >= L'\xD800')
	    && (cw <= L'\xDFFF');
#elif 0
	return (cw & L'\xD800') == L'\xD800';
#else
	return cw - L'\xD800' <= L'\xDFFF' - L'\xD800';
#endif
}

#define iswsurrogate	issurrogateW

__inline
int	ishighsurrogateW(wchar_t cw)
{
#if 0
	unsigned short	us;

	// https://msdn.microsoft.com/en-us/library/dd318118.aspx

	return GetStringTypeExW(LOCALE_USER_DEFAULT, CT_CTYPE3, &cw, 1, &us)
	    && (us & C3_HIGHSURROGATE);
#elif 0
	return (cw >= L'\xD800') && (cw <= L'\xDBFF');
#elif 0
	return (cw - L'\xD800') <= (L'\xDBFF' - L'\xD800');
#else
	return (cw & L'\xDC00') == L'\xD800';
#endif
}

#define iswhighsurrogate	ishighsurrogateW

__inline
int	islowsurrogateW(wchar_t cw)
{
#if 0
	unsigned short	us;

	// https://msdn.microsoft.com/en-us/library/dd318118.aspx

	return GetStringTypeExW(LOCALE_USER_DEFAULT, CT_CTYPE3, &cw, 1, &us)
	    && (us & C3_LOWSURROGATE);
#elif 0
	return (cw >= L'\xDC00') && (cw <= L'\xDFFF');
#elif 0
	return (cw - L'\xDC00') <= (L'\xDFFF' - L'\xDC00');
#else
	return (cw & L'\xDC00') == L'\xDC00';
#endif
}

#define iswlowsurrogate	islowsurrogateW

// https://msdn.microsoft.com/en-us/library/dd318683.aspx

#ifndef IS_HIGH_SURROGATE
#define IS_HIGH_SURROGATE	((cw - L'\xD800') < L'\x0400')
#endif

// https://msdn.microsoft.com/en-us/library/dd318686.aspx

#ifndef IS_LOW_SURROGATE
#define IS_LOW_SURROGATE	((cw - L'\xDC00') < L'\x0400')
#endif

// https://msdn.microsoft.com/en-us/library/dd318687.aspx

#ifndef IS_SURROGATE_PAIR
#define IS_SURROGATE_PAIR(cwHigh, cwLow)	(((cwHigh - L'\xD800') | (cwLow - L'\xDC00')) < L'\x0400')
#endif

#ifdef _M_IX86
__declspec(naked)
int	__fastcall	iswhitespaceA(char cb)
{
	__asm
	{
		xor	eax, eax
		cmp	eax, ecx	; CF = (cb > '\0')
		adc	eax, 3E00h	; eax = (cb > '\0') | (1 << '\r') | (1 << '\f')
					;     | (1 << '\v') | (1 << '\n') | (1 << '\t')
		shr	eax, cl		; eax = eax >> (cb % ' ')
		xor	edx, edx
		cmp	ecx, 33		; CF = (cb <= ' ')
		adc	edx, edx	; edx = (cb <= ' ') ? 1 : 0
		and	eax, edx
		ret
	}
}
#elif 0
__inline
int	iswhitespaceA(char cb)
{
	return (cb == ' ') || (cb == '\r') || (cb == '\f')
	    || (cb == '\v') || (cb == '\n') || (cb == '\t');
}
#endif

#ifdef _M_IX86
__declspec(naked)
int	__fastcall	iswhitespaceW(wchar_t cw)
{
	__asm
	{
		xor	eax, eax
		cmp	eax, ecx	; CF = (cw > L'\0')
		adc	eax, 3E00h	; eax = (cw > L'\0') | (1 << L'\r') | (1 << L'\f')
					;     | (1 << L'\v') | (1 << L'\n') | (1 << L'\t')
		shr	eax, cl		; eax = eax >> (cb % L' ')
		xor	edx, edx
		cmp	ecx, 33		; CF = (cw <= L' ')
		adc	edx, edx	; edx = (cw <= L' ') ? 1 : 0
		and	eax, edx
		ret
	}
}
#elif 0
__inline
int	iswhitespaceW(wchar_t cw)
{
	return (cw == L' ') || (cw == L'\r') || (cw == L'\f')
	    || (cw == L'\v') || (cw == L'\n') || (cw == L'\t');
}
#endif

#ifdef _M_IX86
// https://msdn.microsoft.com/en-us/library/d2s81xt0.aspx

unsigned long long	__emulu(unsigned long multiplicand, unsigned long multiplier);

#pragma intrinsic(__emulu)
#elif _M_AMD64
// https://msdn.microsoft.com/en-us/library/67xf5cy5.aspx

unsigned long long	__umulh(unsigned long long multiplicand, unsigned long long multiplier);

#pragma intrinsic(__umulh)
#endif

// https://msdn.microsoft.com/en-us/library/3f8w183e.aspx

__declspec(noalias) __declspec(restrict)
__inline
void	*calloc(size_t count,
	        size_t size)
{
	extern	const	HANDLE	hHeap;

	LPVOID	block;
#ifdef _M_IX86
	if ((__emulu(count, size) >> 32) != 0)
		return NULL;
#elif _M_AMD64
	if (__umulh(count, size) != 0)
		return NULL;
#elif _WIN64
	if (~0ULL / count < size)
		return NULL;
#else
	if (~0UL / count < size)
		return NULL;
#endif
	// https://msdn.microsoft.com/en-us/library/aa366597.aspx

	block = HeapAlloc(hHeap, HEAP_ZERO_MEMORY, count * size);

	if (block == NULL)

		// https://msdn.microsoft.com/en-us/library/aa363362.aspx

		OutputDebugStringA("NOMSVCRT: HeapAlloc() returned NULL\n");

	return block;
}

// https://msdn.microsoft.com/en-us/library/6ewkz86d.aspx

__declspec(noalias) __declspec(restrict)
__inline
void	*malloc(size_t size)
{
	extern	const	HANDLE	hHeap;

	// https://msdn.microsoft.com/en-us/library/aa366597.aspx

	LPVOID	block = HeapAlloc(hHeap, 0, size);

	if (block == NULL)

		// https://msdn.microsoft.com/en-us/library/aa363362.aspx

		OutputDebugStringA("NOMSVCRT: HeapAlloc() returned NULL\n");

	return block;
}

// https://msdn.microsoft.com/en-us/library/xbebcx7d.aspx

__declspec(noalias) __declspec(restrict)
__inline
void	*realloc(void   *block,
	         size_t size)
{
	extern	const	HANDLE	hHeap;

	if (block == NULL)
	{
		// https://msdn.microsoft.com/en-us/library/aa366597.aspx

		block = HeapAlloc(hHeap, 0, size);

		if (block == NULL)

			// https://msdn.microsoft.com/en-us/library/aa363362.aspx

			OutputDebugStringA("NOMSVCRT: HeapAlloc() returned NULL\n");
	}
	else
		if (size == 0)
		{
			// https://msdn.microsoft.com/en-us/library/aa366701.aspx

			if (!HeapFree(hHeap, 0, block))
				DebugFormatA("NOMSVCRT: HeapFree() returned error %lu\n", GetLastError());

			block = NULL;
		}
		else
		{	// https://msdn.microsoft.com/en-us/library/aa366704.aspx

			block = HeapReAlloc(hHeap, 0, block, size);

			if (block == NULL)

				// https://msdn.microsoft.com/en-us/library/aa363362.aspx

				OutputDebugStringA("NOMSVCRT: HeapReAlloc() returned NULL\n");
		}

	return block;
}

// https://msdn.microsoft.com/en-us/library/we1whae7.aspx

__declspec(noalias)
__inline
void	free(void *block)
{
	extern	const	HANDLE	hHeap;

	if (block != NULL)

		// https://msdn.microsoft.com/en-us/library/aa366701.aspx

		if (!HeapFree(hHeap, 0, block))
			DebugFormatA("NOMSVCRT: HeapFree() returned error %lu\n", GetLastError());
}

#ifdef __cplusplus
}

// https://msdn.microsoft.com/en-us/library/kftdy56f.aspx

// https://msdn.microsoft.com/en-us/library/t48aek43.aspx

__inline
void *	operator new(size_t size)
{
	extern	const	HANDLE	hHeap;

	// https://msdn.microsoft.com/en-us/library/aa366597.aspx
#ifdef THROW
	return HeapAlloc(hHeap, HEAP_GENERATE_EXCEPTIONS, size);
#else
	LPVOID	block = HeapAlloc(hHeap, 0, size);

	if (block == NULL)

		// https://msdn.microsoft.com/en-us/library/aa363362.aspx

		OutputDebugStringA("NOMSVCRT: HeapAlloc() returned NULL\n");

	return block;
#endif
}

// https://msdn.microsoft.com/en-us/library/t48aek43.aspx

__inline
void *	operator new[](size_t size)
{
	extern	const	HANDLE	hHeap;

	// https://msdn.microsoft.com/en-us/library/aa366597.aspx
#ifdef THROW
	return HeapAlloc(hHeap, HEAP_GENERATE_EXCEPTIONS, size);
#else
	LPVOID	block = HeapAlloc(hHeap, 0, size);

	if (block == NULL)

		// https://msdn.microsoft.com/en-us/library/aa363362.aspx

		OutputDebugStringA("NOMSVCRT: HeapAlloc() returned NULL\n");

	return block;
#endif
}

// https://msdn.microsoft.com/en-us/library/248aa748.aspx

__inline
void	operator delete(void *block)
{
	extern	const	HANDLE	hHeap;

	if (block != NULL)

		// https://msdn.microsoft.com/en-us/library/aa366701.aspx

		if (!HeapFree(hHeap, 0, block))
			DebugFormatA("NOMSVCRT: HeapFree() returned error %lu\n", GetLastError());
}

// https://msdn.microsoft.com/en-us/library/248aa748.aspx

__inline
void	operator delete[](void *block)
{
	extern	const	HANDLE	hHeap;

	if (block != NULL)

		// https://msdn.microsoft.com/en-us/library/aa366701.aspx

		if (!HeapFree(hHeap, 0, block))
			DebugFormatA("NOMSVCRT: HeapFree() returned error %lu\n", GetLastError());
}
#endif // __cplusplus
#endif // __NOMSVCRT_H__
